/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.adapters.input;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.dcg.p2p.exceptions.InvalidFormatException;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.utilities.ServerSocketListener;

/**
 * <code>InputAdapter</code> to get data from a Shoutcast server.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see InputAdapter
 * @see VlcHttpAdapter
 */
public class ShoutCastAdapter extends InputAdapter {

	protected final Log logger = LogFactory.getLog( this.getClass() );

	/**
	 * The owner of this adapter.
	 */
	private Node source;

	/**
	 * URL of the Shoutcast server.
	 */
	private URL url;
	private InputStream stationIn;

	/**
	 * The local port to which a mp3 media player can connect to listen to the
	 * received stream.
	 */
	private int localPort;
	private ServerSocketListener socketListener;

	private ICYInfo icy = new ICYInfo();
	private boolean icySent = false;

	private boolean running = false;


	/**
	 * Creates a new <code>InputAdapter</code> for Shoutcast streams.<br>
	 * 
	 * @param url
	 *            a string representation of the URL of the Shoutcast station<br>
	 *            (example: "http://64.236.34.97:80/stream/1074")
	 * @param localPort
	 *            the local port to which the stream will be forwarded if a
	 *            media player is connected
	 * @param source
	 *            the owner of this <code>ShoutCastAdapter</code>
	 */
	public ShoutCastAdapter( String url, int localPort, Node source ){

		try{
			this.url = new URL( url );
			this.localPort = localPort;
			this.source = source;
		}catch( MalformedURLException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( "invalid url:" + url );
			}
		}
	}


	/**
	 * Creates a new <code>InputAdapter</code> for Shoutcast streams.<br>
	 * 
	 * @param url
	 *            the URL of the Shoutcast station<br>
	 * @param localPort
	 *            the local port to which the stream will be forwarded if a
	 *            media player is connected
	 * @param source
	 *            the owner of this <code>ShoutCastAdapter</code>
	 */
	public ShoutCastAdapter( URL url, int localPort, Node source ){

		this.url = url;
		this.localPort = localPort;
		this.source = source;
	}


	/**
	 * Starts this <code>InputAdapter</code>.<br>
	 * The data received is expected to start with an icy message which will be
	 * parsed. The Shoutcast stream thereafter will be forwarded to the target
	 * <code>Packetizer</code> as well as to a local port to which a media
	 * player can connect and listen to what is received.<br>
	 * ICY information will be provided on the local socket only.
	 */
	public void run(){

		int len;
		byte[] buffer = new byte[24000];
		byte[] icyBuffer = new byte[4096];
		int crlf = 0, icyLen = 0;
		byte b;

		running = true;

		// wait for incoming local connection using a thread
		socketListener = new ServerSocketListener( localPort, source, this );
		socketListener.start();

		try{
			// open stream from shoutcast station
			stationIn = url.openStream();
			if( logger.isDebugEnabled() ){
				logger.debug( "connection to shoutcast station established" );
			}
			// read and parse ICY header
			if( logger.isDebugEnabled() ){
				logger.debug( "receiving icy header" );
			}
			while( ( crlf < 4 ) && ( stationIn.available() != -1 ) ){
				b = (byte)stationIn.read();
				if( ( b == 13 ) || ( b == 10 ) ){
					crlf++;
				}else{
					crlf = 0;
				}
				icyBuffer[icyLen] = b;
				icyLen++;
			}
			try{
				parseIcy( icyBuffer, icyLen - 2 );
			}catch( InvalidFormatException e ){
				if( logger.isErrorEnabled() ){
					logger.error( "cannot parse icy streaming information for " + url.getHost() + "; stream does not seem to have valid ICY information" );
				}
			}
			// read mp3 data
			if( logger.isDebugEnabled() ){
				logger.debug( "receiving mp3 data" );
			}
			while( stationIn.available() != -1 && running ){
				len = stationIn.read( buffer );
				// stream to local socket if available
				if( socketListener.out != null ){
					try{
						if( !icySent ){
							icySent = icy.sendIcy( socketListener.out );
						}
						if( icySent ){
							socketListener.out.write( buffer, 0, len );
						}
					}catch( Exception e ){
						// reset connection to media player
						socketListener.out = null;
						icySent = false;
					}
				}
				// stream to the Packetizer
				if( target != null ){
					target.addData( buffer, len );
				}
			}
			if( logger.isDebugEnabled() ){
				logger.debug( "stream reception stopped or no more data available" );
			}
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}


	/**
	 * Parses the ICY information which was received from the Shoutcast station.
	 * 
	 * @param buffer
	 *            byte array containing the received bytes which are supposed to
	 *            be the ICY information
	 * @param buffLen
	 *            the length of the buffer argument
	 * @throws InvalidFormatException
	 *             if the buffer does not match the ICY pattern
	 */
	private void parseIcy( byte[] buffer, int buffLen ) throws InvalidFormatException{

		if( logger.isDebugEnabled() ){
			logger.debug( "parsing icy header" );
		}

		byte[] tmp = new byte[buffLen];
		int buffPos = 0, curLen = 0;
		// read status (must be 'ICY 200 OK')
		curLen = readUntil( 13, 0, buffer, tmp );
		buffPos += curLen + 2; // skip 'CRLF'
		if( !new String( tmp ).substring( 0, curLen ).equalsIgnoreCase( "icy 200 ok" ) ){
			throw new InvalidFormatException( "icy message invalid - not a valid mp3 stream" );
		}
		// parse ICY message
		while( buffPos < buffLen ){
			// read to next ":"
			curLen = readUntil( 58, buffPos, buffer, tmp );
			buffPos += curLen + 1; // skip ':'
			String headerField = new String( tmp ).substring( 0, curLen );
			// read to next CRLF
			curLen = readUntil( 13, buffPos, buffer, tmp );
			buffPos += curLen + 2; // skip 'CRLF'
			String headerFieldValue = new String( tmp ).substring( 0, curLen );
			// set corresponding field in icy object
			if( headerField.equalsIgnoreCase( "icy-notice1" ) ){
				icy.notice1 = headerFieldValue;
			}else if( headerField.equalsIgnoreCase( "icy-notice2" ) ){
				icy.notice2 = headerFieldValue;
			}else if( headerField.equalsIgnoreCase( "icy-notice3" ) ){
				icy.notice3 = headerFieldValue;
			}else if( headerField.equalsIgnoreCase( "icy-name" ) ){
				icy.name = headerFieldValue;
			}else if( headerField.equalsIgnoreCase( "icy-genre" ) ){
				icy.genre = headerFieldValue;
			}else if( headerField.equalsIgnoreCase( "icy-url" ) ){
				icy.url = headerFieldValue;
			}else if( headerField.equalsIgnoreCase( "content-type" ) ){
				icy.contentType = headerFieldValue;
			}else if( headerField.equalsIgnoreCase( "icy-pub" ) ){
				try{
					icy.pub = Integer.parseInt( headerFieldValue );
				}catch( NumberFormatException e ){
					throw new InvalidFormatException( "icy message invalid - " + e.getMessage() );
				}
			}else if( headerField.equalsIgnoreCase( "icy-br" ) ){
				try{
					icy.br = Integer.parseInt( headerFieldValue );
				}catch( NumberFormatException e ){
					throw new InvalidFormatException( "icy message invalid - " + e.getMessage() );
				}
			}else{ // unknown header info
				curLen = skipUntil( 13, buffPos, buffer );
				buffPos += curLen + 2; // skip CRLF
			}
		}
	}


	/**
	 * Reads to the next target character from the src array and copies the
	 * bytes into the dest array.
	 * 
	 * @param target
	 *            the character at which to stop reading
	 * @param offset
	 *            the position from where to read inside the src array
	 * @param src
	 *            byte array to read from
	 * @param dest
	 *            byte array to write to
	 * @return the number of read bytes
	 * @throws InvalidFormatException
	 *             if the target character cannot be found
	 */
	private int readUntil( int target, int offset, byte[] src, byte[] dest ) throws InvalidFormatException{

		try{
			int j = 0;
			while( src[offset] != target ){
				dest[j++] = src[offset++];
			}
			return j;
		}catch( Exception e ){
			throw new InvalidFormatException( "icy message invalid - " + e.getMessage() );
		}
	}


	/**
	 * Skips all bytes until the target byte is found.
	 * 
	 * @param target
	 *            the character until which bytes should be skipped
	 * @param offset
	 *            the position from where to skip inside the src array
	 * @param src
	 *            byte array in which to skip bytes
	 * @return the number of bytes skipped
	 * @throws InvalidFormatException
	 *             if the target character cannot be found
	 */
	private int skipUntil( int target, int offset, byte[] src ) throws InvalidFormatException{

		try{
			int i = 0;
			while( src[offset] != target ){
				offset++;
				i++;
			}
			return i;
		}catch( Exception e ){
			throw new InvalidFormatException( "icy message invalid - " + e.getMessage() );
		}
	}


	/**
	 * Stops this <code>ShoutCastAdapter</code> if it is run as a thread.<br>
	 * The <code>ServerSocketListener</code> and the <code>InputStream</code>
	 * will be closed.
	 */
	public void stop(){

		running = false;
		try{
			socketListener.close();
			stationIn.close();
		}catch( Exception e ){
		}
	}


	@Override
	public void resendContentInfo(){

		icySent = false;
	}


	/**
	 * Returns the local port number to which a media player can connect.
	 * 
	 * @return the local media port
	 */
	public int getLocalPort(){

		return localPort;
	}


	/**
	 * Sets the local port number to which a media player can connect.
	 * 
	 * @param localPort
	 *            the local media port
	 */
	public void setLocalPort( int localPort ){

		this.localPort = localPort;
	}
}
