/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.adapters.input;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * <code>InputAdapter</code> to get data from a VLC media player which
 * forwards the stream using the HTTP protocol.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see InputAdapter
 * @see ShoutCastAdapter
 */
public class VlcHttpAdapter extends InputAdapter {

	protected final Log logger = LogFactory.getLog( this.getClass() );

	private URL url;
	private InputStream in;

	private boolean running = false;


	/**
	 * Creates a new <code>InputAdapter</code> for HTTP streams forwarded by
	 * VLC.
	 * 
	 * @param url
	 *            the string representation of the URL of the VLC media player
	 *            which is the source of the stream<br>
	 *            (example: "http://localhost:1234")
	 */
	public VlcHttpAdapter( String url ){

		try{
			this.url = new URL( url );
		}catch( MalformedURLException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( "invalid url:" + url );
			}
		}
	}


	/**
	 * Creates a new <code>InputAdapter</code> for HTTP streams forwarded by
	 * VLC.
	 * 
	 * @param url
	 *            the URL of the VLC media player which is the source of the
	 *            stream<br>
	 */
	public VlcHttpAdapter( URL url ){

		this.url = url;
	}


	/**
	 * Starts this <code>InputAdapter</code>.<br>
	 * Receives data from the VLC client and forwards it to the Packetizer.
	 */
	public void run(){

		int len;
		byte[] buffer = new byte[900000];

		running = true;

		try{
			// open stream from VLC client
			in = url.openStream();
			if( logger.isDebugEnabled() ){
				logger.debug( "connection to VLC client established; reading data" );
			}
			while( in.available() != -1 && running ){
				len = in.read( buffer );
				// stream to the packetizer
				if( target != null ){
					target.addData( buffer, len );
				}
			}
			if( logger.isDebugEnabled() ){
				logger.debug( "stream reception stopped; no more data available" );
			}
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}


	/**
	 * Stops this <code>VlcHttpAdapter</code> if it is run as a thread.<br>
	 * The <code>InputStream</code> will be closed.
	 */
	public void stop(){

		running = false;
		try{
			in.close();
		}catch( Exception e ){
		}
	}


	@Override
	public void resendContentInfo(){

		return;
	}

}
