/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.adapters.output;

import java.io.IOException;

import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.utilities.ServerSocketListener;

/**
 * <code>OutputAdapter</code> to forward the media stream to an mp3 media
 * player.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see OutputAdapter
 * @see FileWriterAdapter
 * @see VlcAdapter
 */
public class Mp3Adapter extends OutputAdapter {

	private boolean icySent = false;
	private Node source;


	/**
	 * Creates a new <code>Mp3Adapter</code> and starts listening for incoming
	 * connections from a media player.
	 * 
	 * @param listeningPort
	 *            the port to which an mp3 media player will connect
	 * @param source
	 *            the owner of this <code>OutputAdapter</code>
	 */
	public Mp3Adapter( int listeningPort, Node source ){

		socketListener = new ServerSocketListener( listeningPort, source, this );
		socketListener.start();
		this.source = source;
	}


	@Override
	public void dataAvailable( byte[] data, int len ){

		if( !icySent && socketListener.out != null ){
			sendIcy();
		}
		if( icySent && socketListener.out != null ){
			try{
				socketListener.out.write( data, 0, len );
			}catch( IOException e ){
				if( logger.isErrorEnabled() ){
					logger.error( "cannot sent data to the local socket; closing it" );
				}
				icySent = false;
				socketListener.out = null;
			}
		}
	}


	/**
	 * Sends the ICY information to the media player. It does not forward the ICY information
	 * received from the source but a custom information.
	 */
	private void sendIcy(){

		byte[] crlf = new byte[] { 13, 10 };
		try{
			socketListener.out.write( new String( "ICY 200 OK" ).getBytes() );
			socketListener.out.write( crlf );
			socketListener.out.write( new String( "icy-notice1:<BR>This stream is designed for Winamp<BR>" ).getBytes() );
			socketListener.out.write( crlf );
			socketListener.out.write( new String( "icy-name:Swistry - P2P Streaming by dcg.ethz.ch" ).getBytes() );
			socketListener.out.write( crlf );
			socketListener.out.write( new String( "content-type:audio/mpeg" ).getBytes() );
			socketListener.out.write( crlf );
			socketListener.out.write( crlf );
			icySent = true;
			if( logger.isDebugEnabled() ){
				logger.debug( "icy message sent to local socket" );
			}
			source.write2gui( "content information sent to media player" );
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( "cannot send icy message to local socket; closing socket" );
			}
			source.write2gui( "cannot send content information to media player; closing media-socket" );
			icySent = false;
			socketListener.out = null;
		}
	}


	@Override
	public void resendContentInfo(){

		icySent = false;
	}

}
