/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.adapters.output;

import java.io.IOException;

import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.utilities.ServerSocketListener;

/**
 * <code>OutputAdapter</code> to forward the media stream to a VLC media
 * player.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see OutputAdapter
 * @see FileWriterAdapter
 * @see Mp3Adapter
 */
public class VlcAdapter extends OutputAdapter {

	private byte[] buffer;
	private int bufferPos = 0;
	private boolean bufferFull = false;
	private boolean contentInfoSent = false;
	private Node source;


	/**
	 * Creates a new <code>VlcAdapter</code>.
	 * 
	 * @param listeningPort
	 *            port to which the VLC media player connects
	 * @param bufferSize
	 *            the number of KiloBytes that are stored in a ring buffer until
	 *            the data is forwarded to the VLC media player; this results in
	 *            a higher playback delay because the buffered data is presented
	 *            first but at least there will be enough data available in
	 *            order to prevent lags<br>
	 *            the bufferSize must be > 0; otherwise no buffer will be used
	 */
	public VlcAdapter( int listeningPort, int bufferSize, Node source ){

		if( bufferSize > 0 ){
			buffer = new byte[bufferSize * 1000];
		}
		socketListener = new ServerSocketListener( listeningPort, source, this );
		socketListener.start();
		this.source = source;
	}


	/**
	 * Adds data to the ring buffer.<br>
	 * If a media player connects it will get the buffered data followed by the
	 * new data.
	 */
	public void dataAvailable( byte[] data, int len ){

		if( !contentInfoSent && socketListener.out != null ){
			sendContentInfo();
		}
		if( contentInfoSent && socketListener.out != null ){
			// forward data to the media player
			try{
				if( bufferPos > 0 || bufferFull ){
					// write buffered data first
					if( bufferFull ){
						socketListener.out.write( buffer, bufferPos, buffer.length - bufferPos );
					}
					if( bufferPos > 0 ){
						socketListener.out.write( buffer, 0, bufferPos );
					}
					// disable buffer
					bufferPos = 0;
					bufferFull = false;
				}
				// write new data
				socketListener.out.write( data, 0, len );
			}catch( IOException e ){
				if( logger.isErrorEnabled() ){
					logger.error( "cannot send data to the local socket; closing it. " + e.getMessage() );
				}
				contentInfoSent = false;
				socketListener.out = null;
			}
		}else if( socketListener.out == null ){
			// save data into the ring buffer
			int remainingSpace = buffer.length - bufferPos;
			int remainingData = len;
			int dataPos = 0;
			while( remainingData > 0 ){
				int toCopy = Math.min( remainingSpace, remainingData );
				System.arraycopy( data, dataPos, buffer, bufferPos, toCopy );
				if( ( bufferPos + toCopy ) >= buffer.length ){
					bufferFull = true;
				}
				bufferPos = ( bufferPos + toCopy ) % buffer.length;
				remainingSpace = buffer.length - bufferPos;
				remainingData -= toCopy;
				dataPos += toCopy;
			}
		}

	}


	/**
	 * Sends the content information to the media player.<br>
	 * VLC expects 'HTTP/1.0 200 OK' and Content-type: application/octet-stream
	 * 
	 */
	private void sendContentInfo(){

		byte[] crlf = new byte[] { 13, 10 };
		try{
			socketListener.out.write( new String( "HTTP/1.0 200 OK" ).getBytes() );
			socketListener.out.write( crlf );
			socketListener.out.write( new String( "Content-type: application/octet-stream" ).getBytes() );
			socketListener.out.write( crlf );
			socketListener.out.write( new String( "Cache-Control: no-cache" ).getBytes() );
			socketListener.out.write( crlf );
			socketListener.out.write( crlf );
			contentInfoSent = true;
			if( logger.isDebugEnabled() ){
				logger.debug( "content information sent to local socket" );
			}
			source.write2gui( "content information sent to media player" );
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( "cannot send content information to local socket; closing socket" );
			}
			source.write2gui( "cannot send content information to media player; closing media-socket" );
			contentInfoSent = false;
			socketListener.out = null;
		}
	}


	@Override
	public void resendContentInfo(){

		contentInfoSent = false;
	}

}
