/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.gui;

import java.net.URL;
import java.util.Calendar;

import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JOptionPane;

import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * Defines the GUI interface and implements a few methods.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see NEPGUI
 * @see PeerGUI
 * @see SourceGUI
 */
public abstract class GUI {

	protected JFrame frame;


	/**
	 * Adds a neighbour representation to the gui.
	 * 
	 * @param neighbourKey
	 *            the hash key representing the new neighbour which is 'ip:port'
	 * @return true if the neighbour representation was successfully created;
	 *         false otherwise
	 */
	public abstract boolean addNeighbour( String neighbourKey );


	/**
	 * Removes a neighbour representation from the gui.
	 * 
	 * @param neighbourKey
	 *            the hash key representing the neighbour which is 'ip:port'
	 * @return true if the neighbour representation was successfully removed;
	 *         false otherwise
	 */
	public abstract boolean removeNeighbour( String neighbourKey );


	/**
	 * Updates the status of the gui. This means that certain buttons and tabs
	 * are enabled or disabled according to the current state of the node which
	 * this gui represents.
	 */
	public abstract void updateStatus();


	/**
	 * Sets the specified text in red letters on the status bar at the bottom of
	 * the gui.
	 * 
	 * @param status
	 *            the text to write onto the status bar
	 */
	public abstract void errorStatusBar( String status );


	/**
	 * Adds a line to the top of the status tab.
	 * 
	 * @param line
	 *            the text to add to the status tab; it will be ended by a \n
	 */
	public abstract void addStatusLine( String line );


	/**
	 * Shows an error message dialog.
	 * 
	 * @param message
	 *            the error message to display in the dialog
	 * @param title
	 *            the title of the dialog
	 */
	public abstract void errorDialog( String message, String title );


	/**
	 * Updates the statistics tab in the gui.
	 */
	public abstract void updateTotals();


	/**
	 * Updates the number of packets received in the per neighbour statistics
	 * tab.
	 * 
	 * @param neighbourKey
	 *            the hash key of the neighbour representation which is
	 *            'ip:port'
	 * @param received
	 *            the current amount of received packets
	 * @return true if this neighbour exists and his values have been updated;
	 *         false otherwise
	 */
	public abstract boolean updatePacketsReceived( String neighbourKey, Long received );


	/**
	 * Updates the number of packets sent in the per neighbour statistics tab.
	 * 
	 * @param neighbourKey
	 *            the hash key of the neighbour representation which is
	 *            'ip:port'
	 * @param sent
	 *            the current amount of sent packets
	 * @return true if this neighbour exists and his values have been updated;
	 *         false otherwise
	 */
	public abstract boolean updatePacketsSent( String neighbourKey, Long sent );


	/**
	 * Updates the number of requests received in the per neighbour statistics
	 * tab.
	 * 
	 * @param neighbourKey
	 *            the hash key of the neighbour representation which is
	 *            'ip:port'
	 * @param requests
	 *            the current amount of received requests
	 * @return true if this neighbour exists and his values have been updated;
	 *         false otherwise
	 */
	public abstract boolean updateRequestsReceived( String neighbourKey, Long requests );


	/**
	 * Updates the number of requests sent in the per neighbour statistics tab.
	 * 
	 * @param neighbourKey
	 *            the hash key of the neighbour representation which is
	 *            'ip:port'
	 * @param requests
	 *            the current amount of sent requests
	 * @return true if this neighbour exists and his values have been updated;
	 *         false otherwise
	 */
	public abstract boolean updateRequestsSent( String neighbourKey, Long requests );


	/**
	 * Updates the number of notifications received in the per neighbour
	 * statistics tab.
	 * 
	 * @param neighbourKey
	 *            the hash key of the neighbour representation which is
	 *            'ip:port'
	 * @param notifications
	 *            the current amount of received notifications
	 * @return true if this neighbour exists and his values have been updated;
	 *         false otherwise
	 */
	public abstract boolean updateNotificationsReceived( String neighbourKey, Long notifications );


	/**
	 * Updates the number of notifications sent in the per neighbour statistics
	 * tab.
	 * 
	 * @param neighbourKey
	 *            the hash key of the neighbour representation which is
	 *            'ip:port'
	 * @param notifications
	 *            the current amount of sent notifications
	 * @return true if this neighbour exists and his values have been updated;
	 *         false otherwise
	 */
	public abstract boolean updateNotificationsSent( String neighbourKey, Long notifications );


	/**
	 * Updates the number of duplicate packets received in the per neighbour
	 * statistics tab.
	 * 
	 * @param dupPackets
	 *            the current amount of received duplicate packets
	 * @return true if this neighbour exists and his values have been updated;
	 *         false otherwise
	 */
	public abstract boolean updateDuplicatePacketsReceived( Long dupPackets );


	/**
	 * Updates the number of duplicate notifications received in the per
	 * neighbour statistics tab.
	 * 
	 * @param dupNotifications
	 *            the current amount of received duplicate notifications
	 * @return true if this neighbour exists and his values have been updated;
	 *         false otherwise
	 */
	public abstract boolean updateDuplicateNotificationsReceived( Long dupNotifications );


	/**
	 * Updates the RTT of the specified neighbour in the per neighbour
	 * statistics tab.
	 * 
	 * @param neighbourKey
	 *            the hash key of the neighbour representation which is
	 *            'ip:port'
	 * @param newRTT
	 *            the current RTT of the specified neighbour
	 * @return true if this neighbour exists and his values have been updated;
	 *         false otherwise
	 */
	public abstract boolean updateNeighbourRTT( String neighbourKey, Integer newRTT );


	/**
	 * Updates the ranking of the specified neighbour in the per neighbour
	 * statistics tab.
	 * 
	 * @param neighbourKey
	 *            the hash key of the neighbour representation which is
	 *            'ip:port'
	 * @param newRanking
	 *            the current ranking of the specified neighbour
	 * @return true if this neighbour exists and his values have been updated;
	 *         false otherwise
	 */
	public abstract boolean updateNeighbourRanking( String neighbourKey, Integer newRanking );


	/**
	 * Creates a time stamp for the current time of the form HH:MM:SS.
	 * 
	 * @return a String representing this time stamp
	 */
	protected String getTimeStamp(){

		Calendar cal = Calendar.getInstance();
		String hh = String.valueOf( cal.get( Calendar.HOUR_OF_DAY ) );
		if( hh.length() < 2 ){
			hh = "0" + hh;
		}
		String mm = String.valueOf( cal.get( Calendar.MINUTE ) );
		if( mm.length() < 2 ){
			mm = "0" + mm;
		}
		String ss = String.valueOf( cal.get( Calendar.SECOND ) );
		if( ss.length() < 2 ){
			ss = "0" + ss;
		}
		return hh + ":" + mm + ":" + ss;
	}


	/**
	 * Shows the about dialog which contains the logo, the current version and
	 * the URL of the homepage of Swistry.
	 */
	protected void showAboutDialog(){

		String aboutText = "Swistry v." + ProtocolConstants.VERSION + "\n\nSwistry is being developed by the\nDistributed Computing Group at ETH Zurich\n\nFor more information visit\nhttp://www.dcg.ethz.ch";
		JOptionPane.showMessageDialog( frame, aboutText, "About Swistry", JOptionPane.INFORMATION_MESSAGE, createImageIcon( "images/logo_small.png" ) );
	}


	/**
	 * Returns an ImageIcon, or null if the path was invalid.
	 * 
	 * @param path
	 *            the path to the image
	 * @return an ImageIcon representing the specified image
	 */
	protected static ImageIcon createImageIcon( String path ){

		URL imgURL = Thread.currentThread().getContextClassLoader().getResource( path );
		if( imgURL != null ){
			return new ImageIcon( imgURL );
		}else{
			return null;
		}
	}
}
