/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.gui;

import info.clearthought.layout.TableLayout;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;

import ch.ethz.dcg.p2p.peers.NetworkEntryPoint;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * GUI for the <code>NetworkEntryPoint</code> which allows comfortable user
 * interaction.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class NEPGUI extends GUI implements ActionListener {

	// menu
	private JMenuBar menuBar;
	private JMenu menu;
	private JMenuItem menuItem;

	// tabs
	private JTabbedPane tabbedPane;

	// config
	private JPanel pConfig;
	private JTextField tfNNeighbours;
	private JTextField tfPingTimeout;
	private JTextField tfminNeighs;
	private JTextField tfmaxNeighs;
	private JLabel lConfig;
	private JButton bSet;

	// status
	private JTextArea taStatus;
	private JScrollPane sStatus;

	// nodes
	private JTextArea taNodes;
	private JScrollPane sNodes;

	// status bar
	private JLabel statusBar;

	private NetworkEntryPoint nep;


	public static void main( String[] args ){

		// Schedule a job for the event-dispatching thread:
		// creating and showing this application's GUI.
		javax.swing.SwingUtilities.invokeLater( new Runnable() {

			public void run(){

				createAndShowGUI();
			}
		} );
	}


	/**
	 * Creates and shows this gui.
	 */
	private static void createAndShowGUI(){

		JFrame.setDefaultLookAndFeelDecorated( true );
		new NEPGUI();
	}


	/**
	 * Creates a new <code>NEPGUI</code>.<br>
	 * A <code>NetworkEntryPoint</code> is created and the gui is set up.
	 */
	private NEPGUI(){

		// init NEP
		nep = new NetworkEntryPoint( this );

		// create frame
		frame = new JFrame( "Swistry - Network Entry Point" );
		ImageIcon icon = createImageIcon( "images/swistry_icon.gif");
		if( icon != null ){
			frame.setIconImage( icon.getImage() );
		}
		frame.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
		double size[][] = { { TableLayout.FILL }, { TableLayout.FILL, 15 } };
		frame.setLayout( new TableLayout( size ) );

		// create menu
		createMenus();

		// create tab contents
		createConfigTab();
		createStatusTab();
		createNodesTab();

		// create tabs
		tabbedPane = new JTabbedPane();
		tabbedPane.setBorder( BorderFactory.createEmptyBorder( 10, 0, 0, 0 ) );
		tabbedPane.addTab( "Nodes (0)", sNodes );
		tabbedPane.addTab( "Status", sStatus );
		tabbedPane.addTab( "Config", pConfig );
		tabbedPane.setEnabledAt( 0, false );
		tabbedPane.setSelectedIndex( 2 );

		// set status bar
		statusBar = new JLabel( " idle" );
		statusBar.setFont( new Font( null, Font.PLAIN, 10 ) );

		// add componentes to frame
		frame.setJMenuBar( menuBar );
		frame.getContentPane().add( tabbedPane, "0,0" );
		frame.getContentPane().add( statusBar, "0,1" );

		// show gui
		frame.pack();
		frame.setSize( new Dimension( 350, 350 ) );
		frame.setVisible( true );
	}


	/**
	 * Creates the menus.
	 */
	private void createMenus(){

		menuBar = new JMenuBar();
		// Go!
		menu = new JMenu( "Go!" );
		menuBar.add( menu );
		// start
		menuItem = new JMenuItem( "Start" );
		menuItem.setActionCommand( "start" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( false );
		menu.add( menuItem );
		// stop
		menuItem = new JMenuItem( "Stop" );
		menuItem.setActionCommand( "stop" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( false );
		menu.add( menuItem );
		// recover
		menuItem = new JMenuItem( "Recover" );
		menuItem.setActionCommand( "recover" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( nep.recoveryPossible() );
		menu.add( menuItem );

		// HELP
		menu = new JMenu( "Help" );
		menuBar.add( menu );
		// about
		menuItem = new JMenuItem( "About Swistry" );
		menuItem.setActionCommand( "about" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( true );
		menu.add( menuItem );
	}


	/**
	 * Creates a tab containing the configuration options.
	 */
	private void createConfigTab(){

		// layout
		double size[][] = { { 60, 10, TableLayout.FILL }, { 20, 10, 20, 10, 20, 10, 20, 10, 20, 20, 20 } };
		pConfig = new JPanel( new TableLayout( size ) );
		pConfig.setBorder( BorderFactory.createEmptyBorder( 5, 5, 5, 5 ) );

		// init neighbours
		tfNNeighbours = new JTextField( new Integer( nep.getNInitNeighbours() ).toString(), 2 );
		pConfig.add( tfNNeighbours, "0,0,l,c" );
		lConfig = new JLabel( "number of initial neighbours" );
		pConfig.add( lConfig, "2,0" );

		// ping timeout
		tfPingTimeout = new JTextField( new Integer( nep.getPingTimeout() ).toString(), 3 );
		pConfig.add( tfPingTimeout, "0,2,l,c" );
		lConfig = new JLabel( "ping timeout [ms]" );
		pConfig.add( lConfig, "2,2" );

		// minimal amount of neighbours
		tfminNeighs = new JTextField( new Integer( nep.getMinNeighs() ).toString(), 3 );
		pConfig.add( tfminNeighs, "0,4,l,c" );
		lConfig = new JLabel( "minimum number of neighbours required" );
		pConfig.add( lConfig, "2,4" );

		// minimal amount of neighbours
		tfmaxNeighs = new JTextField( new Integer( nep.getMaxNeighs() ).toString(), 3 );
		pConfig.add( tfmaxNeighs, "0,6,l,c" );
		lConfig = new JLabel( "maximum number of neighbours allowed" );
		pConfig.add( lConfig, "2,6" );

		// button
		bSet = new JButton( "Set" );
		bSet.setActionCommand( "set" );
		bSet.addActionListener( this );
		pConfig.add( bSet, "0,10,2,10,l,c" );
	}


	/**
	 * Creates a tab containing status messages.
	 */
	private void createStatusTab(){

		taStatus = new JTextArea( "[" + getTimeStamp() + "] idle: configuration or recovery required\n" );
		taStatus.setEditable( false );
		taStatus.setFont( new Font( null, Font.PLAIN, 10 ) );
		sStatus = new JScrollPane( taStatus );
	}


	/**
	 * Creates a tab containing the registered peers.
	 */
	private void createNodesTab(){

		taNodes = new JTextArea();
		taNodes.setEditable( false );
		taNodes.setFont( new Font( null, Font.PLAIN, 10 ) );
		sNodes = new JScrollPane( taNodes );
	}
	
	
	/**
	 * Enables or disables the config tab items.
	 * 
	 * @param enabled
	 *            true will enable the elements; false will disable them
	 */
	private void setConfigEnabled( boolean enabled ){
		
		tfNNeighbours.setEnabled( enabled );
		tfPingTimeout.setEnabled( enabled );
		tfminNeighs.setEnabled( enabled );
		tfmaxNeighs.setEnabled( enabled );
		bSet.setEnabled( enabled );
	}


	@Override
	public boolean addNeighbour( String neighbourKey ){
	
		return true;
	}


	@Override
	public boolean removeNeighbour( String neighbourKey ){
	
		return true;
	}


	@Override
	public void addStatusLine( String line ){

		if( taStatus != null ){
			taStatus.setText( "[" + getTimeStamp() + "] " + line + "\n" + taStatus.getText() );
		}
		updateStatus();
		tabbedPane.setSelectedIndex( 1 );
	}


	@Override
	public void errorStatusBar( String s ){

		statusBar.setForeground( Color.RED );
		statusBar.setText( " " + s );
	}


	@Override
	public void errorDialog( String message, String title ){

		JOptionPane.showMessageDialog( frame, message, title, JOptionPane.ERROR_MESSAGE );
	}


	/**
	 * Part of the <code>ActionListener</code> interface.<br>
	 * It is basically a dispatcher that reacts upon gui interaction.
	 */
	public void actionPerformed( ActionEvent e ){

		// SET
		if( "set".equals( e.getActionCommand() ) ){
			// warn the user about what he is doing
			if( nep.recoveryPossible() ){
				int n = JOptionPane.showConfirmDialog( frame, "All previously known peers will be lost.\nDo you want to continue?", "Warning: Overwriting persistent state", JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE );
				if( n == JOptionPane.NO_OPTION ){
					return;
				}
			}
			// get values
			int nNeighs, ping, minNeighs, maxNeighs;
			try{
				nNeighs = Integer.parseInt( tfNNeighbours.getText() );
				ping = Integer.parseInt( tfPingTimeout.getText() );
				minNeighs = Integer.parseInt( tfminNeighs.getText() );
				maxNeighs = Integer.parseInt( tfmaxNeighs.getText() );
				if( minNeighs > maxNeighs ){
					throw new NumberFormatException();
				}
			}catch( NumberFormatException e1 ){
				errorStatusBar( "invalid configuration" );
				return;
			}
			nep.setNInitNeighbours( nNeighs );
			nep.setPingTimeout( ping );
			nep.setMinNeighs( minNeighs );
			nep.setMaxNeighs( maxNeighs );

			// enable start
			menuBar.getMenu( 0 ).getMenuComponent( 0 ).setEnabled( true );
			// disable recover
			menuBar.getMenu( 0 ).getMenuComponent( 2 ).setEnabled( false );
			// update status
			nep.setStatus( NetworkEntryPoint.configured );
			addStatusLine( "configured" );
		}

		// START
		if( "start".equals( e.getActionCommand() ) ){
			// start nep
			new Thread( nep, "NetworkEntryPoint" ).start();
			// disable config
			setConfigEnabled( false );
			// disable start
			menuBar.getMenu( 0 ).getMenuComponent( 0 ).setEnabled( false );
			// enable stop
			menuBar.getMenu( 0 ).getMenuComponent( 1 ).setEnabled( true );
			// diable recover
			menuBar.getMenu( 0 ).getMenuComponent( 2 ).setEnabled( false );
			// update and enable nodes tab
			tabbedPane.setEnabledAt( 0, true );
			updateNodes( "", 0 );
			// update status
			nep.setStatus( NetworkEntryPoint.connected );
			addStatusLine( "connected; accepting incoming connections" );
			addStatusLine( "Version " + ProtocolConstants.VERSION );
			statusBar.setForeground( Color.BLUE );
		}

		// STOP
		if( "stop".equals( e.getActionCommand() ) ){
			// stop nep
			nep.stop();
			// enable config
			setConfigEnabled( true );
			// enable start
			menuBar.getMenu( 0 ).getMenuComponent( 0 ).setEnabled( true );
			// disable stop
			menuBar.getMenu( 0 ).getMenuComponent( 1 ).setEnabled( false );
			// enable recover
			menuBar.getMenu( 0 ).getMenuComponent( 2 ).setEnabled( true );
			// update status tab
			nep.setStatus( NetworkEntryPoint.configured );
			addStatusLine( "disconnected" );
			tabbedPane.setSelectedIndex( 1 );
			// update and disable nodes tab
			updateNodes( "", 0 );
			tabbedPane.setEnabledAt( 0, false );
		}

		// RECOVER
		if( "recover".equals( e.getActionCommand() ) ){
			if( nep.recover() ){
				// start nep
				new Thread( nep, "NetworkEntryPoint" ).start();
				// disable config
				setConfigEnabled( false );
				// disable start menu item
				menuBar.getMenu( 0 ).getMenuComponent( 0 ).setEnabled( false );
				// enable stop menu item
				menuBar.getMenu( 0 ).getMenuComponent( 1 ).setEnabled( true );
				// diable recover menu item
				menuBar.getMenu( 0 ).getMenuComponent( 2 ).setEnabled( false );
				// update and enable nodes tab
				tabbedPane.setEnabledAt( 0, true );
				// enable status tab
				tabbedPane.setEnabledAt( 1, true );
				// update status bar
				nep.setStatus( NetworkEntryPoint.connected );
				// update status tab
				addStatusLine( "last state successfully recovered" );
				addStatusLine( "connected; accepting incoming connections" );
				statusBar.setForeground( Color.BLUE );
			}else{
				addStatusLine( "recovering last state failed; reconfiguration required" );
			}
		}

		// ABOUT
		if( "about".equals( e.getActionCommand() ) ){
			showAboutDialog();
		}
	}


	@Override
	public void updateTotals(){
		
		return;
	}


	@Override
	public void updateStatus(){
	
		statusBar.setForeground( Color.BLACK );
		statusBar.setText( nep.getStatus() );
	}
	
	
	/**
	 * Updates the nodes tab.
	 * 
	 * @param s
	 *            the text to set in the tab
	 * @param n
	 *            the number of nodes currently registered
	 */
	public void updateNodes( String s, int n ){
		
		if( n > 0 ){
			taNodes.setText( s );
		}else{
			taNodes.setText( "No nodes available." );
		}
		tabbedPane.setTitleAt( 0, "Nodes (" + n + ") " );
		tabbedPane.setSelectedIndex( 0 );
	}


	@Override
	public boolean updatePacketsReceived( String neighbourKey, Long received ){

		return true;
	}


	@Override
	public boolean updatePacketsSent( String neighbourKey, Long sent ){

		return true;
	}


	@Override
	public boolean updateRequestsReceived( String neighbourKey, Long requests ){

		return true;
	}


	@Override
	public boolean updateRequestsSent( String neighbourKey, Long requests ){

		return true;
	}


	@Override
	public boolean updateNotificationsReceived( String neighbourKey, Long notifications ){

		return true;
	}


	@Override
	public boolean updateNotificationsSent( String neighbourKey, Long notifications ){

		return true;
	}


	@Override
	public boolean updateDuplicatePacketsReceived( Long dupPackets ){

		return true;
	}


	@Override
	public boolean updateDuplicateNotificationsReceived( Long dupNotifications ){

		return true;
	}


	@Override
	public boolean updateNeighbourRTT( String neighbourKey, Integer newDelay ){

		return true;
	}


	@Override
	public boolean updateNeighbourRanking( String neighbourKey, Integer newRanking ){

		return true;
	}

}
