/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.gui;

import info.clearthought.layout.TableLayout;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.HeadlessException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Hashtable;
import java.util.Iterator;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;

import aw.gui.chart.AbstractDataCollector;
import aw.gui.chart.Chart2D;
import aw.gui.chart.ITrace2D;
import aw.gui.chart.Trace2DLtd;
import ch.ethz.dcg.p2p.adapters.output.FileWriterAdapter;
import ch.ethz.dcg.p2p.exceptions.DispatcherNotRunningException;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.exceptions.UDPSocketManagerException;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.peers.Peer;
import ch.ethz.dcg.p2p.tasks.ByeTask;

/**
 * GUI for the <code>Peer</code> which allows comfortable user interaction.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class PeerGUI extends GUI implements ActionListener {

	// menu
	private JMenuBar menuBar;
	private JMenu menu;
	private JMenuItem menuItem;
	private static final int EXIT = 0;
	private static final int JOINEXIT = 1;
	private static final int LEAVEEXIT = 2;

	// tabs
	private JTabbedPane tabbedPane;

	// neighbours tab
	private JPanel pNeighbours;
	private TableLayout tlNeighbours;
	private JLabel lNeigh;
	private JScrollPane sNeighbours;
	private Hashtable<String, JLabel[]> neighbours = new Hashtable<String, JLabel[]>();
	private static final int ID = 0;
	private static final int RANKING = 1;
	private static final int PACKETSRECEIVED = 2;
	private static final int REQUESTSSENT = 3;
	private static final int NOTIFICATIONSRECEIVED = 4;
	private static final int RTT = 5;
	private static final int NNEIGHCOLS = 6;

	// status tab
	private JTextArea taStatus;
	private JScrollPane sStatus;

	// statistics tab
	private JPanel pStatistics;
	private AbstractDataCollector downCollector;
	private AbstractDataCollector upCollector;
	private JLabel lPackRcvSnd;
	private JLabel lNotiRcvSnd;
	private JLabel lReqRcvSnd;
	private JLabel ldupPack;
	private JLabel ldupNoti;
	private JLabel lpacketId;
	private JLabel lKByteTransfer;

	// config tab
	private JTabbedPane tpConfig;
	private JPanel pBasicConfig;
	private JPanel pAdvConfig;
	private JPanel pRecording;
	private JLabel lConfig;
	private JButton bSet;
	private JButton bSave;
	private JButton bStart;
	private JButton bStop;
	private JTextField tfNepAddr;
	private JTextField tfDownSpeed;
	private JTextField tfUpSpeed;
	private JTextField tfIp;
	private JTextField tfPort;
	private JTextField tfMediaPort;
	private JTextField tfStartThreshold;
	private JTextField tfPath;
	private JFileChooser fc;
	private JLabel lImage;

	// file
	private File recordFile;
	private FileOutputStream fileOut;
	private FileWriterAdapter fileAdapter;

	// the number of seconds that are stored in the buffer before streaming to
	// the media player starts
	private int dePacketizerStartThreshold = 5; // sec

	private JLabel statusBar;

	private Peer peer;

	private final String configFileName = "Swistry.conf";


	public static void main( String[] args ){

		// Schedule a job for the event-dispatching thread:
		// creating and showing this application's GUI.
		javax.swing.SwingUtilities.invokeLater( new Runnable() {

			public void run(){

				createAndShowGUI();
			}
		} );
	}


	/**
	 * Creates and shows this gui.
	 */
	private static void createAndShowGUI(){

		JFrame.setDefaultLookAndFeelDecorated( true );
		new PeerGUI();
	}


	/**
	 * Creates a new <code>PeerGUI</code>.<br>
	 * A <code>Peer</code> is created and the gui is set up.
	 */
	private PeerGUI(){

		// init Peer
		peer = new Peer( this );

		// register GUI with the Statistic class
		Node.statistic.setGui( this );

		// init frame
		frame = new JFrame( "Swistry - Peer" );
		ImageIcon icon = createImageIcon( "images/swistry_icon.gif");
		if( icon != null ){
			frame.setIconImage( icon.getImage() );
		}
		frame.addWindowListener( new WindowAdapter() {

			// handle closing of the window
			public void windowClosing( WindowEvent e ){

				// hide the gui
				frame.dispose();
				try{
					if( peer.getStatus() == Node.connected ){
						new Thread( new ByeTask( peer ), "ByeTask" ).start();
						// wait for BYECONFIRM messages
						Thread.sleep( 3000 );
					}
				}catch( InterruptedException ie ){
				}finally{
					// terminate program
					System.exit( 0 );
				}
			}
		} );

		// define gui layout using a TableLayout
		double size[][] = { { TableLayout.FILL }, { TableLayout.FILL, 15 } };
		frame.setLayout( new TableLayout( size ) );

		// create the menus
		createMenus();

		// create the tab contents
		createNeighboursTab();
		createStatisticsTab();
		createStatusTab();
		createConfigTab();
		createRecordingTab();

		// create the tabs
		tabbedPane = new JTabbedPane();
		tabbedPane.setBorder( BorderFactory.createEmptyBorder( 10, 0, 0, 0 ) );
		tabbedPane.addTab( "Neighbours (0)", sNeighbours );
		tabbedPane.addTab( "Statistic", pStatistics );
		tabbedPane.addTab( "Status", sStatus );
		tabbedPane.addTab( "Config", tpConfig );
		tabbedPane.addTab( "Recording", pRecording );
		tabbedPane.setEnabledAt( 0, false );
		tabbedPane.setEnabledAt( 1, false );
		tabbedPane.setEnabledAt( 2, true );
		tabbedPane.setEnabledAt( 3, true );
		tabbedPane.setEnabledAt( 4, false );
		tabbedPane.setSelectedIndex( 3 );

		// add status bar
		statusBar = new JLabel( " idle" );
		statusBar.setFont( new Font( null, Font.PLAIN, 10 ) );

		// add the above componentes to the frame
		frame.setJMenuBar( menuBar );
		frame.getContentPane().add( tabbedPane, "0,0" );
		frame.getContentPane().add( statusBar, "0,1" );

		// show gui
		frame.pack();
		frame.setSize( new Dimension( 600, 400 ) );
		frame.setVisible( true );
	}


	/**
	 * Creates the menus of the gui.
	 */
	private void createMenus(){

		menuBar = new JMenuBar();

		// Go!
		menu = new JMenu( "Go!" );
		menuBar.add( menu );
		// join
		menuItem = new JMenuItem( "Join Network" );
		menuItem.setActionCommand( "join" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( false );
		menu.add( menuItem );
		// leave
		menuItem = new JMenuItem( "Leave Network" );
		menuItem.setActionCommand( "leave" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( false );
		menu.add( menuItem );
		// exit
		menuItem = new JMenuItem( "Exit" );
		menuItem.setActionCommand( "exit" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( true );
		menu.add( menuItem );

		// HELP
		menu = new JMenu( "Help" );
		menuBar.add( menu );
		// about
		menuItem = new JMenuItem( "About Swistry" );
		menuItem.setActionCommand( "about" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( true );
		menu.add( menuItem );
	}


	/**
	 * Creates the tab containing the per neighbour information.
	 */
	private void createNeighboursTab(){

		// set layout
		double size[][] = { { TableLayout.FILL, 5, 50, 5, 50, 5, 50, 5, 70, 5, 50 }, { 10, 10, 2, 1 } };
		tlNeighbours = new TableLayout( size );
		pNeighbours = new JPanel( tlNeighbours );
		pNeighbours.setBorder( BorderFactory.createEmptyBorder( 5, 0, 0, 0 ) );

		// header
		lNeigh = new JLabel( "Peer" );
		lNeigh.setFont( new Font( null, Font.PLAIN, 12 ) );
		pNeighbours.add( lNeigh, "0,0,0,1,c,c" );
		pNeighbours.add( createHeaderLabel( "Ranking" ), "2,0,2,1,c,c" );
		pNeighbours.add( createHeaderLabel( "Packets" ), "4,0,c,c" );
		pNeighbours.add( createHeaderLabel( "received" ), "4,1,c,c" );
		pNeighbours.add( createHeaderLabel( "Packets" ), "6,0,c,c" );
		pNeighbours.add( createHeaderLabel( "requested" ), "6,1,c,c" );
		pNeighbours.add( createHeaderLabel( "Notifications" ), "8,0,c,c" );
		pNeighbours.add( createHeaderLabel( "received" ), "8,1,c,c" );
		pNeighbours.add( createHeaderLabel( "RTT [ms]" ), "10,0,10,1,c,c" );

		// black horizontal line
		lNeigh = new JLabel();
		lNeigh.setOpaque( true );
		lNeigh.setBackground( Color.BLACK );
		pNeighbours.add( lNeigh, "0,3,10,3" );
		sNeighbours = new JScrollPane( pNeighbours );
	}


	/**
	 * Creates a formated header label.
	 * 
	 * @param text
	 *            the text of the header label
	 * @return a JLabel containing the specified text with the header formatting
	 */
	private JLabel createHeaderLabel( String text ){

		JLabel label = new JLabel( text );
		label.setFont( new Font( null, Font.PLAIN, 10 ) );
		return label;
	}


	/**
	 * Rebuilds the table with the per neighbour information.<br>
	 * After this operation only neighbours which are still present will be left
	 * in the table.
	 */
	private void rebuildNeighbourTable(){

		JLabel labels[];
		Iterator<JLabel[]> it;

		// delete all neighbour components
		it = neighbours.values().iterator();
		while( it.hasNext() ){
			labels = it.next();
			for( JLabel curL : labels ){
				tlNeighbours.removeLayoutComponent( curL );
			}
		}

		// delete all neighbour rows
		for( int i = tlNeighbours.getNumRow() - 1; i >= 4; i-- ){
			tlNeighbours.deleteRow( i );
		}

		// insert all neighbour components that are still available
		if( neighbours.size() > 0 ){
			it = neighbours.values().iterator();
			int rows = tlNeighbours.getNumRow();
			while( it.hasNext() ){
				// create new row
				tlNeighbours.insertRow( rows, 5 );
				tlNeighbours.insertRow( rows + 1, 10 );
				// add labels to row
				labels = it.next();
				for( int i = 0; i < NNEIGHCOLS; i++ ){
					pNeighbours.add( labels[i], 2 * i + "," + ( rows + 1 ) + ",c,c" );
				}
				rows += 2;
			}
		}

		// update tab title
		tabbedPane.setTitleAt( 0, "Neighbours (" + neighbours.size() + ")" );
		tabbedPane.setSelectedIndex( 0 );
	}


	/**
	 * Creates tab which contains status informations.
	 */
	private void createStatusTab(){

		taStatus = new JTextArea();
		taStatus.setEditable( false );
		taStatus.setFont( new Font( null, Font.PLAIN, 10 ) );
		addStatusLine( "idle: config required" );
		sStatus = new JScrollPane( taStatus );
	}


	/**
	 * Create the tab containing the overall statistics.
	 */
	private void createStatisticsTab(){

		// layout
		double size[][] = { { 200, TableLayout.FILL }, { 100, 20, 10, 10, 10, 10, 10, 20, 10, 10, 10, 20, 10, 20, 10 } };
		pStatistics = new JPanel( new TableLayout( size ) );
		pStatistics.setBorder( BorderFactory.createEmptyBorder( 5, 5, 5, 5 ) );

		// chart
		Chart2D loadChart = new Chart2D();
		ITrace2D downTrace = new Trace2DLtd( 100, "download [kbps]" );
		ITrace2D upTrace = new Trace2DLtd( 100, "upload [kbps]" );
		downTrace.setColor( Color.GREEN );
		upTrace.setColor( Color.RED );
		loadChart.addTrace( downTrace );
		loadChart.addTrace( upTrace );
		pStatistics.add( loadChart, "0,0,1,0" );

		// packets
		pStatistics.add( createStatLabel( "packets received / sent: " ), "0,2,l,c" );
		lPackRcvSnd = new JLabel( "0 / 0" );
		lPackRcvSnd.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( lPackRcvSnd, "1,2,l,c" );

		// requests
		pStatistics.add( createStatLabel( "requests received / sent: " ), "0,4,l,c" );
		lReqRcvSnd = new JLabel( "0 / 0" );
		lReqRcvSnd.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( lReqRcvSnd, "1,4,l,c" );

		// notifications
		pStatistics.add( createStatLabel( "notifications received / sent: " ), "0,6,l,c" );
		lNotiRcvSnd = new JLabel( "0 / 0" );
		lNotiRcvSnd.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( lNotiRcvSnd, "1,6,l,c" );

		// duplicate packets
		pStatistics.add( createStatLabel( "duplicate packets received: " ), "0,8,l,c" );
		ldupPack = new JLabel( new Long( Node.statistic.getDuplicatePacketsReceived() ).toString() );
		ldupPack.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( ldupPack, "1,8,l,c" );

		// duplicate notifications
		pStatistics.add( createStatLabel( "duplicate notifications received: " ), "0,10,l,c" );
		ldupNoti = new JLabel( new Long( Node.statistic.getDuplicateNotificationsReceived() ).toString() );
		ldupNoti.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( ldupNoti, "1,10,l,c" );

		// current packet id
		pStatistics.add( createStatLabel( "packet id forwarding to media player: " ), "0,12,l,c" );
		lpacketId = new JLabel( new Long( peer.getCurPacketId() ).toString() );
		lpacketId.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( lpacketId, "1,12,l,c" );

		// total kbytes
		pStatistics.add( createStatLabel( "kilobytes received / sent: " ), "0,14,l,c" );
		lKByteTransfer = new JLabel( "0kB / 0kB" );
		lKByteTransfer.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( lKByteTransfer, "1,14,l,c" );

		// init datacollectors
		downCollector = new DownloadDataCollector( downTrace, 3000 );
		upCollector = new UploadDataCollector( upTrace, 3000 );
	}


	/**
	 * Create a formatted label for the statistic.
	 * 
	 * @param text
	 *            the text of the label
	 * @return JLabel containing the specified text formatted as required
	 */
	private JLabel createStatLabel( String text ){

		JLabel label = new JLabel( text );
		label.setFont( new Font( null, Font.PLAIN, 10 ) );
		return label;
	}


	/**
	 * Creates the tab containing the basic- and advanced configurations.
	 */
	private void createConfigTab(){

		InetAddress nepAddress = null;
		int inSpeed = -1;
		int outSpeed = -1;
		InetAddress ip = null;
		int port = -1;
		int mediaPort = -1;
		int dePacketizerThreshold = -1;

		// read last config from file if available
		try{
			File configFile = new File( configFileName );
			if( !configFile.exists() ){
				peer.write2gui( "no config file found: setting default values" );
			}else{
				FileInputStream fileIn;
				fileIn = new FileInputStream( configFile );
				ObjectInputStream objectIn = new ObjectInputStream( fileIn );
				byte[] tmpAddr = new byte[4];
				// read basic config settings from file
				objectIn.read( tmpAddr, 0, 4 );
				nepAddress = InetAddress.getByAddress( tmpAddr );
				inSpeed = objectIn.readInt();
				outSpeed = objectIn.readInt();
				// read advanced config settings from file
				objectIn.read( tmpAddr, 0, 4 );
				ip = InetAddress.getByAddress( tmpAddr );
				port = objectIn.readInt();
				mediaPort = objectIn.readInt();
				dePacketizerThreshold = objectIn.readInt();
				// close file stream
				objectIn.close();
				fileIn.close();
			}
		}catch( IOException e ){
			peer.write2gui( "cannot read last config: setting default values" );
		}

		// create sub-tabs
		createBasicConfigTab( nepAddress, inSpeed, outSpeed );
		createAdvancedConfigTab( ip, port, mediaPort, dePacketizerThreshold );

		// create config tab
		tpConfig = new JTabbedPane();
		tpConfig.setBorder( BorderFactory.createEmptyBorder( 10, 0, 0, 0 ) );
		tpConfig.add( "Basic", pBasicConfig );
		tpConfig.add( "Advanced", pAdvConfig );
	}


	/**
	 * Creates the basic configuration tab.
	 * 
	 * @param nepAddress
	 *            IP address of the last <code>NetworkEntryPoint</code>
	 * @param inSpeed
	 *            last inbound bandwidth of the peer
	 * @param outSpeed
	 *            last outbound bandwidth of the peer
	 */
	private void createBasicConfigTab( InetAddress nepAddress, int inSpeed, int outSpeed ){

		// layout
		double size[][] = { { 40, 5, 40, 10, 80, 10, TableLayout.FILL }, { 20, 20, 20, 5, 20, TableLayout.FILL } };
		pBasicConfig = new JPanel( new TableLayout( size ) );
		pBasicConfig.setBorder( BorderFactory.createEmptyBorder( 10, 5, 5, 5 ) );

		// NEP address
		tfNepAddr = new JTextField( "", 10 );
		if( nepAddress != null ){
			tfNepAddr.setText( nepAddress.getHostAddress() );
		}
		pBasicConfig.add( tfNepAddr, "0,0,4,0" );
		lConfig = new JLabel( "IP address or hostname of the Network Entry Point" );
		pBasicConfig.add( lConfig, "6,0,l,c" );

		// inbound speed
		tfDownSpeed = new JTextField( "", 3 );
		if( inSpeed != -1 ){
			tfDownSpeed.setText( Integer.toString( inSpeed ) );
		}
		pBasicConfig.add( tfDownSpeed, "0,2,l,c" );
		lConfig = new JLabel( "Kbit/s" );
		lConfig.setFont( new Font( null, Font.PLAIN, 12 ) );
		pBasicConfig.add( lConfig, "2,2,l,c" );
		lConfig = new JLabel( "download speed" );
		pBasicConfig.add( lConfig, "4,2,6,2,l,c" );

		// outbound speed
		tfUpSpeed = new JTextField( "", 3 );
		if( outSpeed != -1 ){
			tfUpSpeed.setText( Integer.toString( outSpeed ) );
		}
		pBasicConfig.add( tfUpSpeed, "0,4,l,c" );
		lConfig = new JLabel( "Kbit/s" );
		lConfig.setFont( new Font( null, Font.PLAIN, 12 ) );
		pBasicConfig.add( lConfig, "2,4,l,c" );
		lConfig = new JLabel( "upload speed" );
		pBasicConfig.add( lConfig, "4,4,6,4,l,c" );

		// button
		bSet = new JButton( "Set" );
		bSet.setActionCommand( "set" );
		bSet.addActionListener( this );
		pBasicConfig.add( bSet, "0,5,2,5,l,b" );

		// image
		lImage = new JLabel( createImageIcon( "images/logo.png" ) );
		pBasicConfig.add( lImage, "6,5,r,b" );
	}


	/**
	 * Creates the advanced configuration tab.
	 * 
	 * @param ip
	 *            last IP address of the peer
	 * @param port
	 *            last port number of the peer
	 * @param mediaPort
	 *            last media port number of the peer
	 * @param dePacketizerThreshold
	 *            last dePacketizerThreshold of the peer
	 */
	private void createAdvancedConfigTab( InetAddress ip, int port, int mediaPort, int dePacketizerThreshold ){

		// layout
		double size[][] = { { 60, 10, 105, 10, TableLayout.FILL }, { 20, 5, 20, 20, 20, 20, 20 } };
		pAdvConfig = new JPanel( new TableLayout( size ) );
		pAdvConfig.setBorder( BorderFactory.createEmptyBorder( 10, 5, 5, 5 ) );

		// host address
		tfIp = new JTextField( 8 );
		if( ip != null ){
			tfIp.setText( ip.getHostAddress() );
		}else{
			tfIp.setText( whatIsMyIp().getHostAddress() );
		}
		pAdvConfig.add( tfIp, "0,0,2,0" );
		lConfig = new JLabel( "public IP address or name of this host" );
		pAdvConfig.add( lConfig, "4,0,l,c" );

		// host port
		tfPort = new JTextField( 4 );
		if( port != -1 ){
			tfPort.setText( Integer.toString( port ) );
		}else{
			tfPort.setText( Integer.toString( 65003 ) );
		}
		pAdvConfig.add( tfPort, "0,2,l,c" );
		lConfig = new JLabel( "port [49152...65535]" );
		pAdvConfig.add( lConfig, "2,2,4,2,l,c" );

		// local media port
		tfMediaPort = new JTextField( 4 );
		if( mediaPort != -1 ){
			tfMediaPort.setText( Integer.toString( mediaPort ) );
		}else{
			tfMediaPort.setText( Integer.toString( 44444 ) );
		}
		pAdvConfig.add( tfMediaPort, "0,4,l,c" );
		lConfig = new JLabel( "local port for media player" );
		pAdvConfig.add( lConfig, "2,4,4,4,l,c" );

		// depacketizer start threshold
		tfStartThreshold = new JTextField( 2 );
		if( dePacketizerThreshold != -1 ){
			tfStartThreshold.setText( Integer.toString( dePacketizerThreshold ) );
		}else{
			tfStartThreshold.setText( Integer.toString( 8 ) );
		}
		pAdvConfig.add( tfStartThreshold, "0,6,l,c" );
		lConfig = new JLabel( "#seconds that are buffered before stream starts" );
		pAdvConfig.add( lConfig, "2,6,4,6,l,c" );
	}


	/**
	 * Gets this hosts IP address by making a request to
	 * http://www.whatismyip.com and parsing the answer.
	 * 
	 * @return the public IP address of this host as reported by whatismyip.com
	 */
	private InetAddress whatIsMyIp(){

		InetAddress ip = null;
		try{
			URL url = new URL( "http://www.whatismyip.com" );
			BufferedReader reader = new BufferedReader( new InputStreamReader( url.openStream() ) );
			String curLine = reader.readLine();
			while( ( curLine != null ) && !curLine.startsWith( "<TITLE>" ) ){
				curLine = reader.readLine();
			}
			if( curLine != null ){
				int begin = curLine.indexOf( " - " ) + 3;
				int end = curLine.indexOf( "</TITLE>" );
				ip = InetAddress.getByName( curLine.substring( begin, end ) );
				return ip;
			}else{
				return null;
			}
		}catch( MalformedURLException e ){
			// should not happen ;)
		}catch( IOException e ){
			try{
				// try the local host address
				return InetAddress.getLocalHost();
			}catch( UnknownHostException uhe ){
				try{
					// try 'localhost'
					return InetAddress.getByName( "localhost" );
				}catch( UnknownHostException e1 ){
					// should not happen ;)
				}
			}
		}
		return null;
	}


	/**
	 * Creates the tab for recording into a file.
	 */
	private void createRecordingTab(){

		// layout
		double size[][] = { { TableLayout.PREFERRED, 5, TableLayout.PREFERRED, TableLayout.FILL }, { 20, 10, 20, 40, 20 } };
		pRecording = new JPanel( new TableLayout( size ) );
		pRecording.setBorder( BorderFactory.createEmptyBorder( 10, 5, 5, 5 ) );

		// buttons
		bSave = new JButton( "Choose File...", createImageIcon( "images/save16.gif" ) );
		bSave.setActionCommand( "save" );
		bSave.addActionListener( this );
		pRecording.add( bSave, "0,2,l,c" );
		bStart = new JButton( "Start Recording", createImageIcon( "images/start16.gif" ) );
		bStart.setActionCommand( "start" );
		bStart.addActionListener( this );
		bStart.setEnabled( false );
		pRecording.add( bStart, "0,4,l,c" );
		bStop = new JButton( "Stop", createImageIcon( "images/stop16.gif" ) );
		bStop.setActionCommand( "stop" );
		bStop.addActionListener( this );
		pRecording.add( bStop, "2,4,l,c" );
		bStop.setEnabled( false );

		// file path
		tfPath = new JTextField();
		tfPath.setEditable( false );
		pRecording.add( tfPath, "0,0,3,0," );

		// file chooser
		fc = new JFileChooser();
	}


	/**
	 * Enables or disable sthe gui elements in the config tab.
	 * 
	 * @param enabled
	 *            true will enable the elements; false will disable them
	 */
	private void setConfigEnabled( boolean enabled ){

		tfNepAddr.setEnabled( enabled );
		tfDownSpeed.setEnabled( enabled );
		tfUpSpeed.setEnabled( enabled );
		tfIp.setEnabled( enabled );
		tfPort.setEnabled( enabled );
		tfMediaPort.setEnabled( enabled );
		tfStartThreshold.setEnabled( enabled );
		bSet.setEnabled( enabled );
	}


	/**
	 * Enables or disables the status tab.
	 * 
	 * @param enabled
	 *            true will enable the elements; false will disable them
	 */
	private void setStatusEnabled( boolean enabled ){

		tabbedPane.setEnabledAt( 2, enabled );
	}


	/**
	 * Enables or disables the statistic tab.
	 * 
	 * @param enabled
	 *            true will enable the elements; false will disable them
	 */
	private void setStatisticEnabled( boolean enabled ){

		tabbedPane.setEnabledAt( 1, enabled );
	}


	/**
	 * Enables or disables the neighbours tab.
	 * 
	 * @param enabled
	 *            true will enable the elements; false will disable them
	 */
	private void setNeighboursEnabled( boolean enabled ){

		tabbedPane.setEnabledAt( 0, enabled );
	}


	/**
	 * Enables or disables the recording tab.
	 * 
	 * @param enabled
	 *            true will enable the elements; false will disable them
	 */
	private void setRecordingEnabled( boolean enabled ){

		tabbedPane.setEnabledAt( 4, enabled );
		tfPath.setEnabled( enabled );
		if( !enabled ){
			bSave.setEnabled( false );
		}else{
			bSave.setEnabled( true );
		}
		bStart.setEnabled( false );
		bStop.setEnabled( false );
	}


	/**
	 * Enables or disables different menu items using the predefined constants.
	 * 
	 * @param enabled
	 *            EXIT = exit only<br>
	 *            JOINEXIT = join and exit<br>
	 *            LEAVEEXIT = leave and exit<br>
	 */
	private void setMenuItemsEnabled( int enabled ){

		JMenu control = menuBar.getMenu( 0 );
		switch( enabled ){
		case EXIT:
			control.getMenuComponent( 0 ).setEnabled( false );
			control.getMenuComponent( 1 ).setEnabled( false );
			control.getMenuComponent( 2 ).setEnabled( true );
			break;
		case JOINEXIT:
			control.getMenuComponent( 0 ).setEnabled( true );
			control.getMenuComponent( 1 ).setEnabled( false );
			control.getMenuComponent( 2 ).setEnabled( true );
			break;
		case LEAVEEXIT:
			control.getMenuComponent( 0 ).setEnabled( false );
			control.getMenuComponent( 1 ).setEnabled( true );
			control.getMenuComponent( 2 ).setEnabled( true );
			break;
		default:
			break;
		}
	}


	@Override
	public boolean addNeighbour( String id ){

		if( neighbours.containsKey( id ) ){
			return false;
		}

		// create labels
		JLabel[] labels = new JLabel[NNEIGHCOLS];
		labels[ID] = new JLabel( id );
		labels[RANKING] = new JLabel( "0" );
		labels[PACKETSRECEIVED] = new JLabel( "0" );
		labels[REQUESTSSENT] = new JLabel( "0" );
		labels[NOTIFICATIONSRECEIVED] = new JLabel( "0" );
		labels[RTT] = new JLabel( "INF" );

		// add neighbour to hashtable
		neighbours.put( id, labels );

		// create new row
		int rows = tlNeighbours.getNumRow();
		tlNeighbours.insertRow( rows++, 5 );
		tlNeighbours.insertRow( rows, 10 );

		// add labels to new row
		for( int i = 0; i < NNEIGHCOLS; i++ ){
			labels[i].setFont( new Font( null, Font.PLAIN, 10 ) );
			pNeighbours.add( labels[i], 2 * i + "," + rows + ",c,c" );
		}
		labels[0].setForeground( Color.BLUE );

		// update tab title
		tabbedPane.setTitleAt( 0, "Neighbours (" + neighbours.size() + ")" );
		tabbedPane.setSelectedIndex( 0 );

		return true;
	}


	@Override
	public boolean removeNeighbour( String id ){

		if( !neighbours.containsKey( id ) ){
			return false;
		}else{
			JLabel[] labels = neighbours.remove( id );
			for( JLabel curL : labels ){
				tlNeighbours.removeLayoutComponent( curL );
			}
			rebuildNeighbourTable();
			return true;
		}
	}


	@Override
	public void addStatusLine( String line ){

		if( taStatus != null ){
			taStatus.setText( "[" + getTimeStamp() + "] " + line + "\n" + taStatus.getText() );
		}
	}


	@Override
	public void errorStatusBar( String s ){

		statusBar.setForeground( Color.RED );
		statusBar.setText( " " + s );
	}


	@Override
	public void errorDialog( String message, String title ){

		JOptionPane.showMessageDialog( frame, message, title, JOptionPane.ERROR_MESSAGE );
	}


	/**
	 * Part of the <code>ActionListener</code> interface.<br>
	 * It is basically a dispatcher that reacts upon gui interaction.
	 */
	public void actionPerformed( ActionEvent e ){

		// SET
		if( "set".equals( e.getActionCommand() ) ){
			// get values
			InetAddress nepAddress = null;
			int inSpeed = 0;
			int outSpeed = 0;
			InetAddress ip = null;
			int port = 65002;
			int mediaPort = 3001;
			// basic config
			// network entry point
			try{
				if( isValid( tfNepAddr.getText() ) ){
					nepAddress = InetAddress.getByName( tfNepAddr.getText() );
				}else{
					tfNepAddr.setForeground( Color.RED );
					errorStatusBar( "invalid network entry point address" );
					tpConfig.setSelectedIndex( 0 );
					return;
				}
			}catch( UnknownHostException e1 ){
				tfNepAddr.setForeground( Color.RED );
				errorStatusBar( "invalid network entry point address" );
				tpConfig.setSelectedIndex( 0 );
				return;
			}
			if( tfNepAddr.getForeground().equals( Color.RED ) ){
				tfNepAddr.setForeground( Color.BLACK );
			}
			// inSpeed
			try{
				inSpeed = Integer.parseInt( tfDownSpeed.getText() );
				if( inSpeed <= 0 ){
					tfDownSpeed.setForeground( Color.RED );
					errorStatusBar( "invalid download speed" );
					return;
				}
			}catch( NumberFormatException nfe ){
				tfDownSpeed.setText( "" );
				errorStatusBar( "invalid download speed" );
				return;
			}
			if( tfDownSpeed.getForeground().equals( Color.RED ) ){
				tfDownSpeed.setForeground( Color.BLACK );
			}
			// outSpeed
			try{
				outSpeed = Integer.parseInt( tfUpSpeed.getText() );
				if( outSpeed <= 0 ){
					tfUpSpeed.setForeground( Color.RED );
					errorStatusBar( "invalid upload speed" );
					return;
				}
			}catch( NumberFormatException nfe ){
				tfUpSpeed.setText( "" );
				errorStatusBar( "invalid upload speed" );
				return;
			}
			if( tfUpSpeed.getForeground().equals( Color.RED ) ){
				tfUpSpeed.setForeground( Color.BLACK );
			}

			// advanced config
			// IP
			try{
				if( isValid( tfIp.getText() ) ){
					ip = InetAddress.getByName( tfIp.getText() );
				}else{
					tfIp.setForeground( Color.RED );
					errorStatusBar( "invalid IP address" );
					tpConfig.setSelectedIndex( 1 );
					return;
				}
			}catch( UnknownHostException e1 ){
				tfIp.setForeground( Color.RED );
				errorStatusBar( "invalid IP address" );
				tpConfig.setSelectedIndex( 1 );
				return;
			}
			if( tfIp.getForeground().equals( Color.RED ) ){
				tfIp.setForeground( Color.BLACK );
			}
			// port
			try{
				port = Integer.parseInt( tfPort.getText() );
				if( ( port < 0 ) || ( port > 65535 ) ){
					tfPort.setForeground( Color.RED );
					errorStatusBar( "invalid port" );
					tpConfig.setSelectedIndex( 1 );
					return;
				}
			}catch( NumberFormatException nfe ){
				tfPort.setForeground( Color.RED );
				tfPort.setText( "65002" );
				errorStatusBar( "invalid port" );
				tpConfig.setSelectedIndex( 1 );
				return;
			}
			if( tfPort.getForeground().equals( Color.RED ) ){
				tfPort.setForeground( Color.BLACK );
			}
			// media player port
			try{
				mediaPort = Integer.parseInt( tfMediaPort.getText() );
				if( ( mediaPort < 0 ) || ( mediaPort > 65535 ) ){
					tfMediaPort.setForeground( Color.RED );
					errorStatusBar( "invalid media player port" );
					tpConfig.setSelectedIndex( 1 );
					return;
				}
			}catch( NumberFormatException nfe ){
				tfMediaPort.setForeground( Color.RED );
				tfMediaPort.setText( "3001" );
				errorStatusBar( "invalid media player port" );
				tpConfig.setSelectedIndex( 1 );
				return;
			}
			if( tfMediaPort.getForeground().equals( Color.RED ) ){
				tfMediaPort.setForeground( Color.BLACK );
			}
			// start threshold of the depacketizer (buffer time)
			try{
				dePacketizerStartThreshold = Integer.parseInt( tfStartThreshold.getText() );
				if( dePacketizerStartThreshold < 3 ){
					tfStartThreshold.setForeground( Color.RED );
					errorStatusBar( "invalid buffer time: must be at least 3" );
					tpConfig.setSelectedIndex( 1 );
					return;
				}
			}catch( NumberFormatException nfe ){
				tfStartThreshold.setForeground( Color.RED );
				tfStartThreshold.setText( "5" );
				errorStatusBar( "invalid start threshold" );
				tpConfig.setSelectedIndex( 1 );
				return;
			}
			if( tfStartThreshold.getForeground().equals( Color.RED ) ){
				tfStartThreshold.setForeground( Color.BLACK );
			}

			// clean up if we are in an error state
			if( peer.getStatus().equals( Node.error ) ){
				reset();
			}

			// if this point is reached, all the config input is valid
			writeConfigToFile( nepAddress, inSpeed, outSpeed, ip, port, mediaPort, dePacketizerStartThreshold );
			peer.configure( ip, port, inSpeed, outSpeed, nepAddress, dePacketizerStartThreshold );
			peer.setMediaPort( mediaPort );
		}

		// JOIN
		if( "join".equals( e.getActionCommand() ) ){
			// join network
			setMenuItemsEnabled( EXIT );
			if( peer.getStatus().equals( Node.error ) ){
				reset();
			}
			try{
				peer.startDispatchers();
				peer.joinNetwork();
				peer.startRequestTask();
				downCollector.start();
				upCollector.start();
				// peer.unleashZombieKiller();
				tabbedPane.setSelectedIndex( 0 );
			}catch( DispatcherNotRunningException e1 ){
				peer.setStatus( Node.error );
				addStatusLine( "[ERROR] " + e1.getMessage() );
			}catch( NodeNotConfiguredException e1 ){
				peer.setStatus( Node.error );
				addStatusLine( "[ERROR] " + e1.getMessage() );
			}catch( UDPSocketManagerException e1 ){
				peer.setStatus( Node.error );
				addStatusLine( "[ERROR] " + e1.getMessage() );
			}
		}

		// LEAVE
		if( "leave".equals( e.getActionCommand() ) ){
			reset();
		}

		// EXIT
		if( "exit".equals( e.getActionCommand() ) ){
			frame.dispose();
			new Thread( new ByeTask( peer ), "ByeTask" ).start();
			try{
				Thread.sleep( 2000 );
			}catch( InterruptedException ie ){
			}finally{
				System.exit( 0 );
			}
		}

		// FILE RECORDING
		// CHOOSE FILE
		if( "save".equals( e.getActionCommand() ) ){
			int returnVal = fc.showSaveDialog( frame );
			if( returnVal == JFileChooser.APPROVE_OPTION ){
				try{
					recordFile = fc.getSelectedFile();
					// check if the file already exists
					if( recordFile.exists() ){
						int answer = JOptionPane.showConfirmDialog( frame, "File already exists!\nDo you want to overwrite it?", "Warning: Overwriting existing file", JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE );
						if( answer == JOptionPane.NO_OPTION ){
							return;
						}
					}
					tfPath.setText( recordFile.getAbsolutePath() );
					bStart.setEnabled( true );
				}catch( HeadlessException e1 ){
					peer.write2gui( e1.getMessage() );
				}
			}
		}
		// START
		if( "start".equals( e.getActionCommand() ) ){
			if( recordFile != null ){
				try{
					if( recordFile.exists() ){
						int answer = JOptionPane.showConfirmDialog( frame, "File already exists!\nDo you want to overwrite it?", "Warning: Overwriting existing file", JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE );
						if( answer == JOptionPane.NO_OPTION ){
							return;
						}
					}
					fileOut = new FileOutputStream( recordFile );
					fileAdapter = new FileWriterAdapter( fileOut, peer );
					peer.addOutputAdapter( fileAdapter );
					bStop.setEnabled( true );
					bSave.setEnabled( false );
				}catch( FileNotFoundException e1 ){
					peer.write2gui( e1.getMessage() );
					recordFile = null;
				}
			}else{
				peer.write2gui( "no file specified for recording" );
			}
			bStart.setEnabled( false );
		}
		// STOP
		if( "stop".equals( e.getActionCommand() ) ){
			try{
				peer.removeOutputAdapter( fileAdapter );
				fileOut.close();
				fileOut = null;
			}catch( IOException e1 ){
				peer.write2gui( e1.getMessage() );
			}
			bStop.setEnabled( false );
			bStart.setEnabled( true );
			bSave.setEnabled( true );
		}

		// ABOUT
		if( "about".equals( e.getActionCommand() ) ){
			showAboutDialog();
		}
	}


	/**
	 * Tests whether the given string represents a valid IP address.
	 * 
	 * @param ip
	 * @return true if the IP address is valid; false otherwise
	 */
	private static boolean isValid( String ip ){

		try{
			if( ip.equals( "" ) || ip == null ){
				return false;
			}
			new URL( "http://" + ip );
			return true;
		}catch( MalformedURLException e ){
			return false;
		}
	}


	/**
	 * Cleans up the peer after leaving the network or when recovering from an
	 * error.
	 * 
	 */
	private void reset(){

		// leave network
		peer.leaveNetwork();
		// reset neighbours heap
		neighbours = new Hashtable<String, JLabel[]>();
		// reset neighbours
		tabbedPane.remove( 0 );
		createNeighboursTab();
		tabbedPane.add( pNeighbours, 0 );
		tabbedPane.setTitleAt( 0, "Neighbours (0)" );
		tabbedPane.setEnabledAt( 0, false );
		// reset statistics
		downCollector.stop();
		upCollector.stop();
		tabbedPane.remove( 1 );
		createStatisticsTab();
		tabbedPane.add( pStatistics, 1 );
		tabbedPane.setTitleAt( 1, "Statistic" );
		tabbedPane.setEnabledAt( 1, false );
		// reset file
		recordFile = null;
		fileOut = null;
		fileAdapter = null;
		tfPath.setText( "" );
		tabbedPane.setEnabledAt( 4, false );
		// select config tab
		tabbedPane.setSelectedIndex( 3 );
	}


	/**
	 * Write the current peer settings into a file.
	 * 
	 * @param nepAddress
	 *            IP address of the </code>NetworkEntryPoint</code>
	 * @param inSpeed
	 *            inbound bandwidth
	 * @param outSpeed
	 *            outbound bandwidth
	 * @param ip
	 *            IP address of this peer
	 * @param port
	 *            port number of this peer
	 * @param mediaPort
	 *            port to connect the media player to
	 * @param dePacketizerThreshold
	 *            threshold of the <code>DePacketizer</code> in seconds
	 * @return true if the config was written down successfully; false otherwise
	 */
	private boolean writeConfigToFile( InetAddress nepAddress, int inSpeed, int outSpeed, InetAddress ip, int port, int mediaPort, int dePacketizerThreshold ){

		try{
			// create a new file (delete any old ones)
			File configFile = new File( configFileName );
			if( configFile.exists() ){
				configFile.delete();
			}
			configFile.createNewFile();
			FileOutputStream fileOut = new FileOutputStream( configFile );
			ObjectOutputStream objectOut = new ObjectOutputStream( fileOut );
			// write current config to file
			objectOut.write( nepAddress.getAddress() );
			objectOut.writeInt( inSpeed );
			objectOut.writeInt( outSpeed );
			objectOut.write( ip.getAddress() );
			objectOut.writeInt( port );
			objectOut.writeInt( mediaPort );
			objectOut.writeInt( dePacketizerThreshold );
			// close the file stream
			objectOut.close();
			fileOut.close();
			return true;
		}catch( IOException e ){
			peer.write2gui( "[ERROR] cannot persistently store the current configuration settings into a file" );
			return false;
		}
	}


	@Override
	public void updateTotals(){

		// update global statistic
		lPackRcvSnd.setText( Node.statistic.getTotalPacketsReceived() + " / " + Node.statistic.getTotalPacketsSent() );
		lReqRcvSnd.setText( Node.statistic.getTotalRequestsReceived() + " / " + Node.statistic.getTotalRequestsSent() );
		lNotiRcvSnd.setText( Node.statistic.getTotalNotificationsReceived() + " / " + Node.statistic.getTotalNotificationsSent() );
		ldupPack.setText( String.valueOf( Node.statistic.getDuplicatePacketsReceived() ) );
		ldupNoti.setText( String.valueOf( Node.statistic.getDuplicateNotificationsReceived() ) );
		lpacketId.setText( String.valueOf( peer.getCurPacketId() ) );
		lKByteTransfer.setText( Node.statistic.getTotalKBytesReceived() + "kB / " + Node.statistic.getTotalKBytesSent() + "kB" );
	}


	@Override
	public void updateStatus(){

		String status = peer.getStatus();
		statusBar.setForeground( Color.BLACK );
		if( status.equals( Node.idle ) ){
			setNeighboursEnabled( false );
			setStatisticEnabled( false );
			setStatusEnabled( true );
			setConfigEnabled( true );
			setRecordingEnabled( false );
			setMenuItemsEnabled( EXIT );
		}else if( status.equals( Node.configured ) ){
			setNeighboursEnabled( false );
			setStatisticEnabled( false );
			setStatusEnabled( true );
			setConfigEnabled( true );
			setRecordingEnabled( false );
			setMenuItemsEnabled( JOINEXIT );
		}else if( status.equals( Node.connected ) ){
			setNeighboursEnabled( true );
			setStatisticEnabled( true );
			setStatusEnabled( true );
			setConfigEnabled( false );
			setRecordingEnabled( true );
			setMenuItemsEnabled( LEAVEEXIT );
		}else if( status.equals( Node.error ) ){
			statusBar.setForeground( Color.RED );
			setNeighboursEnabled( false );
			setStatisticEnabled( false );
			setStatusEnabled( true );
			setConfigEnabled( true );
			setRecordingEnabled( false );
			setMenuItemsEnabled( JOINEXIT );
			tabbedPane.setSelectedIndex( 2 );
		}
		statusBar.setText( " " + status );
	}


	@Override
	public boolean updatePacketsReceived( String neighbourKey, Long received ){

		if( neighbours.containsKey( neighbourKey ) ){
			neighbours.get( neighbourKey )[PACKETSRECEIVED].setText( received.toString() );
			return true;
		}else{
			return false;
		}
	}


	@Override
	public boolean updatePacketsSent( String neighbourKey, Long sent ){

		return true;
	}


	@Override
	public boolean updateRequestsReceived( String neighbourKey, Long received ){

		return true;
	}


	@Override
	public boolean updateRequestsSent( String neighbourKey, Long sent ){

		if( neighbours.containsKey( neighbourKey ) ){
			neighbours.get( neighbourKey )[REQUESTSSENT].setText( sent.toString() );
			return true;
		}else{
			return false;
		}
	}


	@Override
	public boolean updateNotificationsReceived( String neighbourKey, Long received ){

		if( neighbours.containsKey( neighbourKey ) ){
			neighbours.get( neighbourKey )[NOTIFICATIONSRECEIVED].setText( received.toString() );
			return true;
		}else{
			return false;
		}
	}


	@Override
	public boolean updateNotificationsSent( String neighbourKey, Long sent ){

		return true;
	}


	@Override
	public boolean updateDuplicatePacketsReceived( Long dupPackets ){

		ldupPack.setText( dupPackets.toString() );
		return true;
	}


	@Override
	public boolean updateDuplicateNotificationsReceived( Long dupNotifications ){

		ldupNoti.setText( dupNotifications.toString() );
		return true;
	}


	@Override
	public boolean updateNeighbourRTT( String neighbourKey, Integer newDelay ){

		if( neighbours.containsKey( neighbourKey ) ){
			neighbours.get( neighbourKey )[RTT].setText( newDelay.toString() );
			return true;
		}else{
			return false;
		}
	}


	@Override
	public boolean updateNeighbourRanking( String neighbourKey, Integer newRanking ){

		if( neighbours.containsKey( neighbourKey ) ){
			neighbours.get( neighbourKey )[RANKING].setText( newRanking.toString() );
			return true;
		}else{
			return false;
		}
	}

}
