/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.gui;

import info.clearthought.layout.TableLayout;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Hashtable;
import java.util.Iterator;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;

import aw.gui.chart.AbstractDataCollector;
import aw.gui.chart.Chart2D;
import aw.gui.chart.ITrace2D;
import aw.gui.chart.Trace2DLtd;
import ch.ethz.dcg.p2p.adapters.input.InputAdapter;
import ch.ethz.dcg.p2p.adapters.input.ShoutCastAdapter;
import ch.ethz.dcg.p2p.adapters.input.VlcHttpAdapter;
import ch.ethz.dcg.p2p.exceptions.DispatcherNotRunningException;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.exceptions.UDPSocketManagerException;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.peers.Source;
import ch.ethz.dcg.p2p.tasks.ByeTask;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * GUI for the <code>Source</code> which allows comfortable user interaction.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class SourceGUI extends GUI implements ActionListener {

	// menu
	private JMenuBar menuBar;
	private JMenu menu;
	private JMenuItem menuItem;
	private static final boolean[] EXIT = new boolean[] { false, false, false, false, true };
	private static final boolean[] JOIN_EXIT = new boolean[] { true, false, false, false, true };
	private static final boolean[] START_LEAVE_EXIT = new boolean[] { false, true, false, true, true };
	private static final boolean[] STOP_EXIT = new boolean[] { false, false, true, false, true };

	// tabs
	private JTabbedPane tabbedPane;

	// neighbours tab
	private JPanel pNeighbours;
	private TableLayout tlNeighbours;
	private JLabel lNeigh;
	private JScrollPane sNeighbours;
	private Hashtable<String, JLabel[]> neighbours = new Hashtable<String, JLabel[]>();
	private static final int ID = 0;
	private static final int PACKETSSENT = 1;
	private static final int NNEIGHCOLS = 2;

	// status tab
	private JTextArea taStatus;
	private JScrollPane sStatus;

	// statistics tab
	private JPanel pStatistics;
	private AbstractDataCollector downCollector;
	private AbstractDataCollector upCollector;
	private JLabel lPackRcvSnd;
	private JLabel lNotiRcvSnd;
	private JLabel lReqRcvSnd;
	private JLabel ldupPack;
	private JLabel ldupNoti;
	private JLabel lpacketId;
	private JLabel lKByteTransfer;

	// config tab
	private JTabbedPane tpConfig;
	private JPanel pBasicConfig;
	private JPanel pAdvConfig;
	private JLabel lConfig;
	private JButton bSet;
	private JTextField tfNepAddr;
	private JTextField tfStreamURL;
	private JComboBox cbStreamTypes;
	private JTextField tfStreamBitRate;
	private JTextField tfPacketSize;
	private JTextField tfIp;
	private JTextField tfPort;

	private JLabel statusBar;

	private Source source;


	public static void main( String[] args ){

		// Schedule a job for the event-dispatching thread:
		// creating and showing this application's GUI.
		javax.swing.SwingUtilities.invokeLater( new Runnable() {

			public void run(){

				createAndShowGUI();
			}
		} );
	}


	/**
	 * Creates and shows this gui.
	 */
	private static void createAndShowGUI(){

		JFrame.setDefaultLookAndFeelDecorated( true );
		new SourceGUI();
	}


	/**
	 * Creates a new <code>SourceGUI</code>.<br>
	 * A <code>Source</code> is created and the gui is set up.
	 */
	private SourceGUI(){

		// init Source
		source = new Source( this );

		// register GUI with the Statistic class
		Node.statistic.setGui( this );

		// init frame
		frame = new JFrame( "Swistry - Source" );
		ImageIcon icon = createImageIcon( "images/swistry_icon.gif");
		if( icon != null ){
			frame.setIconImage( icon.getImage() );
		}
		frame.addWindowListener( new WindowAdapter() {

			public void windowClosing( WindowEvent e ){

				frame.dispose();
				try{
					if( source.getStatus().equals( Node.started ) ){
						source.stop();
						new Thread( new ByeTask( source ), "ByeTask" ).start();
						Thread.sleep( 3000 );
					}else if( source.getStatus().equals( Node.connected ) ){
						new Thread( new ByeTask( source ), "ByeTask" ).start();
						Thread.sleep( 3000 );
					}
				}catch( InterruptedException ie ){
				}finally{
					System.exit( 0 );
				}
			}
		} );

		// define gui layout using a TableLayout
		double size[][] = { { TableLayout.FILL }, { TableLayout.FILL, 15 } };
		frame.setLayout( new TableLayout( size ) );

		// create the menus
		createMenus();

		// create the tab contents
		createNeighboursTab();
		createStatisticsTab();
		createStatusTab();
		createConfigTab();

		// create the tabs
		tabbedPane = new JTabbedPane();
		tabbedPane.setBorder( BorderFactory.createEmptyBorder( 10, 0, 0, 0 ) );
		tabbedPane.addTab( "Neighbours (0)", sNeighbours );
		tabbedPane.addTab( "Statistic", pStatistics );
		tabbedPane.addTab( "Status", sStatus );
		tabbedPane.addTab( "Config", tpConfig );
		tabbedPane.setEnabledAt( 0, false );
		tabbedPane.setEnabledAt( 1, false );
		tabbedPane.setEnabledAt( 2, true );
		tabbedPane.setEnabledAt( 3, true );
		tabbedPane.setSelectedIndex( 3 );

		// add status bar
		statusBar = new JLabel( " idle" );
		statusBar.setFont( new Font( null, Font.PLAIN, 10 ) );

		// add the above componentes to the frame
		frame.setJMenuBar( menuBar );
		frame.getContentPane().add( tabbedPane, "0,0" );
		frame.getContentPane().add( statusBar, "0,1" );

		// show gui
		frame.pack();
		frame.setSize( new Dimension( 600, 400 ) );
		frame.setVisible( true );
	}


	/**
	 * Creates the menus of the gui.
	 */
	private void createMenus(){

		menuBar = new JMenuBar();

		// Go!
		menu = new JMenu( "Go!" );
		menuBar.add( menu );
		// join
		menuItem = new JMenuItem( "Join Network" );
		menuItem.setActionCommand( "join" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( false );
		menu.add( menuItem );
		// start
		menuItem = new JMenuItem( "Start Stream" );
		menuItem.setActionCommand( "start" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( false );
		menu.add( menuItem );
		// stop
		menuItem = new JMenuItem( "Stop Stream" );
		menuItem.setActionCommand( "stop" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( false );
		menu.add( menuItem );
		// leave
		menuItem = new JMenuItem( "Leave Network" );
		menuItem.setActionCommand( "leave" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( false );
		menu.add( menuItem );
		// exit
		menuItem = new JMenuItem( "Exit" );
		menuItem.setActionCommand( "exit" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( true );
		menu.add( menuItem );

		// HELP
		menu = new JMenu( "Help" );
		menuBar.add( menu );
		// about
		menuItem = new JMenuItem( "About Swistry" );
		menuItem.setActionCommand( "about" );
		menuItem.addActionListener( this );
		menuItem.setEnabled( true );
		menu.add( menuItem );
	}


	/**
	 * Creates the tab containing the per neighbour information.
	 */
	private void createNeighboursTab(){

		double size[][] = { { 150, 5, 150, TableLayout.FILL }, { 20, 1 } };
		tlNeighbours = new TableLayout( size );
		pNeighbours = new JPanel( tlNeighbours );
		pNeighbours.setBorder( BorderFactory.createEmptyBorder( 5, 0, 0, 0 ) );

		// header
		lNeigh = new JLabel( "Peer" );
		lNeigh.setFont( new Font( null, Font.PLAIN, 12 ) );
		pNeighbours.add( lNeigh, "0,0,c,c" );

		pNeighbours.add( createHeaderLabel( "Packets sent" ), "2,0,c,c" );

		lNeigh = new JLabel();
		lNeigh.setOpaque( true );
		lNeigh.setBackground( Color.BLACK );
		pNeighbours.add( lNeigh, "0,1,3,1" );
		sNeighbours = new JScrollPane( pNeighbours );
	}


	/**
	 * Creates a formated header label.
	 * 
	 * @param text
	 *            the text of the header label
	 * @return a JLabel containing the specified text with the header formatting
	 */
	private JLabel createHeaderLabel( String text ){

		JLabel label = new JLabel( text );
		label.setFont( new Font( null, Font.PLAIN, 10 ) );
		return label;
	}


	/**
	 * Rebuilds the table with the per neighbour information.<br>
	 * After this operation only neighbours which are still present will be left
	 * in the table.
	 */
	private void rebuildNeighbourTable(){

		JLabel labels[];
		Iterator<JLabel[]> it;

		if( neighbours.size() == 0 ){
			// rebuild empty table
			for( int i = tlNeighbours.getNumRow() - 1; i >= 2; i-- ){
				tlNeighbours.deleteRow( i );
			}
		}else{
			// delete all still existing neighbour components
			it = neighbours.values().iterator();
			while( it.hasNext() ){
				labels = it.next();
				for( JLabel curL : labels ){
					tlNeighbours.removeLayoutComponent( curL );
				}
			}

			// delete all neighbour rows (except for the header)
			for( int i = tlNeighbours.getNumRow() - 1; i >= 2; i-- ){
				tlNeighbours.deleteRow( i );
			}

			// insert all neighbour components again
			it = neighbours.values().iterator();
			int rows = tlNeighbours.getNumRow();
			while( it.hasNext() ){
				// create new row
				tlNeighbours.insertRow( rows, 5 );
				tlNeighbours.insertRow( rows + 1, 10 );
				// add labels to row
				labels = it.next();
				for( int i = 0; i < NNEIGHCOLS; i++ ){
					pNeighbours.add( labels[i], 2 * i + "," + ( rows + 1 ) + ",c,c" );
				}
				rows += 2;
			}
		}
		// update tab title
		tabbedPane.setTitleAt( 0, "Neighbours (" + neighbours.size() + ")" );
		tabbedPane.setSelectedIndex( 0 );
		pNeighbours.repaint();
	}


	/**
	 * Creates tab which contains status informations.
	 */
	private void createStatusTab(){

		taStatus = new JTextArea();
		taStatus.setEditable( false );
		taStatus.setFont( new Font( null, Font.PLAIN, 10 ) );
		addStatusLine( "idle: configuration required" );
		sStatus = new JScrollPane( taStatus );
	}


	/**
	 * Create the tab containing the overall statistics.
	 */
	private void createStatisticsTab(){

		// panel
		double size[][] = { { 200, TableLayout.FILL }, { 100, 20, 10, 10, 10, 10, 10, 20, 10, 10, 10, 20, 10, 20, 10 } };
		pStatistics = new JPanel( new TableLayout( size ) );
		pStatistics.setBorder( BorderFactory.createEmptyBorder( 5, 5, 5, 5 ) );

		// chart
		Chart2D loadChart = new Chart2D();
		ITrace2D downTrace = new Trace2DLtd( 100, "download [kbps]" );
		ITrace2D upTrace = new Trace2DLtd( 100, "upload [kbps]" );
		downTrace.setColor( Color.GREEN );
		upTrace.setColor( Color.RED );
		loadChart.addTrace( downTrace );
		loadChart.addTrace( upTrace );
		pStatistics.add( loadChart, "0,0,1,0" );

		// packets
		pStatistics.add( createStatLabel( "packets received / sent: " ), "0,2,l,c" );
		lPackRcvSnd = new JLabel( "0 / 0" );
		lPackRcvSnd.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( lPackRcvSnd, "1,2,l,c" );

		// requests
		pStatistics.add( createStatLabel( "requests received / sent: " ), "0,4,l,c" );
		lReqRcvSnd = new JLabel( "0 / 0" );
		lReqRcvSnd.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( lReqRcvSnd, "1,4,l,c" );

		// notifications
		pStatistics.add( createStatLabel( "notifications received / sent: " ), "0,6,l,c" );
		lNotiRcvSnd = new JLabel( "0 / 0" );
		lNotiRcvSnd.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( lNotiRcvSnd, "1,6,l,c" );

		// duplicate packets
		pStatistics.add( createStatLabel( "duplicate packets received: " ), "0,8,l,c" );
		ldupPack = new JLabel( new Long( Node.statistic.getDuplicatePacketsReceived() ).toString() );
		ldupPack.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( ldupPack, "1,8,l,c" );

		// duplicate notifications
		pStatistics.add( createStatLabel( "duplicate notifications received: " ), "0,10,l,c" );
		ldupNoti = new JLabel( new Long( Node.statistic.getDuplicateNotificationsReceived() ).toString() );
		ldupNoti.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( ldupNoti, "1,10,l,c" );

		// current packet id
		pStatistics.add( createStatLabel( "latest available packet id: " ), "0,12,l,c" );
		lpacketId = new JLabel( new Long( source.getCurPacketId() ).toString() );
		lpacketId.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( lpacketId, "1,12,l,c" );

		// total kbytes
		pStatistics.add( createStatLabel( "kilobytes received / sent: " ), "0,14,l,c" );
		lKByteTransfer = new JLabel( "0kB / 0kB" );
		lKByteTransfer.setFont( new Font( null, Font.PLAIN, 10 ) );
		pStatistics.add( lKByteTransfer, "1,14,l,c" );

		// init datacollectors
		downCollector = new DownloadDataCollector( downTrace, 3000 );
		upCollector = new UploadDataCollector( upTrace, 3000 );
	}


	/**
	 * Create a formatted label for the statistic.
	 * 
	 * @param text
	 *            the text of the label
	 * @return JLabel containing the specified text formatted as required
	 */
	private JLabel createStatLabel( String text ){

		JLabel label = new JLabel( text );
		label.setFont( new Font( null, Font.PLAIN, 10 ) );
		return label;
	}


	/**
	 * Creates the tab containing the basic- and advanced configurations.
	 */
	private void createConfigTab(){

		createBasicConfigTab();
		createAdvancedConfigTab();

		tpConfig = new JTabbedPane();
		tpConfig.setBorder( BorderFactory.createEmptyBorder( 10, 0, 0, 0 ) );
		tpConfig.add( "Basic", pBasicConfig );
		tpConfig.add( "Advanced", pAdvConfig );
	}


	/**
	 * Creates the basic configuration tab.
	 */
	private void createBasicConfigTab(){

		// panel
		double size[][] = { { 50, 5, 40, 10, 100, 10, TableLayout.FILL }, { 20, 20, 20, 10, 20, 10, 20, 10, 20, TableLayout.FILL } };
		pBasicConfig = new JPanel( new TableLayout( size ) );
		pBasicConfig.setBorder( BorderFactory.createEmptyBorder( 10, 5, 5, 5 ) );

		// NEP address
		tfNepAddr = new JTextField( "", 10 );
		pBasicConfig.add( tfNepAddr, "0,0,4,0" );
		lConfig = new JLabel( "IP address or hostname of the Network Entry Point" );
		pBasicConfig.add( lConfig, "6,0,l,c" );

		// stream URL
		tfStreamURL = new JTextField( "http://", 20 );
		pBasicConfig.add( tfStreamURL, "0,2,4,2,l,c" );
		lConfig = new JLabel( "URL of the live input stream" );
		pBasicConfig.add( lConfig, "6,2,l,c" );

		// stream types
		String[] inputStreamTypes = { " Shoutcast ", " VLC " };
		cbStreamTypes = new JComboBox( inputStreamTypes );
		cbStreamTypes.setSelectedIndex( 0 );
		pBasicConfig.add( cbStreamTypes, "0,4,2,4,l,c" );
		lConfig = new JLabel( "stream type" );
		pBasicConfig.add( lConfig, "4,4,6,4,l,c" );

		// stream bit rate
		tfStreamBitRate = new JTextField( "", 4 );
		pBasicConfig.add( tfStreamBitRate, "0,6,l,c" );
		lConfig = new JLabel( "Kbit/s" );
		lConfig.setFont( new Font( null, Font.PLAIN, 12 ) );
		pBasicConfig.add( lConfig, "2,6,l,c" );
		lConfig = new JLabel( "stream bitrate" );
		pBasicConfig.add( lConfig, "4,6,6,6,l,c" );

		// packet size
		tfPacketSize = new JTextField( "16000", 4 );
		pBasicConfig.add( tfPacketSize, "0,8,l,c" );
		lConfig = new JLabel( "bytes" );
		lConfig.setFont( new Font( null, Font.PLAIN, 12 ) );
		pBasicConfig.add( lConfig, "2,8,l,c" );
		lConfig = new JLabel( "packet size (payload only - max 60000)" );
		pBasicConfig.add( lConfig, "4,8,6,8,l,c" );

		// button
		bSet = new JButton( "Set" );
		bSet.setActionCommand( "set" );
		bSet.addActionListener( this );
		pBasicConfig.add( bSet, "0,9,2,9,l,b" );
	}


	/**
	 * Creates the advanced configuration tab.
	 */
	private void createAdvancedConfigTab(){

		// panel
		double size[][] = { { 60, 10, 105, 10, TableLayout.FILL }, { 20, 5, 20, 20, 20 } };
		pAdvConfig = new JPanel( new TableLayout( size ) );
		pAdvConfig.setBorder( BorderFactory.createEmptyBorder( 10, 5, 5, 5 ) );

		// host address
		tfIp = new JTextField( whatIsMyIp().getHostAddress(), 8 );
		pAdvConfig.add( tfIp, "0,0,2,0" );
		lConfig = new JLabel( "public IP address or hostname of this host" );
		pAdvConfig.add( lConfig, "4,0,l,c" );

		// host port
		tfPort = new JTextField( new Integer( 65001 ).toString(), 4 );
		pAdvConfig.add( tfPort, "0,2,l,c" );
		lConfig = new JLabel( "port [49152...65535]" );
		pAdvConfig.add( lConfig, "2,2,4,2,l,c" );
	}


	/**
	 * Gets this hosts IP address by making a request to
	 * http://www.whatismyip.com and parsing the answer.
	 * 
	 * @return the public IP address of this host as reported by whatismyip.com
	 */
	private InetAddress whatIsMyIp(){

		InetAddress ip = null;
		try{
			URL url = new URL( "http://www.whatismyip.com" );
			BufferedReader reader = new BufferedReader( new InputStreamReader( url.openStream() ) );
			String curLine = reader.readLine();
			while( ( curLine != null ) && !curLine.startsWith( "<TITLE>" ) ){
				curLine = reader.readLine();
			}
			if( curLine != null ){
				int begin = curLine.indexOf( " - " ) + 3;
				int end = curLine.indexOf( "</TITLE>" );
				ip = InetAddress.getByName( curLine.substring( begin, end ) );
				return ip;
			}else{
				return null;
			}
		}catch( MalformedURLException e ){
			// should not happen ;)
		}catch( IOException e ){
			try{
				// try the local host address
				return InetAddress.getLocalHost();
			}catch( UnknownHostException uhe ){
				try{
					// try 'localhost'
					return InetAddress.getByName( "localhost" );
				}catch( UnknownHostException e1 ){
					// should not happen ;)
				}
			}
		}
		return null;
	}


	/**
	 * Enables or disable sthe gui elements in the config tab.
	 * 
	 * @param enabled
	 *            true will enable the elements; false will disable them
	 */

	private void setConfigEnabled( boolean enabled ){

		tfNepAddr.setEnabled( enabled );
		tfStreamURL.setEnabled( enabled );
		cbStreamTypes.setEnabled( enabled );
		tfStreamBitRate.setEnabled( enabled );
		tfPacketSize.setEnabled( enabled );
		tfIp.setEnabled( enabled );
		tfPort.setEnabled( enabled );
		bSet.setEnabled( enabled );
	}


	/**
	 * Enables or disables the status tab.
	 * 
	 * @param enabled
	 *            true will enable the elements; false will disable them
	 */
	private void setStatusEnabled( boolean enabled ){

		tabbedPane.setEnabledAt( 2, enabled );
	}


	/**
	 * Enables or disables the statistic tab.
	 * 
	 * @param enabled
	 *            true will enable the elements; false will disable them
	 */
	private void setStatisticEnabled( boolean enabled ){

		tabbedPane.setEnabledAt( 1, enabled );
	}


	/**
	 * Enables or disables the neighbours tab.
	 * 
	 * @param enabled
	 *            true will enable the elements; false will disable them
	 */
	private void setNeighboursEnabled( boolean enabled ){

		tabbedPane.setEnabledAt( 0, enabled );
	}


	/**
	 * Enables or disables different menu items using the predefined constants.
	 * 
	 * @param enabled
	 *            boolean array that specifies which menuitems should be enabled
	 *            and which should be disabled:<br>
	 *            EXIT = exit only<br>
	 *            JOIN_EXIT = join and exit<br>
	 *            START_LEAVE_EXIT = start and leave and exit<br>
	 *            STOP_EXIT = stop and exit<br>
	 */
	private void setMenuItemsEnabled( boolean[] enabled ){

		for( int i = 0; i < enabled.length; i++ ){
			menuBar.getMenu( 0 ).getMenuComponent( i ).setEnabled( enabled[i] );
		}
	}


	@Override
	public boolean addNeighbour( String id ){

		if( neighbours.containsKey( id ) ){
			return false;
		}

		// create labels
		JLabel[] labels = new JLabel[NNEIGHCOLS];
		labels[ID] = new JLabel( id );
		labels[PACKETSSENT] = new JLabel( "0" );

		// add neighbour to hashtable
		neighbours.put( id, labels );

		// create new row
		int rows = tlNeighbours.getNumRow();
		tlNeighbours.insertRow( rows++, 5 );
		tlNeighbours.insertRow( rows, 10 );

		// add labels to new row
		for( int i = 0; i < NNEIGHCOLS; i++ ){
			labels[i].setFont( new Font( null, Font.PLAIN, 10 ) );
			pNeighbours.add( labels[i], 2 * i + "," + rows + ",c,c" );
		}
		labels[0].setForeground( Color.BLUE );

		// update tab title
		tabbedPane.setTitleAt( 0, "Neighbours (" + neighbours.size() + ")" );
		tabbedPane.setSelectedIndex( 0 );
		return true;
	}


	@Override
	public boolean removeNeighbour( String id ){

		if( !neighbours.containsKey( id ) ){
			return false;
		}else{
			JLabel[] labels = neighbours.remove( id );
			for( JLabel curL : labels ){
				pNeighbours.remove( curL );
				tlNeighbours.removeLayoutComponent( curL );
			}
			rebuildNeighbourTable();
			return true;
		}
	}


	@Override
	public void addStatusLine( String line ){

		if( taStatus != null ){
			taStatus.setText( "[" + getTimeStamp() + "] " + line + "\n" + taStatus.getText() );
		}
	}


	@Override
	public void errorStatusBar( String s ){

		statusBar.setForeground( Color.RED );
		statusBar.setText( " " + s );
	}


	@Override
	public void errorDialog( String message, String title ){

		JOptionPane.showMessageDialog( frame, message, title, JOptionPane.ERROR_MESSAGE );
	}


	/**
	 * Part of the <code>ActionListener</code> interface.<br>
	 * It is basically a dispatcher that reacts upon gui interaction.
	 */
	public void actionPerformed( ActionEvent e ){

		// SET
		if( "set".equals( e.getActionCommand() ) ){
			// get values
			InetAddress nepAddress = null;
			URL streamURL = null;
			InputAdapter adapter = null;
			byte streamType = ProtocolConstants.UNKNOWN_TYPE;
			int streamBitRate = 0;
			int packetSize = 0;
			InetAddress ip = null;
			int port = 65001;
			// basic config
			// network entry point
			try{
				if( isValid( tfNepAddr.getText() ) ){
					nepAddress = InetAddress.getByName( tfNepAddr.getText() );
				}else{
					tfNepAddr.setForeground( Color.RED );
					errorStatusBar( "invalid network entry point address" );
					tpConfig.setSelectedIndex( 0 );
					return;
				}
			}catch( UnknownHostException e1 ){
				tfNepAddr.setForeground( Color.RED );
				errorStatusBar( "invalid network entry point address" );
				tpConfig.setSelectedIndex( 0 );
				return;
			}
			if( tfNepAddr.getForeground().equals( Color.RED ) ){
				tfNepAddr.setForeground( Color.BLACK );
			}
			// URL of the input stream
			try{
				streamURL = new URL( tfStreamURL.getText() );
				if( tfStreamURL.getText().equalsIgnoreCase( "http://" ) ){
					throw new MalformedURLException( "empty url" );
				}
			}catch( MalformedURLException e1 ){
				tfStreamURL.setForeground( Color.RED );
				errorStatusBar( "invalid input stream url" );
				tpConfig.setSelectedIndex( 0 );
				return;
			}
			if( tfStreamURL.getForeground().equals( Color.RED ) ){
				tfStreamURL.setForeground( Color.BLACK );
			}
			// stream type (cannot be invalid since it is a combo box)
			String selected = (String)cbStreamTypes.getSelectedItem();
			if( selected.equalsIgnoreCase( " Shoutcast " ) ){
				streamType = ProtocolConstants.HTTP_MP3;
				adapter = new ShoutCastAdapter( streamURL, 3000, source );
			}else if( selected.equalsIgnoreCase( " VLC " ) ){
				streamType = ProtocolConstants.HTTP_VLC;
				adapter = new VlcHttpAdapter( streamURL );
			}
			// stream bitrate
			try{
				streamBitRate = Integer.parseInt( tfStreamBitRate.getText() );
				if( streamBitRate <= 0 ){
					throw new NumberFormatException( "stream bitrate <= 0" );
				}
			}catch( NumberFormatException nfe ){
				tfStreamBitRate.setForeground( Color.RED );
				tfStreamBitRate.setText( "0" );
				errorStatusBar( "invalid stream bitrate" );
				tpConfig.setSelectedIndex( 0 );
				return;
			}
			if( tfStreamBitRate.getForeground().equals( Color.RED ) ){
				tfStreamBitRate.setForeground( Color.BLACK );
			}
			// packet size
			try{
				packetSize = Integer.parseInt( tfPacketSize.getText() );
				if( packetSize <= 0 ){
					throw new NumberFormatException( "packet size <=0" );
				}else if( packetSize > 60000 ){
					throw new NumberFormatException( "packet size >600000" );
				}
			}catch( NumberFormatException nfe ){
				tfPacketSize.setForeground( Color.RED );
				tfPacketSize.setText( "0" );
				errorStatusBar( "invalid packet size" );
				tpConfig.setSelectedIndex( 0 );
				return;
			}
			if( tfPacketSize.getForeground().equals( Color.RED ) ){
				tfPacketSize.setForeground( Color.BLACK );
			}

			// advanced config
			// IP
			try{
				if( isValid( tfIp.getText() ) ){
					ip = InetAddress.getByName( tfIp.getText() );
				}else{
					tfIp.setForeground( Color.RED );
					errorStatusBar( "invalid IP address" );
					tpConfig.setSelectedIndex( 1 );
					return;
				}
			}catch( UnknownHostException e1 ){
				tfIp.setForeground( Color.RED );
				errorStatusBar( "invalid IP address" );
				tpConfig.setSelectedIndex( 1 );
				return;
			}
			if( tfIp.getForeground().equals( Color.RED ) ){
				tfIp.setForeground( Color.BLACK );
			}
			// port
			try{
				port = Integer.parseInt( tfPort.getText() );
				if( ( port < 0 ) || ( port > 65535 ) ){
					tfPort.setForeground( Color.RED );
					errorStatusBar( "invalid port" );
					tpConfig.setSelectedIndex( 1 );
					return;
				}
			}catch( NumberFormatException nfe ){
				tfPort.setForeground( Color.RED );
				tfPort.setText( "65002" );
				errorStatusBar( "invalid port" );
				tpConfig.setSelectedIndex( 1 );
				return;
			}
			if( tfPort.getForeground().equals( Color.RED ) ){
				tfPort.setForeground( Color.BLACK );
			}
			// clean up if we are in an error state
			if( source.getStatus().equals( Node.error ) ){
				reset();
			}

			// if this point is reached, all the config input is valid
			// configure and initialise the source
			source.configure( ip, port, streamBitRate, nepAddress, streamType );
			source.initPacketizer( packetSize );
			source.setInputAdapter( adapter );
		}

		// JOIN
		if( "join".equals( e.getActionCommand() ) ){
			// join network
			setMenuItemsEnabled( EXIT );
			if( source.getStatus().equals( Node.error ) ){
				reset();
			}
			try{
				source.startDispatchers();
				source.joinNetwork();
				//source.unleashZombieKiller();
			}catch( DispatcherNotRunningException e1 ){
				source.setStatus( Node.error );
				addStatusLine( "[ERROR] " + e1.getMessage() );
			}catch( NodeNotConfiguredException e1 ){
				source.setStatus( Node.error );
				addStatusLine( "[ERROR] " + e1.getMessage() );
			}catch( UDPSocketManagerException e1 ){
				source.setStatus( Node.error );
				addStatusLine( "[ERROR] " + e1.getMessage() );
			}
			tabbedPane.setSelectedIndex( 0 );
		}

		// START
		if( "start".equals( e.getActionCommand() ) ){
			// start the stream
			downCollector.start();
			upCollector.start();
			source.start();
			tabbedPane.setSelectedIndex( 1 );
		}

		// STOP
		if( "stop".equals( e.getActionCommand() ) ){
			// stop the stream
			source.stop();
			tabbedPane.setSelectedIndex( 1 );
		}

		// LEAVE
		if( "leave".equals( e.getActionCommand() ) ){
			reset();
		}

		// EXIT
		if( "exit".equals( e.getActionCommand() ) ){
			frame.dispose();
			try{
				if( source.getStatus().equals( Node.connected ) || source.getStatus().equals( Node.started ) ){
					new Thread( new ByeTask( source ), "ByeTask" ).start();
					Thread.sleep( 3000 );
				}
			}catch( InterruptedException ie ){
			}finally{
				System.exit( 0 );
			}
		}

		// ABOUT
		if( "about".equals( e.getActionCommand() ) ){
			showAboutDialog();
		}
	}


	/**
	 * Tests whether the given string represents a valid IP address.
	 * 
	 * @param ip
	 * @return true if the IP address is valid; false otherwise
	 */
	private static boolean isValid( String ip ){

		try{
			if( ip.equals( "" ) || ip == null ){
				return false;
			}
			new URL( "http://" + ip );
			return true;
		}catch( MalformedURLException e ){
			return false;
		}
	}


	/**
	 * Cleans up the peer after leaving the network or when recovering from an
	 * error.
	 * 
	 */
	private void reset(){

		// leave network
		source.leaveNetwork();
		// reset neighbours heap
		neighbours = new Hashtable<String, JLabel[]>();
		// reset neighbours
		tabbedPane.remove( 0 );
		createNeighboursTab();
		tabbedPane.add( pNeighbours, 0 );
		tabbedPane.setTitleAt( 0, "Neighbours (0)" );
		tabbedPane.setEnabledAt( 0, false );
		// reset statistics
		downCollector.stop();
		upCollector.stop();
		tabbedPane.remove( 1 );
		createStatisticsTab();
		tabbedPane.add( pStatistics, 1 );
		tabbedPane.setTitleAt( 1, "Statistic" );
		tabbedPane.setEnabledAt( 1, false );
	}


	@Override
	public void updateTotals(){

		// update global statistic
		lPackRcvSnd.setText( Node.statistic.getTotalPacketsReceived() + " / " + Node.statistic.getTotalPacketsSent() );
		lReqRcvSnd.setText( Node.statistic.getTotalRequestsReceived() + " / " + Node.statistic.getTotalRequestsSent() );
		lNotiRcvSnd.setText( Node.statistic.getTotalNotificationsReceived() + " / " + Node.statistic.getTotalNotificationsSent() );
		ldupPack.setText( String.valueOf( Node.statistic.getDuplicatePacketsReceived() ) );
		ldupNoti.setText( String.valueOf( Node.statistic.getDuplicateNotificationsReceived() ) );
		lpacketId.setText( String.valueOf( source.getCurPacketId() ) );
		lKByteTransfer.setText( Node.statistic.getTotalKBytesReceived() + "kB / " + Node.statistic.getTotalKBytesSent() + "kB" );
	}


	@Override
	public void updateStatus(){

		String status = source.getStatus();
		statusBar.setForeground( Color.BLACK );
		if( status.equals( Node.idle ) ){
			setNeighboursEnabled( false );
			setStatisticEnabled( false );
			setStatusEnabled( true );
			setConfigEnabled( true );
			setMenuItemsEnabled( EXIT );
		}else if( status.equals( Node.configured ) ){
			setNeighboursEnabled( false );
			setStatisticEnabled( false );
			setStatusEnabled( true );
			setConfigEnabled( true );
			setMenuItemsEnabled( JOIN_EXIT );
		}else if( status.equals( Node.connected ) ){
			setNeighboursEnabled( true );
			setStatisticEnabled( true );
			setStatusEnabled( true );
			setConfigEnabled( false );
			setMenuItemsEnabled( START_LEAVE_EXIT );
		}else if( status.equals( Node.started ) ){
			setNeighboursEnabled( true );
			setStatisticEnabled( true );
			setStatusEnabled( true );
			setConfigEnabled( false );
			setMenuItemsEnabled( STOP_EXIT );
		}else if( status.equals( Node.error ) ){
			statusBar.setForeground( Color.RED );
			if( source.getOldStatus().equals( Node.configured ) ){
				// go back to configured state
				setNeighboursEnabled( false );
				setStatisticEnabled( false );
				setStatusEnabled( true );
				setConfigEnabled( true );
				setMenuItemsEnabled( JOIN_EXIT );
			}else if( source.getOldStatus().equals( Node.connected ) ){
				// go back to connected state
				setNeighboursEnabled( true );
				setStatisticEnabled( true );
				setStatusEnabled( true );
				setConfigEnabled( false );
				setMenuItemsEnabled( START_LEAVE_EXIT );
			}
			tabbedPane.setSelectedIndex( 2 );
		}
		statusBar.setText( " " + status );
	}


	@Override
	public boolean updatePacketsReceived( String neighbourKey, Long received ){

		return true;
	}


	@Override
	public boolean updatePacketsSent( String neighbourKey, Long sent ){

		if( neighbours.containsKey( neighbourKey ) ){
			neighbours.get( neighbourKey )[PACKETSSENT].setText( sent.toString() );
			return true;
		}else{
			return false;
		}
	}


	@Override
	public boolean updateRequestsReceived( String neighbourKey, Long received ){

		return true;
	}


	@Override
	public boolean updateRequestsSent( String neighbourKey, Long sent ){

		return true;
	}


	@Override
	public boolean updateNotificationsReceived( String neighbourKey, Long received ){

		return true;
	}


	@Override
	public boolean updateNotificationsSent( String neighbourKey, Long sent ){

		return true;
	}


	@Override
	public boolean updateDuplicatePacketsReceived( Long dupPackets ){

		ldupPack.setText( dupPackets.toString() );
		return true;
	}


	@Override
	public boolean updateDuplicateNotificationsReceived( Long dupNotifications ){

		ldupNoti.setText( dupNotifications.toString() );
		return true;
	}


	@Override
	public boolean updateNeighbourRTT( String neighbourKey, Integer newDelay ){

		return true;
	}


	@Override
	public boolean updateNeighbourRanking( String neighbourKey, Integer newRanking ){

		return true;
	}
}
