/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;

import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * Handler for incoming BYECONFIRM messages.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class ByeConfirmHandler extends Task {

	/**
	 * Creates a new <code>ByeConfirmHandler</code>.
	 * 
	 * @param socket
	 *            the TCP connection connection to the sender
	 * @param source
	 *            the owner of this task
	 */
	public ByeConfirmHandler( Socket socket, Node source ){

		this.socket = socket;
		this.source = source;
	}


	/**
	 * Handles the BYECONFIRM message:<br>
	 * respond with a BYEACK or BYENACK message.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "ByeConfirmHandler started" );
		}

		try{
			DataInputStream in = new DataInputStream( socket.getInputStream() );
			DataOutputStream out = new DataOutputStream( socket.getOutputStream() );
			// read BYECONFIRM message
			InetAddress askingIp = socket.getInetAddress();
			int askingPort = in.readInt();
			if( logger.isInfoEnabled() ){
				logger.info( "BYECONFIRM message received from " + askingIp.getHostAddress() + ":" + askingPort );
			}
			if( !source.inStatus( Node.configured ) ){
				// I am not leaving the network; send NACK
				out.write( ProtocolConstants.BYENACK );
				if( logger.isInfoEnabled() ){
					logger.info( "BYENACK message sent to " + askingIp.getHostAddress() + ":" + askingPort );
				}
			}else{
				// send BYEACK
				out.write( ProtocolConstants.BYEACK );
				if( logger.isInfoEnabled() ){
					logger.info( "BYEACK message sent to " + askingIp.getHostAddress() + ":" + askingPort );
				}
			}
			out.flush();
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( "cannot confirm BYECONFIRM message ( " + e.getMessage() + " )" );
			}
		}
	}
}
