/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketTimeoutException;

import ch.ethz.dcg.p2p.exceptions.ByeHandlerException;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.NEPNeighbour;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.NetworkEntryPoint;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Handler for incoming BYE messages.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class ByeHandler extends Task {

	private InetAddress leavingIp;
	private int leavingPort;
	private NetworkEntryPoint nep = null;
	private int timeout = 1000; // ms


	/**
	 * Creates a new <code>ByeHandler</code>.<br>
	 * This task can either be started by a <code>Node</code> or by a
	 * <code>NetworkEntryPoint</code>. If it is started by a
	 * <code>Node</code> then the nep argument is null and vice versa.
	 * 
	 * @param socket
	 *            the TCP connection connection to the sender
	 * @param source
	 *            possible owner of this task
	 * @param nep
	 *            possible owner of this task
	 */
	public ByeHandler( Socket socket, Node source, NetworkEntryPoint nep ){

		this.socket = socket;
		this.source = source;
		this.nep = nep;
	}


	/**
	 * Handles the BYE message:<br>
	 * confirm the BYE message using a BYECONFIRM and delete the sender if it
	 * responds with a BYEACK message.
	 */
	public void run(){

		if( source == null && nep == null ){
			if( logger.isErrorEnabled() ){
				logger.error( "cannot start ByeHandler: nep and source pointers are both null" );
			}
			return;
		}
		if( logger.isDebugEnabled() ){
			logger.debug( "ByeHandler started" );
		}

		try{
			// receive the bye message
			getByeMessage();
			if( source != null ){
				// ByeHandler started by Node
				if( !source.hasNeighbour( leavingIp, leavingPort ) ){
					if( logger.isWarnEnabled() ){
						logger.warn( "BYE message from unknown node received; ignoring it" );
					}
					return;
				}
				if( confirmBye() ){
					source.removeNeighbour( leavingIp, leavingPort );
					source.write2gui( "peer " + leavingIp.getHostAddress() + ":" + leavingPort + " has left" );
				}
				source.printNeighbours();
			}else if( nep != null ){
				// ByeHandler started by NetworkEntryPoint
				NEPNeighbour oldNeighbour = nep.containsNEPNeighbour( leavingIp, leavingPort );
				if( oldNeighbour == null ){
					if( logger.isWarnEnabled() ){
						logger.warn( "BYE message from node received that is not present; ignoring it" );
					}
					return;
				}
				if( confirmBye() ){
					if( oldNeighbour.getNodeType() == ProtocolConstants.SOURCE ){
						nep.removeLevel( oldNeighbour.getLevel() );
					}
					nep.removeNEPNeighbour( leavingIp, leavingPort );
					nep.write2gui( "peer " + leavingIp.getHostAddress() + ":" + leavingPort + " has left" );
				}
				nep.printNeighbours();
			}
		}catch( ByeHandlerException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}


	/**
	 * Receives the BYE message.
	 * 
	 * @throws ByeHandlerException
	 *             if a communication error occurred
	 */
	private void getByeMessage() throws ByeHandlerException{

		try{
			DataInputStream in = new DataInputStream( socket.getInputStream() );
			leavingIp = Utilities.readIP( in );
			leavingPort = in.readInt();
			if( logger.isInfoEnabled() ){
				logger.info( "BYE message received from " + leavingIp.getHostAddress() + ":" + leavingPort );
			}
		}catch( IOException e ){
			throw new ByeHandlerException( "cannot get BYE message from " + leavingIp.getHostAddress() + ":" + leavingPort );
		}
	}


	/**
	 * Confirms the received BYE message (security feature).
	 * 
	 * @return true if the BYE message was confirmed; false otherwise
	 * @throws ByeHandlerException
	 *             if a communication error occurred
	 */
	private boolean confirmBye() throws ByeHandlerException{

		if( !isReachable( timeout ) ){
			// test if the sender is still available
			return true;
		}
		Socket s = null;
		try{
			// send BYECONFIRM message
			s = new Socket( leavingIp, leavingPort );
			s.setSoTimeout( 3000 );
			DataOutputStream out = new DataOutputStream( s.getOutputStream() );
			out.write( ProtocolConstants.BYECONFIRM );
			if( source != null ){
				out.writeInt( source.getPort() );
			}else{
				out.writeInt( ProtocolConstants.NEPPORT );
			}
			out.flush();
			if( logger.isInfoEnabled() ){
				logger.info( "BYECONFIRM message sent to " + leavingIp.getHostAddress() + ":" + leavingPort );
			}
			// read BYEACK or BYENACK
			DataInputStream in = new DataInputStream( s.getInputStream() );
			byte answer = (byte)in.read();
			if( answer == ProtocolConstants.BYEACK ){
				if( logger.isInfoEnabled() ){
					logger.info( "BYEACK message received from " + leavingIp.getHostAddress() + ":" + leavingPort );
				}
				return true;
			}else if( answer == ProtocolConstants.BYENACK ){
				if( logger.isInfoEnabled() ){
					logger.info( "BYENACK message received from " + leavingIp.getHostAddress() + ":" + leavingPort );
				}
				return false;
			}else{
				if( logger.isInfoEnabled() ){
					logger.info( "unknown answer type received; ignoring BYE message" );
				}
				return false;
			}
		}catch( SocketTimeoutException e ){
			return true;
		}catch( IOException e ){
			throw new ByeHandlerException( "cannot confirm BYE message; ignoring BYE message" );
		}finally{
			try{
				s.close();
			}catch( Exception e ){
			}
		}
	}


	/**
	 * Tests if the sender is responding to a PING message.
	 * 
	 * @param timeout
	 *            ping timeout
	 * @return true if the sender is responding; false otherwise
	 */
	private boolean isReachable( int timeout ){

		return new Neighbour( leavingIp, leavingPort ).isReachable( timeout );
	}
}
