/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.Utilities;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * Handler for incoming DROPPEDCONFIRM messages.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class DroppedConfirmHandler extends Task {

	/**
	 * Creates a new <code>DroppedConfirmHandler</code>.
	 * 
	 * @param socket
	 *            the TCP connection connection to the sender
	 * @param source
	 *            the owner of this task
	 */
	public DroppedConfirmHandler( Socket socket, Node source ){

		this.socket = socket;
		this.source = source;
	}


	/**
	 * Handles the DROPPEDCONFIRM message:<br>
	 * respond with a DROPPEDACK or DROPPEDNACK message.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "DroppedConfirmHandler started" );
		}

		try{
			DataInputStream in = new DataInputStream( socket.getInputStream() );
			DataOutputStream out = new DataOutputStream( socket.getOutputStream() );
			// read DROPPEDCONFIRM message
			InetAddress askingIp = Utilities.readIP( in );
			int askingPort = in.readInt();
			if( logger.isInfoEnabled() ){
				logger.info( "DROPPEDCONFIRM message received from " + askingIp.getHostAddress() + ":" + askingPort );
			}
			if( source.hasNeighbour( askingIp, askingPort ) ){
				// send DROPPEDNACK; I still have him as my neighbour
				out.write( ProtocolConstants.DROPPEDNACK );
				if( logger.isInfoEnabled() ){
					logger.info( "DROPPEDNACK message sent to " + askingIp.getHostAddress() + ":" + askingPort );
				}
			}else{
				// send DROPPEDACK; I do not have him in my neighbourhood anymore
				out.write( ProtocolConstants.DROPPEDACK );
				if( logger.isInfoEnabled() ){
					logger.info( "DROPPEDACK message sent to " + askingIp.getHostAddress() + ":" + askingPort );
				}
			}
			Utilities.writeIP( source.getIp(), out );
			out.writeInt( source.getPort() );
			out.flush();
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( "cannot confirm DROPPEDCONFIRM message ( " + e.getMessage() + " )" );
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}
}
