/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;

import ch.ethz.dcg.p2p.exceptions.DroppedHandlerException;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Handler for incoming DROPPED messages.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class DroppedHandler extends Task {

	private InetAddress droppedIp;
	private int droppedPort;


	/**
	 * Creates a new <code>DroppedHandler</code>.
	 * 
	 * @param socket
	 *            the TCP connection connection to the sender
	 * @param source
	 *            the owner of this task
	 */
	public DroppedHandler( Socket socket, Node source ){

		this.socket = socket;
		this.source = source;
	}


	/**
	 * Handles the DROPPED message:<br>
	 * confirm the DROPPED message using a DROPPEDCONFIRM and delete the sender
	 * if it responds with a DROPPEDACK message.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "DroppedHandler started" );
		}

		try{
			// get the DROPPED message
			getDropped();
			if( !source.hasNeighbour( droppedIp, droppedPort ) ){
				if( logger.isWarnEnabled() ){
					logger.warn( "DROPPED message from unknown node received; ignore it" );
				}
				return;
			}
			// confirm the message
			if( confirmDropped() ){
				source.removeNeighbour( droppedIp, droppedPort );
			}
		}catch( DroppedHandlerException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}finally{
			source.printNeighbours();
		}
	}


	/**
	 * Receives the DROPPED message.
	 * 
	 * @throws DroppedHandlerException
	 *             if a communication error occurred
	 */
	private void getDropped() throws DroppedHandlerException{

		try{
			DataInputStream in = new DataInputStream( socket.getInputStream() );
			droppedIp = Utilities.readIP( in );
			droppedPort = in.readInt();
			if( logger.isInfoEnabled() ){
				logger.info( "DROPPED message received from " + droppedIp.getHostAddress() + ":" + droppedPort );
			}
		}catch( IOException e ){
			throw new DroppedHandlerException( "cannot get DROPPED message from " + socket.getInetAddress().getHostAddress() );
		}
	}


	/**
	 * Confirms the DROPPED message.
	 * 
	 * @return true if the sender responds with a DROPPEDACK; false otherwise
	 * @throws DroppedHandlerException
	 *             if a communication error occurred
	 */
	private boolean confirmDropped() throws DroppedHandlerException{

		Socket s = null;
		try{
			// send DROPPEDCONFIRM message
			s = new Socket( droppedIp, droppedPort );
			DataOutputStream out = new DataOutputStream( s.getOutputStream() );
			out.write( ProtocolConstants.DROPPEDCONFIRM );
			Utilities.writeIP( source.getIp(), out );
			out.writeInt( source.getPort() );
			out.flush();
			if( logger.isInfoEnabled() ){
				logger.info( "DROPPEDCONFIRM message sent to " + droppedIp.getHostAddress() + ":" + droppedPort );
			}
			// read DROPPEDACK or DROPPEDNACK
			DataInputStream in = new DataInputStream( s.getInputStream() );
			byte answer = (byte)in.read();
			if( ( answer != ProtocolConstants.DROPPEDACK ) && ( answer != ProtocolConstants.DROPPEDNACK ) ){
				throw new DroppedHandlerException( "cannot confirm DROPPED message; invalid answer received; discard DROPPED message" );
			}
			InetAddress droppingIp = Utilities.readIP( in );
			int droppingPort = in.readInt();
			if( droppedIp.equals( droppingIp ) && ( droppedPort == droppingPort ) && ( answer == ProtocolConstants.DROPPEDACK ) ){
				if( logger.isInfoEnabled() ){
					logger.info( "DROPPEDACK message received from " + droppedIp.getHostAddress() + ":" + droppedPort );
				}
				return true;
			}else{
				if( logger.isInfoEnabled() ){
					logger.info( "DROPPEDNACK message received from " + droppedIp.getHostAddress() + ":" + droppedPort );
				}
				return false;
			}
		}catch( IOException e ){
			throw new DroppedHandlerException( "cannot confirm DROPPED message; will not remove " + droppedIp.getHostAddress() + ":" + droppedPort );
		}finally{
			try{
				s.close();
			}catch( Exception e ){
			}
		}
	}
}
