/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.util.ConcurrentModificationException;

import ch.ethz.dcg.p2p.exceptions.FloodHandlerException;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Handler for incoming FLOOD messages.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class FloodHandler extends Task {

	private Node source;
	private int floodId;
	private byte msgType;
	private int diagPort = 0;
	private InetAddress diagIp = null;
	private DataInputStream in = null;
	private DataOutputStream out = null;


	/**
	 * Creates a new <code>FloodHandler</code>.
	 * 
	 * @param socket
	 *            the TCP connection connection to the sender
	 * @param source
	 *            the owner of this task
	 */
	public FloodHandler( Socket socket, Node source ){

		this.socket = socket;
		this.source = source;
	}


	/**
	 * Handles the FLOOD message:<br>
	 * get the type of the message that is beeing flooded and act accordingly.
	 * Each FLOOD message contains a unique ID. If a FLOOD message with this ID
	 * has already been received this method will do nothing. Otherwise it
	 * forwards the message to all its neighbours.
	 */
	public void run(){

		// open input stream
		try{
			in = new DataInputStream( socket.getInputStream() );
		}catch( IOException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( "cannot start FloodHandler: cannot open InputStream" );
			}
			return;
		}

		if( logger.isDebugEnabled() ){
			logger.debug( "FloodHandler started" );
		}

		// dispatch type of the flood message
		try{
			// check if I have seen this flood message already
			floodId = in.readInt();
			if( source.seenFloodId( floodId ) ){
				if( logger.isWarnEnabled() ){
					logger.warn( "flood id already seen; do nothing and return" );
				}
				return;
			}
			source.addFloodId( floodId );

			// handle the flood message according to its type
			msgType = in.readByte();
			switch( msgType ){
			case ProtocolConstants.CONNECTIVITYTOKEN:
				if( logger.isInfoEnabled() ){
					logger.info( "CONNECTIVITYTOKEN FLOOD message received from " + socket.getInetAddress().getHostAddress() );
				}
				handleConnectivityToken();
				break;
			case ProtocolConstants.REPORTNEIGHBOURS:
				if( logger.isInfoEnabled() ){
					logger.info( "REPORTNEIGHBOURS FLOOD message received from " + socket.getInetAddress() );
				}
				handleReportNeighbours();
				break;
			default:
				if( logger.isErrorEnabled() ){
					logger.error( "invalid message type received " + msgType );
				}
			}
		}catch( FloodHandlerException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}


	/**
	 * Handles the CONNECTIVITYTOKEN message.
	 * 
	 * @throws FloodHandlerException
	 *             if a communication error occurred
	 */
	private void handleConnectivityToken() throws FloodHandlerException{

		// get ip and port of diagnostics node
		getIpAndPort();

		// send ack to the diagnostic node
		Socket diagSocket = null;
		try{
			diagSocket = new Socket( diagIp, diagPort );
			out = new DataOutputStream( diagSocket.getOutputStream() );
			out.writeByte( ProtocolConstants.CONNECTIVITYACK );
			Utilities.writeIP( source.getIp(), out );
			out.writeInt( source.getPort() );
			out.flush();
			if( logger.isInfoEnabled() ){
				logger.info( "CONNECTIVITYACK sent to " + diagIp.getHostAddress() + ":" + diagPort );
			}
		}catch( IOException e ){
			throw new FloodHandlerException( "cannot send CONNECTIVITYACK message to " + diagIp.getHostAddress() + ":" + diagPort + " ( " + e.getMessage() + " )" );
		}finally{
			try{
				diagSocket.close();
			}catch( Exception e ){
			}
		}

		// forward message to my neighbours
		forwardMessage( diagIp, diagPort );
	}


	/**
	 * Handles the REPORTNEIGHBOURS message.
	 * 
	 * @throws FloodHandlerException
	 *             if a communication error occurred
	 */
	private void handleReportNeighbours() throws FloodHandlerException{

		// get ip and port of diagnostics node
		getIpAndPort();

		// send neighbours message to diagnostics node
		Socket diagSocket = null;
		try{
			diagSocket = new Socket( diagIp, diagPort );
			out = new DataOutputStream( diagSocket.getOutputStream() );
			out.writeByte( ProtocolConstants.NEIGHBOURS );
			Utilities.writeIP( source.getIp(), out );
			out.writeInt( source.getPort() );
			out.writeInt( source.getNNeighbours() );
			for( Neighbour curN : source.getNeighboursCopy() ){
				Utilities.writeIP( curN.getIp(), out );
				out.writeInt( curN.getPort() );
			}
			out.flush();
			if( logger.isInfoEnabled() ){
				logger.info( "NEIGHBOURS message sent to " + diagSocket.getInetAddress().getHostAddress() + ":" + diagPort );
			}
		}catch( IOException e ){
			throw new FloodHandlerException( "cannot send NEIGHBOURS message to " + diagIp.getHostAddress() + ":" + diagPort + " : " + e.getMessage() );
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}catch( ConcurrentModificationException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}finally{
			try{
				diagSocket.close();
			}catch( Exception e ){
			}
		}

		// forward message to my neighbours
		forwardMessage( diagIp, diagPort );
	}


	/**
	 * Forwards the message to all neighbours.
	 */
	private void forwardMessage( InetAddress diagIp, int diagPort ){

		Socket fwdSocket = null;
		try{
			for( Neighbour curN : source.getNeighboursCopy() ){
				try{
					fwdSocket = new Socket( curN.getIp(), curN.getPort() );
					out = new DataOutputStream( fwdSocket.getOutputStream() );
					out.writeByte( ProtocolConstants.FLOOD );
					out.writeInt( floodId );
					out.writeByte( msgType );
					Utilities.writeIP( diagIp, out );
					out.writeInt( diagPort );
					out.flush();
					if( logger.isInfoEnabled() ){
						logger.info( "FLOOD message forwarded to " + curN.getIp().getHostAddress() + ":" + curN.getPort() );
					}
				}catch( IOException e ){
					if( logger.isWarnEnabled() ){
						logger.warn( "cannot forward FLOOD message to " + curN.getIp().getHostAddress() + ":" + curN.getPort() );
					}
				}catch( ConcurrentModificationException e ){
					if( logger.isErrorEnabled() ){
						logger.error( e.getMessage() );
					}
				}finally{
					try{
						fwdSocket.close();
					}catch( Exception e ){
					}
				}
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}


	/**
	 * Reads the next 8 bytes from the socket assuming they are the ip and port
	 * of the diagnostics node.
	 * 
	 * @throws FloodHandlerException
	 *             if a communication error occurred
	 */
	private void getIpAndPort() throws FloodHandlerException{

		try{
			diagIp = Utilities.readIP( in );
			diagPort = in.readInt();
		}catch( IOException e ){
			throw new FloodHandlerException( "cannot get ip and port of the Diagnostics node: " + e.getMessage() );
		}
	}

}
