/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;

import ch.ethz.dcg.p2p.exceptions.ForwardHandlerException;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.Utilities;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * Handler for incoming FORWARD messages.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class ForwardHandler extends Task {

	private InetAddress forwardIp;
	private int forwardPort;


	/**
	 * Creates a new <code>ByeHandler</code>.
	 * 
	 * @param socket
	 *            the TCP connection connection to the sender
	 * @param source
	 *            possible owner of this task
	 */
	public ForwardHandler( Socket socket, Node source ){

		this.socket = socket;
		this.source = source;
	}


	/**
	 * Handles the FORWARD message:<br>
	 * respond with an ACCEPTFORWARD or REJECTFORWARD message.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "ForwardHandler started" );
		}

		try{
			getForward();
			if( source.hasNeighbour( forwardIp, forwardPort ) ){
				sendAnswer( ProtocolConstants.REJECTFORWARD );
			}else if( source.hasSpareCapacity() || !allNeighboursAvailable() ){
				sendAnswer( ProtocolConstants.ACCEPTFORWARD );
				source.addNeighbour( new Neighbour( forwardIp, forwardPort ) );
			}else{
				sendAnswer( ProtocolConstants.REJECTFORWARD );
			}
		}catch( ForwardHandlerException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}finally{
			source.printNeighbours();
		}
	}


	/**
	 * Gets the FORWARD message.
	 * 
	 * @throws ForwardHandlerException
	 *             if a communication error occurred
	 */
	private void getForward() throws ForwardHandlerException{

		try{
			DataInputStream in = new DataInputStream( socket.getInputStream() );
			forwardIp = Utilities.readIP( in );
			forwardPort = in.readInt();
			if( logger.isInfoEnabled() ){
				logger.info( "FORWARD message received from " + socket.getInetAddress().getHostAddress() );
			}
		}catch( IOException e ){
			throw new ForwardHandlerException( "cannot get FORWARD message from " + socket.getInetAddress().getHostAddress() );
		}
	}


	/**
	 * Sends back the answer to the FORWARD request.
	 * 
	 * @param answer
	 *            either ACCEPTFORWARD or REJECTFORWARD
	 * @throws ForwardHandlerException
	 *             if a communication error occurred
	 */
	private void sendAnswer( byte answer ) throws ForwardHandlerException{

		try{
			DataOutputStream out = new DataOutputStream( socket.getOutputStream() );
			out.writeByte( answer );
			Utilities.writeIP( source.getIp(), out );
			out.writeInt( source.getPort() );
			out.flush();
			if( logger.isInfoEnabled() ){
				if( answer == ProtocolConstants.ACCEPTFORWARD ){
					logger.info( "ACCEPTFORWARD message sent to " + socket.getInetAddress().getHostAddress() );
				}else{
					logger.info( "REJECTFORWARD message sent to " + socket.getInetAddress().getHostAddress() );
				}
			}
		}catch( IOException e ){
			throw new ForwardHandlerException( "cannot send answer to " + socket.getInetAddress().getHostAddress() );
		}
	}

}
