/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Collections;

import ch.ethz.dcg.p2p.exceptions.HelloHandlerException;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.exceptions.WrongProtocolMessageException;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Handler for incoming HELLO messages.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class HelloHandler extends Task {

	private InetAddress helloIp;
	private int helloPort;


	/**
	 * Creates a new <code>HelloHandler</code>.
	 * 
	 * @param socket
	 *            the TCP connection connection to the sender
	 * @param source
	 *            the owner of this task
	 */
	public HelloHandler( Socket socket, Node source ){

		this.socket = socket;
		this.source = source;
	}


	/**
	 * Handles the HELLO message:<br>
	 * respond with an ACCEPT or REDIRECT message.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "HelloHandler started" );
		}

		// get the HELLO message
		// check if node has spare capacity
		// -> if so send the nodes ip and port back in an ACCEPT message and add
		// the requesting node to the nodes neighbours
		// -> if not find a neighbour that has spare capacity
		// --> if none of the neighbours has spare capacity, remove the worst
		// one and accept the new node as direct neighbour
		// notify the dropped neighbour

		try{
			getHello();
			if( source.hasNeighbour( helloIp, helloPort ) ){
				sendAccept();
				if( logger.isDebugEnabled() ){
					logger.debug( "peer " + helloIp.getHostAddress() + ":" + helloPort + " is already a neighbour; ACCEPT sent" );
				}
			}
			if( !allNeighboursAvailable() || source.hasSpareCapacity() ){
				sendAccept();
				source.addNeighbour( new Neighbour( helloIp, helloPort ) );
			}else{
				Neighbour accepting = findAcceptingNeighbour();
				if( accepting == null ){
					Neighbour worst = source.removeWorstNeighbour();
					sendDropped( worst );
					sendAccept();
					source.addNeighbour( new Neighbour( helloIp, helloPort ) );
				}else{
					sendRedirect( accepting );
				}
			}
		}catch( HelloHandlerException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}finally{
			source.printNeighbours();
		}
	}


	/**
	 * Gets the HELLO message.
	 * 
	 * @throws HelloHandlerException
	 *             if a communication error occurred
	 */
	private void getHello() throws HelloHandlerException{

		try{
			DataInputStream inStream = new DataInputStream( socket.getInputStream() );
			helloIp = Utilities.readIP( inStream );
			helloPort = inStream.readInt();
			if( logger.isInfoEnabled() ){
				logger.info( "HELLO message received from " + socket.getInetAddress().getHostAddress() + ":" + helloPort );
			}
		}catch( IOException e ){
			throw new HelloHandlerException( "cannot get HELLO message from " + socket.getInetAddress().getHostAddress() );
		}
	}


	/**
	 * Replies to the node that has sent the HELLO message with an ACCEPT message.
	 * He will become a direct neighbour of me.
	 * 
	 * @throws HelloHandlerException
	 *             if a communication error occurred
	 */
	private void sendAccept() throws HelloHandlerException{

		try{
			DataOutputStream out = new DataOutputStream( socket.getOutputStream() );
			out.write( ProtocolConstants.ACCEPT );
			Utilities.writeIP( source.getIp(), out );
			out.writeInt( source.getPort() );
			out.writeInt( source.getPacketSize() );
			out.writeInt( source.getBitrate() );
			out.flush();
			if( logger.isInfoEnabled() ){
				logger.info( "ACCEPT message sent to " + socket.getInetAddress().getHostAddress() );
			}
			source.write2gui( helloIp.getHostAddress() + ":" + helloPort + " accepted as new neighbour" );
		}catch( IOException e ){
			throw new HelloHandlerException( "cannot send ACCEPT message back to " + socket.getInetAddress().getHostAddress() );
		}
	}


	/**
	 * Finds a neighbour of the source that is willing to accept another
	 * neighbour.
	 * 
	 * @return first neighbour that is willing to accept another neighbour or
	 *         null if none exists
	 */
	private Neighbour findAcceptingNeighbour(){

		boolean accepting = false;

		// tmp vector for random neighbour selection
		ArrayList<Neighbour> tmpN;
		try{
			tmpN = source.getNeighboursCopy();
			Collections.shuffle( tmpN );
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
			return null;
		}

		// ask all neighbour if they have spare capacity (in random order)
		for( Neighbour curN : tmpN ){
			try{
				accepting = isWillingToAccept( curN );
				if( accepting ){
					if( logger.isDebugEnabled() ){
						logger.debug( "node " + curN.getIp().getHostAddress() + " is willing to accept " + helloIp.getHostAddress() + ":" + helloPort );
					}
					return curN;
				}
			}catch( WrongProtocolMessageException e ){
				if( logger.isErrorEnabled() ){
					logger.error( curN.getIp() + " did not behave as expected: " + e.getMessage() );
				}
			}catch( HelloHandlerException e ){
				if( logger.isErrorEnabled() ){
					logger.error( curN.getIp() + " did not behave as expected: " + e.getMessage() );
				}
			}
		}
		if( logger.isDebugEnabled() ){
			logger.debug( "no accepting neighbour found" );
		}
		return null;
	}


	/**
	 * Checks if neighbour is willing to accept another neighbour.
	 * 
	 * @param newN
	 *            the neighbour that we want to test.
	 * @return true if he is willing to accept another neighbour; false
	 *         otherwise
	 * @throws WrongProtocolMessageException
	 *             if an unexpected answer was received
	 * @throws HelloHandlerException
	 *             if a communication error occurred
	 */
	private boolean isWillingToAccept( Neighbour newN ) throws WrongProtocolMessageException, HelloHandlerException{

		Socket tmpSocket = null;
		DataOutputStream tmpOut = null;
		DataInputStream tmpIn = null;

		try{
			// open connection to Neighbour n
			// send FORWARD
			// receive ACCEPT or REJECT
			tmpSocket = new Socket( newN.getIp(), newN.getPort() );
			tmpOut = new DataOutputStream( tmpSocket.getOutputStream() );
			tmpIn = new DataInputStream( tmpSocket.getInputStream() );

			tmpOut.writeByte( ProtocolConstants.FORWARD );
			Utilities.writeIP( helloIp, tmpOut );
			tmpOut.writeInt( helloPort );
			tmpOut.flush();
			if( logger.isInfoEnabled() ){
				logger.info( "FORWARD message sent to " + newN.getIp().getHostAddress() + ":" + newN.getPort() );
			}

			byte answer = tmpIn.readByte();
			if( answer == ProtocolConstants.ACCEPTFORWARD ){
				if( Utilities.readIP( tmpIn ).equals( newN.getIp() ) && ( tmpIn.readInt() == newN.getPort() ) ){
					if( logger.isInfoEnabled() ){
						logger.info( "ACCEPTFORWARD message received from " + newN.getIp().getHostAddress() + ":" + newN.getPort() );
					}
					return true;
				}else{
					throw new WrongProtocolMessageException( "ACCEPTFORWARD message was received from an unexpected host" );
				}
			}else if( answer == ProtocolConstants.REJECTFORWARD ){
				if( logger.isInfoEnabled() ){
					logger.info( "REJECTFORWARD message received from " + newN.getIp().getHostAddress() + ":" + newN.getPort() );
				}
				return false;
			}else{
				throw new WrongProtocolMessageException( "wrong message type received: " + answer );
			}
		}catch( IOException e ){
			throw new HelloHandlerException( "cannot check if " + newN.getIp().getHostAddress() + ":" + newN.getPort() + " is willing to accept another neighbour" );
		}finally{
			try{
				tmpSocket.close();
			}catch( Exception e ){
			}
		}
	}


	/**
	 * Sends a REDIRECT message to the node that has sent the HELLO message. The
	 * IP address and the port are those of the node that is willing to accept
	 * him as a neighbour.
	 * 
	 * @param accepting
	 *            Neighbour that is willing to accept the node that has sent the
	 *            HELLO message.
	 * @throws HelloHandlerException if a communication error occurred
	 */
	private void sendRedirect( Neighbour accepting ) throws HelloHandlerException{

		try{
			DataOutputStream out = new DataOutputStream( socket.getOutputStream() );
			out.write( ProtocolConstants.REDIRECT );
			Utilities.writeIP( accepting.getIp(), out );
			out.writeInt( accepting.getPort() );
			out.writeInt( source.getPacketSize() );
			out.writeInt( source.getBitrate() );
			out.flush();
			if( logger.isInfoEnabled() ){
				logger.info( "REDIRECT message sent to " + helloIp.getHostAddress() + ":" + helloPort );
			}
		}catch( IOException e ){
			throw new HelloHandlerException( "cannot send REDIRECT message back to " + socket.getInetAddress().getHostAddress() );
		}
	}


	/**
	 * Notifies a neighbour that he has been dropped.
	 * 
	 * @param dropped
	 *            neighbour that has been dropped
	 * @throws HelloHandlerException if a communication error occurred
	 */
	private void sendDropped( Neighbour dropped ) throws HelloHandlerException{

		Socket tmpSocket = null;
		DataOutputStream tmpOut = null;

		try{
			tmpSocket = new Socket( dropped.getIp(), dropped.getPort() );
			tmpOut = new DataOutputStream( tmpSocket.getOutputStream() );
			tmpOut.write( ProtocolConstants.DROPPED );
			Utilities.writeIP( source.getIp(), tmpOut );
			tmpOut.writeInt( source.getPort() );
			tmpOut.flush();
			if( logger.isInfoEnabled() ){
				logger.info( "DROPPED message sent to " + dropped.getIp().getHostAddress() + ":" + dropped.getPort() );
			}
		}catch( IOException e ){
			throw new HelloHandlerException( "cannot send DROPPED message to " + dropped.getIp().getHostAddress() + ":" + dropped.getPort() );
		}finally{
			try{
				tmpSocket.close();
			}catch( Exception e ){
			}
		}
	}

}
