/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.Socket;
import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.dcg.p2p.exceptions.WrongProtocolMessageException;
import ch.ethz.dcg.p2p.peers.NEPNeighbour;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.NetworkEntryPoint;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.Utilities;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * The NEPJoinByeHandler is used at the NetworkEntryPoint to handle incoming
 * connections from new peers that want to join the network. It also starts a
 * ByeHandler if a node is leaving the network or reports the statistics if
 * requested.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * 
 */
public class NEPJoinByeHandler extends Task {

	private static final Log logger = LogFactory.getLog( NEPJoinByeHandler.class );

	private NetworkEntryPoint nep;
	private DataOutputStream outStream;


	public NEPJoinByeHandler( Socket socket, NetworkEntryPoint nep ){

		this.socket = socket;
		this.nep = nep;
	}


	/**
	 * Handles incoming JOIN requests. Peers can re-join. Sources cannot join a
	 * level which already has a source but they can re-join their own level.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "NEPJoinByeHandler started" );
		}

		NEPNeighbour newNode;
		try{

			// receive the request from the new node
			// find his initial neighbours
			// send a INITNEIGHBOURS message back including his initial
			// neighbours

			DataInputStream inStream = new DataInputStream( socket.getInputStream() );
			outStream = new DataOutputStream( socket.getOutputStream() );

			newNode = getRequest( inStream );
			if( newNode == null ){
				// it was not a JOIN message
				return;
			}
			// add new NEPNeighbour to NetworkEntryPoint
			if( newNode.getNodeType() == ProtocolConstants.SOURCE ){
				// create new level with this source if non-existent
				if( nep.containsSource( newNode.getLevel() ) ){
					// duplicate source
					NEPNeighbour existing = nep.containsNEPNeighbourAtLevel( newNode.getIp(), newNode.getPort(), newNode.getLevel() );
					if( existing == null || existing.getNodeType() == ProtocolConstants.PEER ){
						sendJoinError( outStream, ProtocolConstants.DUPLICATESOURCE );
						nep.write2gui( "duplicate join attempt by " + newNode.getIp().getHostAddress() + ":" + newNode.getPort() + " prohibited" );
						return;
					}
				}else{
					// insert source as source in hashtable at nep
					nep.insertSource( newNode );
					// create an appropriate level for the new source
					nep.createLevel( newNode.getLevel() );
				}
				// new level was created; new source will be inserted in
				// this level
			}else if( newNode.getNodeType() == ProtocolConstants.PEER ){
				// check if there is an appropriate level for the Peer
				if( newNode.getSpeed() < nep.getMinSpeed() * nep.getCapacityRequired() ){
					// peer too slow for all levels; he cannot join
					sendJoinError( outStream, ProtocolConstants.PEERTOOSLOW );
					nep.write2gui( "peer " + newNode.getIp().getHostAddress() + ":" + newNode.getPort() + " is too slow to join (reported " + newNode.getSpeed() + "kbps)" );
					return;
				}
			}

			Vector<Neighbour> initNeighbours = nep.getInitNeighbours( newNode );
			if( initNeighbours.isEmpty() && ( newNode.getNodeType() == ProtocolConstants.PEER ) ){
				// there is a ghost level without source; remove it and the ghost level as well
				sendJoinError( outStream, ProtocolConstants.PEERTOOSLOW );
				nep.removeLevel( newNode.getLevel() );
				return;
			}
			sendInitialNeighboursResponse( outStream, initNeighbours );
			nep.insertNEPNeighbour( newNode );
			if( newNode.getNodeType() == ProtocolConstants.PEER ){
				nep.write2gui( newNode.getIp().getHostAddress() + ":" + newNode.getPort() + " joined with speed " + newNode.getSpeed() );
			}else if( newNode.getNodeType() == ProtocolConstants.SOURCE ){
				nep.write2gui( newNode.getIp().getHostAddress() + ":" + newNode.getPort() + "(Source) joined and created new level" );
			}

		}catch( WrongProtocolMessageException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}catch( Exception e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}finally{
			nep.printNeighbours();
		}
	}


	/**
	 * Parses the input received from the joining peer.
	 * 
	 * @param inStream
	 * @return NEPNeighbour if it is a JOIN message; null otherwise
	 * @throws WrongProtocolMessageException
	 *             if an unexpected message was received
	 * @throws IOException
	 *             if a communication error occurred
	 */
	private NEPNeighbour getRequest( DataInputStream inStream ) throws WrongProtocolMessageException, IOException{

		NEPNeighbour newNode;
		int port, speed;
		String version;

		byte type = (byte)inStream.readByte();
		if( type != ProtocolConstants.JOIN ){
			if( type == ProtocolConstants.BYE ){
				// BYE
				new Thread( new ByeHandler( socket, null, nep ), "ByeHandler" ).start();
				return null;
			}else if( type == ProtocolConstants.REPORTSTATISTIC ){
				// REPORTSTATISTIC
				new Thread( new ReportStatisticHandler( socket, null, nep ), "ReportStatisticHandler" ).start();
				return null;
			}else{
				throw new WrongProtocolMessageException( "wrong message type received: " + type );
			}
		}
		port = inStream.readInt();
		speed = inStream.readInt();
		byte nodeType = inStream.readByte();
		if( ( nodeType == ProtocolConstants.SOURCE ) || ( nodeType == ProtocolConstants.PEER ) ){
			newNode = new NEPNeighbour( socket.getInetAddress(), port, nodeType, speed );
		}else{
			throw new WrongProtocolMessageException( "invalid node type received: " + nodeType );
		}
		if( nodeType == ProtocolConstants.SOURCE ){
			nep.setStreamType( inStream.readByte() );
		}
		byte[] v = new byte[255];
		int len = inStream.read( v );
		version = new String( v, 0, len );
		if( !version.equals( ProtocolConstants.VERSION ) ){
			sendVersionMismatch();
			if( logger.isInfoEnabled() ){
				logger.info( "JOIN message with invalid version received" );
			}
			return null;
		}

		if( logger.isInfoEnabled() ){
			logger.info( "JOIN message received from " + socket.getInetAddress().getHostAddress() + ":" + port );
		}

		return newNode;
	}


	/**
	 * Sends the message containing all initial neighbours to the joining peer.
	 * 
	 * @param outStream
	 *            the TCP connection to the joining peer
	 * @param initNeighs
	 *            the initial neighbours to send
	 */
	private void sendInitialNeighboursResponse( DataOutputStream outStream, Vector<Neighbour> initNeighs ) throws IOException{

		outStream.writeByte( ProtocolConstants.INITNEIGHBOURS );
		outStream.writeInt( nep.getMinNeighs() );
		outStream.writeInt( nep.getMaxNeighs() );
		outStream.writeByte( nep.getStreamType() );
		outStream.writeInt( initNeighs.size() );
		for( Neighbour curN : initNeighs ){
			Utilities.writeIP( curN.getIp(), outStream );
			outStream.writeInt( curN.getPort() );
		}
		outStream.flush();
		if( logger.isInfoEnabled() ){
			logger.info( "INITNEIGHBOURS message sent to " + socket.getInetAddress().getHostAddress() );
		}
	}


	/**
	 * Sends a DUPLICATESOURCE message or a PEERTOOSLOW message back to the peer
	 * that wants to join.
	 * 
	 * @param outStream
	 *            the TCP connection to the joining peer
	 * @param error
	 *            the type of error that should be sent
	 * @throws IOException
	 *             if a communication error occurred
	 */
	private void sendJoinError( DataOutputStream outStream, byte error ) throws IOException{

		outStream.writeByte( error );
		outStream.flush();
		if( logger.isInfoEnabled() ){
			switch( error ){
			case ProtocolConstants.DUPLICATESOURCE:
				logger.info( "DUPLICATESOURCE message sent to " + socket.getInetAddress().getHostAddress() );
				break;
			case ProtocolConstants.PEERTOOSLOW:
				logger.info( "PEERTOOSLOW message sent to " + socket.getInetAddress().getHostAddress() );
				break;
			default:
				logger.warn( "unknown message sent to " + socket.getInetAddress().getHostAddress() );
				break;
			}
		}
	}


	/**
	 * Sends a VERSIONMISMATCH message to the joining peer.
	 * 
	 * @throws IOException
	 *             if a communication error occurred
	 */
	private void sendVersionMismatch() throws IOException{

		outStream.writeByte( ProtocolConstants.VERSIONMISMATCH );
		outStream.write( ProtocolConstants.VERSION.getBytes() );
		outStream.flush();
		if( logger.isInfoEnabled() ){
			logger.info( "VERSIONMISMATCH message sent to " + socket.getInetAddress().getHostAddress() );
		}
	}
}
