/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.net.InetAddress;
import java.net.UnknownHostException;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Handler for incoming NOTIFICATION messages.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class NotificationHandler extends Task {

	private long packetId;
	private InetAddress holderIp;
	private int holderPort;
	private byte[] udpData;


	/**
	 * Creates a new <code>NotificationHandler</code>.
	 * 
	 * @param udpData
	 *            the received UDP data
	 * @param source
	 *            the owner of this task
	 */
	public NotificationHandler( byte[] udpData, Node source ){

		this.udpData = udpData;
		this.source = source;
	}


	/**
	 * Handles the NOTIFICATION message:<br>
	 * add it to the owners notifications if it is valid
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "NotificationHandler started" );
		}

		try{
			packetId = Utilities.getLong( udpData, 1, 8 );
			if( packetId < 0 ){
				if( logger.isWarnEnabled() ){
					logger.warn( "NOTIFICATION for negative packet id received; dropping it" );
				}
				source.write2gui( "NOTIFICATION with negative id received" );
				Node.statistic.incInvalidNotificationReceived();
				return;
			}
			holderIp = InetAddress.getByAddress( Utilities.subArray( udpData, 9, 4 ) );
			holderPort = Utilities.getInt( udpData, 13, 4 );
			if( logger.isInfoEnabled() ){
				logger.info( "NOTIFICATION for packet " + packetId + " received from " + holderIp.getHostAddress() + ":" + holderPort );
			}
			source.addNotification( packetId, holderIp, holderPort );
		}catch( UnknownHostException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}
}
