/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.io.IOException;
import java.net.InetAddress;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.exceptions.P2PPacketException;
import ch.ethz.dcg.p2p.packets.P2PPacket;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Handler for incoming P2PPACKET messages.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class PacketHandler extends Task {

	private long packetId;
	private int dataLength;
	private byte[] data;
	private InetAddress senderIp;
	private int senderPort;
	private byte[] udpData;
	private long receptionTime;


	/**
	 * Creates a new <code>PacketHandler</code>.
	 * 
	 * @param udpData
	 *            the received UDP data
	 * @param source
	 *            the owner of this task
	 */
	public PacketHandler( byte[] udpData, Node source ){

		this.udpData = udpData;
		this.source = source;
	}


	/**
	 * Handles the P2PPACKET message:<br>
	 * add it to the owners packets if it is valid
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "PacketHandler started" );
		}

		try{
			receptionTime = System.currentTimeMillis();
			senderIp = InetAddress.getByAddress( Utilities.subArray( udpData, 1, 4 ) );
			senderPort = Utilities.getInt( udpData, 5, 4 );
			if( !source.hasNeighbour( senderIp, senderPort ) ){
				// do not know sending peer; discard packet
				if( logger.isWarnEnabled() ){
					logger.warn( "P2PPACKET received from unknown neighbour " + senderIp.getHostAddress() + ":" + senderPort );
				}
				source.write2gui( "P2PPACKET received from unknown neighbour " + senderIp.getHostAddress() + ":" + senderPort );
				Node.statistic.incUnknownPacketsReceived();
				return;
			}
			packetId = Utilities.getLong( udpData, 9, 8 );
			if( logger.isInfoEnabled() ){
				logger.info( "P2PPACKET " + packetId + " received from " + senderIp.getHostAddress() + ":" + senderPort );
			}
			if( packetId < 0 ){
				if( logger.isWarnEnabled() ){
					logger.warn( "P2PPACKET with negative id received; dropping it" );
				}
				source.write2gui( "P2PPACKET with negative id received" );
				Node.statistic.incInvalidPacketReceived();
				return;
			}
			if( source.hasPacket( packetId ) ){
				if( logger.isWarnEnabled() ){
					logger.warn( "duplicate P2PPACKET " + packetId + " received; dropping it" );
				}
				Node.statistic.incDuplicatePacketsReceived();
				return;
			}
			dataLength = Utilities.getInt( udpData, 17, 4 );
			if( dataLength < 0 ){
				if( logger.isWarnEnabled() ){
					logger.warn( "P2PPACKET with negative size received; dropping it" );
				}
				source.write2gui( "P2PPACKET with negative size received" );
				Node.statistic.incInvalidPacketReceived();
				return;
			}
			if( dataLength > 65500 ){
				if( logger.isWarnEnabled() ){
					logger.warn( "P2PPACKET with oversize received; dropping it" );
				}
				source.write2gui( "P2PPACKET with oversize received" );
				Node.statistic.incInvalidPacketReceived();
				return;
			}
			data = Utilities.subArray( udpData, 21, dataLength );
			P2PPacket packet = new P2PPacket( packetId, dataLength, source.getNeighbour( senderIp, senderPort ) );
			packet.setData( data );
			source.addPacket( packet, receptionTime );
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( "cannot create packet: " + e.getMessage() );
			}
		}catch( P2PPacketException e ){
			if( logger.isErrorEnabled() ){
				logger.error( "cannot create packet: " + e.getMessage() );
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}
}
