/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.io.IOException;
import java.net.InetAddress;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Handler for incoming PACKETNOTAVAILABLE messages.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class PacketNotAvailableHandler extends Task {

	private long packetId;
	private InetAddress informingIp;
	private int informingPort;
	private byte[] udpData;


	/**
	 * Creates a new <code>PacketNotAvailableHandler</code>.
	 * 
	 * @param udpData
	 *            the received UDP data
	 * @param source
	 *            the owner of this task
	 */
	public PacketNotAvailableHandler( byte[] udpData, Node source ){

		this.udpData = udpData;
		this.source = source;
	}


	/**
	 * Handles the PACKETNOTAVAILABLE message:<br>
	 * reschedule the packet for re-request from another neighbour
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.fatal( "PacketNotAvailableHandler started" );
		}

		try{
			packetId = Utilities.getLong( udpData, 1, 8 );
			if( packetId < 0 ){
				if( logger.isWarnEnabled() ){
					logger.warn( "PACKETNOTAVAILABLE message for negative packet id received; dropping it" );
				}
				source.write2gui( "PACKETNOTAVAILABLE mesage with negative id received" );
				return;
			}
			informingIp = InetAddress.getByAddress( Utilities.subArray( udpData, 9, 4 ) );
			informingPort = Utilities.getInt( udpData, 13, 4 );
			if( !source.hasNeighbour( informingIp, informingPort ) ){
				if( logger.isWarnEnabled() ){
					logger.warn( "PACKETNOTAVAILABLE message from unknown neighbour received; dropping it" );
				}
				source.write2gui( "PACKETNOTAVAILABLE message received from " + informingIp.getHostAddress() + ":" + informingPort + " which is NOT my neighbour" );
				return;
			}
			if( !source.hasRequestedPacket( packetId ) ){
				if( logger.isWarnEnabled() ){
					logger.warn( "PACKETNOTAVAILABLE message for not requested packet received; dropping it" );
				}
				source.write2gui( "PACKETNOTAVAILABLE message for not requested packet received" );
				return;
			}
			if( logger.isInfoEnabled() ){
				logger.info( "PACKETNOTAVAILABLE message for packet " + packetId + " received from " + informingIp.getHostAddress() + ":" + informingPort );
			}
			source.moveToNotifications( packetId );
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( "cannot receive PACKETNOTAVAILABLE message ( " + e.getMessage() + " )" );
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}
}
