/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.io.DataOutputStream;
import java.io.IOException;
import java.net.Socket;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.NetworkEntryPoint;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.peers.Peer;
import ch.ethz.dcg.p2p.tasks.PingTask;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.DummyNode;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Handler for incoming REPORTSTATISTIC messages.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class ReportStatisticHandler extends Task {

	private NetworkEntryPoint nep = null;


	/**
	 * Creates a new <code>ReportStatisticHandler</code>.<br>
	 * This task can either be started by a <code>Node</code> or by a
	 * <code>NetworkEntryPoint</code>. If it is started by a
	 * <code>Node</code> then the nep argument is null and vice versa.
	 * 
	 * @param socket
	 *            the TCP connection connection to the sender
	 * @param source
	 *            possible owner of this task
	 * @param nep
	 *            possible owner of this task
	 */
	public ReportStatisticHandler( Socket socket, Node source, NetworkEntryPoint nep ){

		this.socket = socket;
		this.source = source;
		this.nep = nep;
	}


	/**
	 * Handles the REPORTSTATISTIC message:<br>
	 * respond with a NEPSTATISTIC or PEERSTATISTIC message depending on who
	 * started this handler.
	 */
	public void run(){

		if( source == null && nep == null ){
			if( logger.isErrorEnabled() ){
				logger.error( "cannot start ReportStatisticHandler: nep and source pointers are both null" );
			}
			return;
		}

		if( logger.isDebugEnabled() ){
			logger.debug( "ReportStatisticHandler started" );
		}

		if( source != null ){
			// ReportStatisticHandler started by a Node
			try{
				DataOutputStream out = new DataOutputStream( socket.getOutputStream() );
				// TYPE
				out.writeByte( ProtocolConstants.PEERSTATISTIC );
				// nNeighs
				out.writeInt( source.getNNeighbours() );
				// packetSize
				out.writeInt( source.getPacketSize() );
				// averageRTT
				if( source instanceof Peer ){
					int count = 0;
					int rttTotal = 0;
					for( Neighbour curN : source.getNeighboursCopy() ){
						if( curN.getRTT() != -1 ){
							rttTotal += curN.getRTT();
							count++;
						}
					}
					if( count > 0 ){
						out.writeFloat( rttTotal / count );
					}else{
						out.writeFloat( -1 );
					}
				}else{
					out.writeFloat( -1 );
				}
			}catch( IOException e ){
				if( logger.isErrorEnabled() ){
					logger.error( e.getMessage() );
				}
			}catch( NodeNotConfiguredException e ){
				if( logger.isErrorEnabled() ){
					logger.error( e.getMessage() );
				}
			}

		}else if( nep != null ){
			// ReportStatisticHandler started by NetworkEntryPoint
			try{
				DataOutputStream out = new DataOutputStream( socket.getOutputStream() );
				// TYPE
				out.writeByte( ProtocolConstants.NEPSTATISTIC );
				// nInitNeighs
				out.writeInt( nep.getNInitNeighbours() );
				// maxNeighs
				out.writeInt( nep.getMaxNeighs() );
				// minNeighs
				out.writeInt( nep.getMinNeighs() );
				// nLevels
				Hashtable<Integer, Vector<DummyNode>> levels = nep.copyLevels();
				out.writeInt( levels.size() );
				// for each level
				Iterator<Integer> it = levels.keySet().iterator();
				while( it.hasNext() ){
					int levelSpeed = it.next();
					Vector<DummyNode> peers = levels.get( levelSpeed );
					removeZombies( peers, 2000, 2 );
					int nPeers = peers.size();
					// levelSpeed
					out.writeInt( levelSpeed );
					// nPeers
					out.writeInt( nPeers );
					// for each peer in this level
					for( int i = 0; i < nPeers; i++ ){
						DummyNode curN = peers.elementAt( i );
						// ip
						Utilities.writeIP( curN.getIp(), out );
						// port
						out.writeInt( curN.getPort() );
					}
					// avgLevelSpeed
					out.writeFloat( nep.getAvgSpeedForLevel( levelSpeed ) );
				}
				// avgTotalSpeed
				out.writeFloat( nep.getAvgOverallSpeed() );
			}catch( IOException e ){
				if( logger.isErrorEnabled() ){
					logger.error( e.getMessage() );
				}
			}
		}
	}


	/**
	 * Do not report peers that are no longer available. This method pings all
	 * neighbours and removes the ones that are not responding.
	 * 
	 * @param peers
	 *            the neighbours to ping
	 * @param timeout
	 *            ping timeout
	 * @param retries
	 *            the number of tries to reach a neighbour before he gets
	 *            removed
	 */
	private void removeZombies( Vector<DummyNode> peers, int timeout, int retries ){

		int i = 0;
		DummyNode curP;
		PingTask ping;
		boolean success;
		while( i < peers.size() ){
			curP = peers.elementAt( i );
			ping = new PingTask( curP.getIp(), curP.getPort(), timeout );
			success = false;
			for( int retry = 0; retry < retries && !success; retry++ ){
				success = ping.sendPINGreceivePONG();
			}
			if( !success ){
				peers.remove( i );
			}else{
				i++;
			}
		}
	}
}
