/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.handlers;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.tasks.PacketTask;
import ch.ethz.dcg.p2p.tasks.Task;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.UDPSocketManager;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Handler for incoming HELLO messages.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class RequestHandler extends Task {

	private long packetId;
	private InetAddress requestingIp;
	private int requestingPort;
	private byte[] udpData;
	private UDPSocketManager socketManager;


	/**
	 * Creates a new <code>RequestHandler</code>.
	 * 
	 * @param udpData
	 *            the received UDP data
	 * @param source
	 *            the owner of this task
	 */
	public RequestHandler( byte[] udpData, Node source ){

		this.udpData = udpData;
		this.source = source;
		this.socketManager = source.getSocketManager();
	}


	/**
	 * Handles the REQUEST message:<br>
	 * respond with a P2PPACKET message if it available or with a
	 * PACKETNOTAVAILABLE otherwise.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "RequestHandler started" );
		}

		try{
			packetId = Utilities.getLong( udpData, 1, 8 );
			if( packetId < 0 ){
				if( logger.isWarnEnabled() ){
					logger.warn( "REQUEST for negative packet id received; dropping it" );
				}
				source.write2gui( "REQUEST with negative id received" );
				Node.statistic.incInvalidPacketReceived();
				return;
			}
			requestingIp = InetAddress.getByAddress( Utilities.subArray( udpData, 9, 4 ) );
			requestingPort = Utilities.getInt( udpData, 13, 4 );
			target = source.getNeighbour( requestingIp, requestingPort );
			if( logger.isInfoEnabled() ){
				logger.info( "REQUEST for packet " + packetId + " received from " + requestingIp.getHostAddress() + ":" + requestingPort );
			}
			if( target == null ){
				if( logger.isWarnEnabled() ){
					logger.warn( "node " + requestingIp.getHostAddress() + ":" + requestingPort + " is not among my neighbours; will not respond to packt request" );
				}
				source.write2gui( "REQUEST from " + requestingIp.getHostAddress() + ":" + requestingPort + " received which is NOT my neighbour" );
				Node.statistic.incUnknownRequestsReceived();
				return;
			}
			Node.statistic.incRequestsReceivedFrom( requestingIp.getHostAddress() + ":" + requestingPort );
			if( !source.hasPacket( packetId ) ){
				if( logger.isWarnEnabled() ){
					logger.warn( "do not have P2PPACKET " + packetId + " for which I have received a REQUEST from " + requestingIp.getHostAddress() + ":" + requestingPort );
				}
				source.write2gui( "do not have P2PPACKET " + packetId + " anymore for which I have just received a REQUEST" );
				Node.statistic.incPacketNotAvailable();
				sendPacketNotAvailable();
				return;
			}
			new Thread( new PacketTask( source, target, packetId ), "PacketTask" ).start();
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( "cannot get the REQUEST message: (id=" + packetId + "),(host=" + requestingIp.getHostAddress() + ":" + requestingPort + ")" );
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}


	/**
	 * Sends a PACKETNOTAVAILABLE message to the requesting peer to inform him
	 * that this packet is no longe available here.
	 */
	private void sendPacketNotAvailable(){

		try{
			byte[] buffer = new byte[17];
			buffer[0] = ProtocolConstants.PACKETNOTAVAILABLE;
			System.arraycopy( Utilities.longToByteArray( packetId ), 0, buffer, 1, 8 );
			System.arraycopy( source.getIp().getAddress(), 0, buffer, 9, 4 );
			System.arraycopy( Utilities.intToByteArray( source.getPort() ), 0, buffer, 13, 4 );
			DatagramPacket packet = new DatagramPacket( buffer, buffer.length, requestingIp, requestingPort );
			socketManager.sendPacket( packet );
			if( logger.isInfoEnabled() ){
				logger.info( "PACKETNOTAVAILABLE message for packet " + packetId + " sent to " + requestingIp.getHostAddress() + ":" + requestingPort );
			}
		}catch( Exception e ){
			if( logger.isWarnEnabled() ){
				logger.warn( "cannot send PACKETNOTAVAILABLEMESSAGE to " + requestingIp.getHostAddress() + ":" + requestingPort );
			}
		}
	}
}
