/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.main;

import java.net.UnknownHostException;

import ch.ethz.dcg.p2p.adapters.input.InputAdapter;
import ch.ethz.dcg.p2p.adapters.input.ShoutCastAdapter;
import ch.ethz.dcg.p2p.adapters.input.VlcHttpAdapter;
import ch.ethz.dcg.p2p.exceptions.DispatcherNotRunningException;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.exceptions.UDPSocketManagerException;
import ch.ethz.dcg.p2p.peers.NetworkEntryPoint;
import ch.ethz.dcg.p2p.peers.Peer;
import ch.ethz.dcg.p2p.peers.Source;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * Operate the Swistry application from the console instead of the gui. Is is
 * possible to start all three components: a NetworkEntryPoint, a Peer or a
 * Source.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class SwistryConsole {

	Peer p;
	Source s;
	NetworkEntryPoint nep;


	public static void main( String[] args ){

		SwistryConsole tester = new SwistryConsole();
		InputAdapter inAdapter = null;

		if( args.length < 1 ){
			printUsage();
			return;
		}

		try{
			if( args[0].equals( "peer" ) ){
				// start a peer
				if( args.length < 8 ){
					printUsagePeer();
					return;
				}
				// get arguments
				String ip = args[1];
				int port = Integer.parseInt( args[2] );
				int inSpeed = Integer.parseInt( args[3] );
				int outSpeed = Integer.parseInt( args[4] );
				String nepAddress = args[5];
				int dePacketizerThreshold = Integer.parseInt( args[6] );
				int localMediaPort = Integer.parseInt( args[7] );
				// init peer
				tester.p = new Peer( ip, port, inSpeed, outSpeed, nepAddress, dePacketizerThreshold );
				tester.p.setMediaPort( localMediaPort );
				tester.p.startDispatchers();
				tester.p.joinNetwork();
				//tester.p.unleashZombieKiller();
				tester.p.startRequestTask();
			}else if( args[0].equals( "nep" ) ){
				// start a NetworkEntryPoint
				if( args.length < 5 ){
					printUsageNEP();
					return;
				}
				// get arguments
				int nInitNeighbours = Integer.parseInt( args[1] );
				int minNeighbours = Integer.parseInt( args[2] );
				int maxNeighbours = Integer.parseInt( args[3] );
				boolean recover = Boolean.parseBoolean( args[4] );
				// init NetworkEntryPoint
				tester.nep = new NetworkEntryPoint( nInitNeighbours, minNeighbours, maxNeighbours, recover );
				new Thread( tester.nep, "NetworkEntryPoint" ).start();
			}else if( args[0].equals( "source" ) ){
				// start a source
				if( args.length < 8 ){
					printUsageSource();
					return;
				}
				// get arguments
				String ip = args[1];
				int port = Integer.parseInt( args[2] );
				int bitRate = Integer.parseInt( args[3] );
				int packetSize = Integer.parseInt( args[4] );
				String nepAddress = args[5];
				String streamType = args[6];
				String inputStreamURL = args[7];
				if( streamType.equalsIgnoreCase( "audio" ) ){
					// init audio source
					tester.s = new Source( ip, port, bitRate, packetSize, nepAddress, ProtocolConstants.HTTP_MP3 );
					inAdapter = new ShoutCastAdapter( inputStreamURL, 2999, tester.s );
				}else if( streamType.equalsIgnoreCase( "video" ) ){
					// init video source
					tester.s = new Source( ip, port, bitRate, packetSize, nepAddress, ProtocolConstants.HTTP_VLC );
					inAdapter = new VlcHttpAdapter( inputStreamURL );
				}else{
					printUsage();
					return;
				}
				tester.s.initPacketizer( packetSize );
				tester.s.setInputAdapter( inAdapter );
				tester.s.startDispatchers();
				tester.s.joinNetwork();
				tester.s.unleashZombieKiller();
				tester.s.start();
			}else{
				printUsage();
			}
		}catch( UnknownHostException e ){
			System.out.println( "UnknownHostException" );
		}catch( DispatcherNotRunningException e ){
			System.out.println( "Dispatcher cannot be started " + e.getMessage() );
		}catch( NodeNotConfiguredException e ){
			System.out.println( e.getMessage() );
		}catch( UDPSocketManagerException e ){
			System.out.println( "UDPSocketManger exception: " + e.getMessage() );
		}

	}


	/**
	 * Prints general usage information.
	 */
	private static void printUsage(){

		System.out.println( "\nArguments: [peer <options> | source <options> | nep <options>]" );
		System.out.println( "(nep is the Network Entry Point)" );
		System.out.println( "There must be a nep before a source can start and there must be a source before a peer can join!" );
		System.out.println();
	}


	/**
	 * Prints usage information concerning the Peer.
	 */
	private static void printUsagePeer(){

		System.out.println( "\npeer <ip> <port> <inSpeed> <outSpeed> <nepAddress> <dePacketizerThreshold> <mediaPort>" );
		System.out.println( "<ip>: public IP address or hostname of this peer" );
		System.out.println( "<port>: port number on which this peer will listen for incoming connections" );
		System.out.println( "<inSpeed>: inbound speed in kbps" );
		System.out.println( "<outSpeed>: outbound speed in kbps" );
		System.out.println( "<nepAddress>: IP address of the Network Entry Point" );
		System.out.println( "<dePacketizerThreshold>: the number of packets that need to be available in order before the stream starts to play" );
		System.out.println( "<mediaPort>: local port number to which the media player will be connected" );
		System.out.println();
	}


	/**
	 * Prints usage information concerning the Source.
	 */
	private static void printUsageSource(){

		System.out.println( "\nsource <ip> <port> <bitRate> <packetSize> <nepAddress> <streamType> <inputStreamURL>" );
		System.out.println( "<ip>: public IP address or hostname of this source" );
		System.out.println( "<port>: port number on which this source will listen for incoming connections" );
		System.out.println( "<bitRate>: the bit rate of the stream in kbps" );
		System.out.println( "<packetSize>: the size of the packets that will be sent over the network in bytes (payload only)" );
		System.out.println( "<nepAddress>: IP address of the Network Entry Point" );
		System.out.println( "<streamType>: the type of the stream - either 'audio' or 'video'" );
		System.out.println( "<inputStreamURL>: the URL from which the stream will be received (e.g. 'http://64.236.34.67:80/stream/1074')" );
		System.out.println();
	}


	/**
	 * Prints usage information concerning the NetworkEntryPoint.
	 */
	private static void printUsageNEP(){

		System.out.println( "\nnep: <nInitNeighbours> <minNeighbours> <maxNeighbours> <recover>" );
		System.out.println( "<nInitNeighbours>: the number of neighbours that a joining peer will receive; must be between <minNeighbours> and <maxNeighbours>" );
		System.out.println( "<minNeighbours>: the minimum number of neighbours each peer must have" );
		System.out.println( "<maxNeighbours>: the maximum number of neighbours each peer can have" );
		System.out.println( "<recover>: whether or not to recover a previous state - either 'true' or 'false'" );
		System.out.println();
	}

}
