/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.packets;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.Peer;
import ch.ethz.dcg.p2p.utilities.DePacketizerListener;

/**
 * Responsible for transforming the discrete <code>P2PPacket</code>s back
 * into a byte stream which thereafter will be forwarded to the media player.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class DePacketizer implements Runnable {

	protected final Log logger = LogFactory.getLog( this.getClass() );

	private Hashtable<Long, P2PPacket> availablePackets;
	private int startThreshold = 5;
	private boolean streamStarted = false;
	private Vector<DePacketizerListener> dpls = new Vector<DePacketizerListener>();
	private Peer owner;
	private RingBuffer buffer;
	private int forwardInterval = 1000; // milliseconds
	private int enoughDataToForward;

	private boolean running = true;


	/**
	 * Create a new <code>DePacketizer</code> instance. It depacketises the
	 * packets found in availablePackets and waits until startThreshold
	 * consecutive packets are available.
	 * 
	 * @param availablePackets
	 *            Hashtable that contains the available packets
	 * @param startThreshold
	 *            number of seconds that have to be available before forwarding
	 *            to OutputAdapter(s) starts
	 * @param owner
	 *            the peer that owns this <code>DePacketizer</code>
	 */
	public DePacketizer( Hashtable<Long, P2PPacket> availablePackets, int startThreshold, Peer owner ){

		this.availablePackets = availablePackets;
		if( startThreshold > 0 ){
			this.startThreshold = startThreshold;
		}
		this.owner = owner;
		buffer = new RingBuffer( 2 * owner.getBitrate() * 125 * startThreshold );
		enoughDataToForward = owner.getBitrate() / 8 * forwardInterval;
	}


	/**
	 * Starts this <code>DePacketizer</code>. Data is beeing ordered and
	 * forwarded to the media player.
	 */
	public void run(){

		P2PPacket nextPacket;
		long nextPacketId;

		if( logger.isDebugEnabled() ){
			logger.debug( "DePacketizer running" );
		}

		// wait for the first packet to become available
		while( availablePackets.size() < 5 ){
			synchronized( this ){
				try{
					owner.write2gui( "waiting for packets" );
					wait();
					if( !running ){
						return;
					}
				}catch( InterruptedException e ){
				}
			}
		}

		// get first packet
		owner.write2gui( "start buffering packets..." );
		Iterator<Long> it = availablePackets.keySet().iterator();
		long firstId = it.next();
		owner.setStartPacketId( firstId );
		owner.setCurPacketId( firstId );
		nextPacket = availablePackets.get( firstId );
		buffer.put( nextPacket.getData(), 0, nextPacket.getSize() );

		// start transferring data to the buffer and to the registered adapters
		long lastForwardTime = 0;
		long startTime = System.currentTimeMillis();
		while( running ){
			nextPacketId = owner.getCurPacketId() + 1;
			if( availablePackets.containsKey( nextPacketId ) ){
				owner.setCurPacketId( nextPacketId );
				nextPacket = availablePackets.get( nextPacketId );
				buffer.put( nextPacket.getData(), 0, nextPacket.getSize() );
				if( !streamStarted && ( buffer.size() >= ( owner.getBitrate() * 125 * startThreshold ) ) ){
					// stream not started and buffer not yet full
					streamStarted = true;
					owner.write2gui( "...buffer full; starting the stream" );
					owner.setStartTime( System.currentTimeMillis() );
				}
				if( streamStarted && ( buffer.size() >= enoughDataToForward ) ){
					// stream started and enough data in buffer to forward next
					// junk to the media player
					lastForwardTime = System.currentTimeMillis();
					forwardDataToListeners( buffer.get( enoughDataToForward ) );
				}
				// update ranking of providing peer
				owner.honorPeer( nextPacket.getReceivedFromKey() );
			}else if( streamStarted && ( System.currentTimeMillis() - lastForwardTime ) >= 1.5 * forwardInterval ){
				// replace missing packt in buffer with 0-bytes
				buffer.put( new byte[owner.getPacketSize()], 0, owner.getPacketSize() );
				// forward data from the buffer if there is some
				if( buffer.size() >= enoughDataToForward ){
					forwardDataToListeners( buffer.get( enoughDataToForward ) );
					lastForwardTime = System.currentTimeMillis();
				}
				// skip packet that is blocking the depacketizer
				owner.setCurPacketId( nextPacketId );
				owner.write2gui( "skipping packet " + nextPacketId );
				// blame the peer that was asked for the missing packet
				owner.blamePeer( nextPacketId - 1 );
			}else if( !streamStarted && ( System.currentTimeMillis() - startTime ) >= ( 2 * startThreshold * 1000 ) ){
				// replace missing packt in buffer with 0-bytes
				buffer.put( new byte[owner.getPacketSize()], 0, owner.getPacketSize() );
				// skip packet that is blocking the depacketizer
				owner.setCurPacketId( nextPacketId );
				synchronized( this ){
					try{
						wait();
					}catch( InterruptedException e ){
					}
				}
			}else{
				synchronized( this ){
					try{
						owner.moveToNotifications( nextPacketId );
						wait();
					}catch( InterruptedException e ){
					}catch( NodeNotConfiguredException e ){
					}
				}
			}
		}
	}


	/**
	 * Forwards the data of the specified packet to the registered
	 * <code>DePacketizerListener</code>s.
	 * 
	 */
	private void forwardDataToListeners( byte[] data ){

		if( !dpls.isEmpty() ){
			for( int i = 0; i < dpls.size(); i++ ){
				dpls.elementAt( i ).dataAvailable( data, data.length );
			}
		}
	}


	/**
	 * Adds a <code>DePacketizerListener</code> to this
	 * <code>DePacketizer</code>.
	 * 
	 * @param dpl
	 *            the <code>DePacketizerListener</code> to add
	 */
	public void addDePacketizerListener( DePacketizerListener dpl ){

		dpls.add( dpl );
	}


	/**
	 * Removes a <code>DePacketizerListener</code> from this
	 * <code>DePacketizer</code>.
	 * 
	 * @param dpl
	 *            the <code>DePacketizerListener</code> to remove
	 */
	public void removeDePacketizerListener( DePacketizerListener dpl ){

		dpls.remove( dpl );
	}


	/**
	 * Stops this <code>DePacketizer</code>.
	 */
	public void close(){

		running = false;
		synchronized( this ){
			notify();
		}
	}
}
