/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.packets;

import java.util.Hashtable;

import ch.ethz.dcg.p2p.peers.Neighbour;

/**
 * Represents notifications received from the neighbours which contains
 * information about who has which packets. There exists a
 * <code>Notification</code> object for each received notification for which
 * the corresponding packet has not yet been requested or received.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class Notification {

	private long packetId;
	// nodes that sent a notification
	private Hashtable<String, Neighbour> holders = new Hashtable<String, Neighbour>();
	// node to which a request has been sent
	private Neighbour provider;
	private long startTime;


	/**
	 * Creates a <code>Notification</code>.
	 * 
	 * @param packetId
	 *            the id of the packet which this notification represents
	 * @param holder
	 *            the neighbour that has the packet with the given id; usually
	 *            the neighbour from which this <code>Notification</code> has
	 *            been received
	 */
	public Notification( long packetId, Neighbour holder ){

		this.packetId = packetId;
		holders.put( hashKey( holder ), holder );
	}


	/**
	 * Generates the hash key for a given Neighbour object. The key is
	 * 'IPAddress:PortNumber'
	 * 
	 * @param n
	 *            the neighbour object for which the key should be generated
	 * @return a String representing the neighbour
	 */
	private String hashKey( Neighbour n ){

		return n.getIp().getHostAddress() + ":" + n.getPort();
	}


	/**
	 * Adds a <code>Neighbour</code> to the vector holding the nodes from
	 * which I have received this <code>Notification</code>. If the specified
	 * node is already known to have sent this <code>Notification</code>, he
	 * will not be added again.
	 * 
	 * @param n
	 *            the neighbour which has the packet with this id
	 * @return true if the neighbour was added; false otherwise
	 */
	public boolean addHolder( Neighbour n ){

		String nKey = hashKey( n );
		if( holders.containsKey( nKey ) ){
			// neighbour already registered as holder
			return false;
		}else{
			holders.put( nKey, n );
			return true;
		}
	}


	/**
	 * Sets the <code>Neighbour</code> to which I have sent a request for this
	 * packet. This <code>Neighbour</code> will be removed from the list of
	 * holders.
	 * 
	 * @param n
	 *            the neighbour to which a request has been sent
	 */
	public void setProvider( Neighbour n ){

		provider = n;
		holders.remove( hashKey( n ) );
	}


	/**
	 * Get the <code>Neighbour</code> to which a request for this packet has
	 * been sent.
	 * 
	 * @return the corresponding <code>Neighbour</code> or null if the packet
	 *         has not yet been requested
	 */
	public Neighbour getProvider(){

		return provider;
	}


	/**
	 * Remove <code>Neighbour</code> n from the list of packet holders.
	 * 
	 * @param n
	 *            the <code>Neighbour</code> to remove
	 */
	public void removeHolder( Neighbour n ){

		holders.remove( hashKey( n ) );
	}


	/**
	 * Get the vector of <code>Neighbours</code> that have sent this
	 * <code>Notification</code>.
	 * 
	 * @return vector of <code>Neighbours</code> that have sent this
	 *         <code>Notification</code>
	 */
	public Hashtable<String, Neighbour> getHolders(){

		return holders;
	}


	/**
	 * Get the id which this <code>Notification</code> represents.
	 * 
	 * @return the packet id of this <code>Notification</code>
	 */
	public long getPacketId(){

		return packetId;
	}


	/**
	 * Get the time when the packet of this notification has been requested.
	 * 
	 * @return time when the packet of this notification has been requested
	 */
	public long getStartTime(){

		return startTime;
	}


	/**
	 * Set the time when the packet of this notification has been requested.
	 * 
	 * @param startTime
	 *            the time when this packet has been requested
	 */
	public void setStartTime( long startTime ){

		this.startTime = startTime;
	}
}
