/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.packets;

import ch.ethz.dcg.p2p.exceptions.P2PPacketException;
import ch.ethz.dcg.p2p.peers.Neighbour;

/**
 * Represents the packets that are sent around in the Swistry network.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class P2PPacket {

	private long id;
	private int size;
	private byte[] data;
	private Neighbour receivedFrom;


	/**
	 * Creates a new P2PPacket which is the media-data container within the
	 * Swistry system.
	 * 
	 * @param id
	 *            serial number of this packet
	 * @param size
	 *            the number of bytes (stream payload) that this packet contains
	 * @param receivedFrom
	 *            the Neighbour from which this packet has come
	 */
	public P2PPacket( long id, int size, Neighbour receivedFrom ){

		this.id = id;
		this.size = size;
		this.receivedFrom = receivedFrom;
	}


	/**
	 * Get the packet id.
	 * 
	 * @return the packet id
	 */
	public long getId(){

		return id;
	}


	/**
	 * Set the packet id.
	 * 
	 * @param id
	 *            packet id
	 */
	public void setId( long id ){

		this.id = id;
	}


	/**
	 * Get the size of this packet.
	 * 
	 * @return the packet size
	 */
	public int getSize(){

		return size;
	}


	/**
	 * Set the size of this packet.
	 * 
	 * @param bytes
	 *            number of bytes this packet contains
	 */
	public void setSize( int bytes ){

		size = bytes;
	}


	/**
	 * Get the media data of this packet.
	 * 
	 * @return the media data of this packet
	 */
	public byte[] getData(){

		return data;
	}


	/**
	 * Set the media data of this packet.
	 * 
	 * @param data
	 *            media data this packet should contain
	 */
	public void setData( byte[] data ) throws P2PPacketException{

		if( size != data.length ){
			throw new P2PPacketException( "specified data size does not match packet data size" );
		}
		this.data = data;
	}


	/**
	 * Get the <code>Neighbour</code> that has sent this packet.
	 * 
	 * @return the <code>Neighbour</code> that has sent this packet
	 */
	public Neighbour getReceivedFrom(){

		return receivedFrom;
	}


	public String getReceivedFromKey(){

		if( receivedFrom != null ){
			return receivedFrom.getIp().getHostAddress() + ":" + receivedFrom.getPort();
		}else{
			return "";
		}
	}
}
