/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.packets;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.exceptions.P2PPacketException;
import ch.ethz.dcg.p2p.peers.Source;

/**
 * Responsible for transforming the byte stream from the media source into
 * discrete packets which are transported through the Swistry network.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class Packetizer {

	protected final Log logger = LogFactory.getLog( this.getClass() );

	private int bufferPos = 0;
	private int packetSize = 16000;
	private byte[] buffer;
	private long lastPacketId = 1;
	private Source sourceNode;


	/**
	 * Creates a new <code>Packetizer</code> that has a buffer into which data
	 * will be stored until there is enough to fill a packet. At that moment a
	 * new <code>P2PPacket</code> will be created and added to the
	 * <code>Source</code> to which this <code>Packetizer</code> belongs.
	 * This packet will cary a default amount of 16kB of data.
	 * 
	 * @param sourceNode
	 *            the <code>Source</code> to which this
	 *            <code>Packetizer</code> belongs
	 */
	public Packetizer( Source sourceNode ){

		buffer = new byte[packetSize];
		this.sourceNode = sourceNode;
	}


	/**
	 * Create a new <code>Packetizer</code> that has a buffer into which data
	 * will be stored until there is enough to fill a packet. At that moment a
	 * new <code>P2PPacket</code> will be created and added to the
	 * <code>Source</code> to which this <code>Packetizer</code> belongs.
	 * 
	 * @param packetSize
	 *            the number of bytes this packet will carry
	 * @param sourceNode
	 *            the <code>Source</code> to which this
	 *            <code>Packetizer</code> belongs
	 */
	public Packetizer( int packetSize, Source sourceNode ){

		this.packetSize = packetSize;
		buffer = new byte[packetSize];
		this.sourceNode = sourceNode;
	}


	/**
	 * Add data to the buffer. If there is enough data available a new
	 * <code>P2PPacket</code> will be created and added to the
	 * <code>Source</code>.
	 * 
	 * @param data
	 *            byte array containing the data
	 * @param dataSize
	 *            the number of bytes to be sent
	 */
	public void addData( byte[] data, int dataSize ){

		if( ( bufferPos + dataSize ) < packetSize ){
			// not enough data to create a new packet -> add it to buffer
			System.arraycopy( data, 0, buffer, bufferPos, dataSize );
			bufferPos += dataSize;
			return;
		}else{
			// create a new packet
			if( logger.isDebugEnabled() ){
				logger.debug( "creating new packet (id = " + lastPacketId + ")" );
			}

			P2PPacket p = new P2PPacket( lastPacketId, packetSize, null );
			byte[] tmp = new byte[packetSize];

			// copy the buffer content into the packet first
			System.arraycopy( buffer, 0, tmp, 0, bufferPos );

			// create as much packets as possible with this data
			int remainingData = dataSize;
			int dataPos = 0;
			while( ( remainingData + bufferPos ) >= packetSize ){
				int remainingSpace = packetSize - bufferPos;
				System.arraycopy( data, dataPos, tmp, bufferPos, remainingSpace );
				bufferPos = 0;
				dataPos += remainingSpace;
				remainingData -= remainingSpace;
				// copy data into the packet
				try{
					p.setData( tmp );
					sourceNode.addPacket( p, 0 );
					sourceNode.setCurPacketId( lastPacketId );
					lastPacketId++;
					p = new P2PPacket( lastPacketId, packetSize, null );
				}catch( P2PPacketException e ){
					if( logger.isErrorEnabled() ){
						logger.error( "cannot create p2p packet: " + e.getMessage() );
					}
				}catch( NodeNotConfiguredException e ){
					if( logger.isErrorEnabled() ){
						logger.error( e.getMessage() );
					}
				}
			}

			// copy the rest that is not enough for another packet into the
			// buffer for the next packet
			System.arraycopy( data, dataPos, buffer, 0, remainingData );
			bufferPos = remainingData;
		}
	}


	/**
	 * Get the size of the packets that are created in this
	 * <code>Packetizer</code>.
	 * 
	 * @return the size of the packets
	 */
	public int getPacketSize(){

		return packetSize;
	}
}
