/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.packets;

/**
 * A RingBuffer to store bytes. It has a fixed size. If there are more bytes
 * added to the buffer than there is free space, the oldest bytes will be
 * replaced in order.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class RingBuffer {

	private int capacity = 0;
	private int first = 0;
	private int last = 0;
	private int nElements = 0;
	private byte[] buffer;


	/**
	 * Initialises the RingBuffer with the specified size.
	 * 
	 * @param capacity
	 *            the number of bytes this buffer can store at any time
	 */
	public RingBuffer( int capacity ){

		if( capacity > 0 ){
			this.capacity = capacity;
		}else{
			capacity = 65000;
		}
		buffer = new byte[capacity];
	}


	/**
	 * Puts new data into this buffer.
	 * 
	 * @param data
	 *            the data to add
	 * @param offset
	 *            the offset in the data array
	 * @param len
	 *            the length of the data to store
	 */
	public synchronized void put( byte[] data, int offset, int len ){

		int copyLen;
		if( isEmpty() ){
			copyLen = Math.min( len, capacity );
			System.arraycopy( data, offset, buffer, 0, copyLen );
			last = copyLen % capacity;
			first = 0;
			nElements = copyLen;
			if( copyLen < len ){
				put( data, offset + copyLen, len - copyLen );
			}
		}else if( isFull() ){
			copyLen = Math.min( len, capacity - last );
			System.arraycopy( data, offset, buffer, last, copyLen );
			last = ( last + copyLen ) % capacity;
			first = last;
			if( copyLen < len ){
				put( data, offset + copyLen, len - copyLen );
			}
		}else if( first < last ){
			copyLen = Math.min( len, capacity - last );
			System.arraycopy( data, offset, buffer, last, copyLen );
			last = ( last + copyLen ) % capacity;
			nElements += copyLen;
			if( copyLen < len ){
				put( data, offset + copyLen, len - copyLen );
			}
		}else if( first > last ){
			copyLen = Math.min( len, first - last );
			System.arraycopy( data, offset, buffer, last, copyLen );
			last = ( last + copyLen ) % capacity;
			nElements += copyLen;
			if( copyLen < len ){
				put( data, offset + copyLen, len - copyLen );
			}
		}
	}


	/**
	 * Gets a certain amount of bytes from this buffer. The read bytes will be
	 * removed from the buffer.
	 * 
	 * @param nBytes
	 *            the number of bytes to read from this RingBuffer
	 * @return a byte array containing the bytes; if there are less than nBytes
	 *         bytes in the buffer the returned byte array will be shorter than
	 *         nBytes (only the available bytes will then be returned)
	 */
	public synchronized byte[] get( int nBytes ){

		int outLen = Math.min( nBytes, nElements );
		byte[] res = new byte[outLen];
		if( isFull() ){
			int outCopy = Math.min( nBytes, capacity - first );
			System.arraycopy( buffer, first, res, 0, outCopy );
			first = ( first + outCopy ) % capacity;
			nElements -= outCopy;
			if( outCopy < outLen ){
				outLen -= outCopy;
				System.arraycopy( buffer, first, res, outCopy, outLen );
				first = ( first + outLen ) % capacity;
				nElements -= outLen;
			}
		}else if( first < last ){
			System.arraycopy( buffer, first, res, 0, outLen );
			first = ( first + outLen ) % capacity;
			nElements -= outLen;
		}else if( last < first ){
			int outCopy = Math.min( nBytes, capacity - first );
			System.arraycopy( buffer, first, res, 0, outCopy );
			first = ( first + outCopy ) % capacity;
			nElements -= outCopy;
			if( outCopy < outLen ){
				outLen -= outCopy;
				System.arraycopy( buffer, first, res, outCopy, outLen );
				first = ( first + outLen ) % capacity;
				nElements -= outLen;
			}
		}
		return res;
	}


	/**
	 * Tests if this buffer is empty.
	 * 
	 * @return true if it contains no elments; false otherwise
	 */
	public synchronized boolean isEmpty(){

		return nElements == 0;
	}


	/**
	 * Tests if this buffer is full.
	 * 
	 * @return true if it is full; false otherwise
	 */
	public synchronized boolean isFull(){

		return nElements == capacity;
	}


	/**
	 * Resets the buffer's content to zero.
	 * 
	 */
	public synchronized void reset(){

		first = 0;
		last = 0;
		nElements = 0;
	}


	/**
	 * Gets the size of this buffer.
	 * 
	 * @return the buffer capacity
	 */
	public synchronized int getCapacity(){

		return capacity;
	}


	/**
	 * Get the number of bytes currently stored in this buffer.
	 * 
	 * @return the number of bytes currently stored in this buffer
	 */
	public synchronized int size(){

		return nElements;
	}
}
