/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.peers;

import java.io.Serializable;
import java.net.InetAddress;
import java.net.UnknownHostException;

import ch.ethz.dcg.p2p.tasks.PingTask;

/**
 * Represents a neighbour. It contains information like IP address, port number,
 * ranking etc.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class Neighbour implements Comparable, Serializable {

	private InetAddress ip;
	private int port;
	private int rtt = -1;
	private long knownSince = System.currentTimeMillis();

	private int ranking = 0;
	private static int max_ranking = 1000;
	private static int min_ranking = -max_ranking;


	/**
	 * Creates a new <code>Neighbour</code> with the given IP address and port
	 * number.
	 * 
	 * @param ip
	 *            String representing the IP address of this
	 *            <code>Neighbour</code>
	 * @param port
	 *            port number of this <code>Neighbour</code>
	 * @throws UnknownHostException
	 *             if the IP address is invalid
	 */
	public Neighbour( String ip, int port ) throws UnknownHostException{

		this.ip = InetAddress.getByName( ip );
		this.port = port;
	}


	/**
	 * Creates a new <code>Neighbour</code> with the given IP address and port
	 * number.
	 * 
	 * @param ip
	 *            IP address of this <code>Neighbour</code>
	 * @param port
	 *            port number of this <code>Neighbour</code>
	 */
	public Neighbour( InetAddress ip, int port ){

		this.ip = ip;
		this.port = port;
	}


	/**
	 * Get the IP address of this <code>Neighbour</code>.
	 * 
	 * @return the IP address of this <code>Neighbour</code>
	 */
	public InetAddress getIp(){

		return ip;
	}


	/**
	 * Get the port of this <code>Neighbour</code>.
	 * 
	 * @return the port of this <code>Neighbour</code>
	 */
	public int getPort(){

		return port;
	}


	/**
	 * Get the ranking of this <code>Neighbour</code>.
	 * 
	 * @return the ranking of this <code>Neighbour</code>
	 */
	public int getRanking(){

		return ranking;
	}


	/**
	 * Set the ranking of this <code>Neighbour</code>.
	 * 
	 * @param ranking
	 *            ranking of this <code>Neighbour</code>
	 */
	public void setRanking( int ranking ){

		this.ranking = ranking;
	}


	/**
	 * Increase the ranking of this <code>Neighbour</code>.
	 * 
	 * @return the current ranking of this <code>Neighbour</code>
	 */
	public int increaseRanking(){

		if( ranking < max_ranking ){
			ranking++;
		}
		return ranking;
	}


	/**
	 * Decrease the ranking of this <code>Neighbour</code>.
	 * 
	 * @return the current ranking of this <code>Neighbour</code>
	 */
	public int decreaseRanking(){

		if( ranking > min_ranking ){
			ranking--;
		}
		return ranking;
	}


	/**
	 * Compare this <code>Neighbour</code> to another one. Part of the
	 * <code>Comparable</code> interface.
	 * 
	 * @param o
	 *            the object to which this <code>Neighbour</code> should be
	 *            compared
	 * @return -1 if this has a lower ranking than the argument. Returns 1 if
	 *         this has a higher ranking than the argument. Returns 0 if this
	 *         and ther argument have equal rankings.
	 */
	public int compareTo( Object o ){

		Neighbour r = (Neighbour)o;

		if( ranking < r.ranking ){
			return -1;
		}else if( ranking > r.ranking ){
			return 1;
		}else{
			return 0;
		}
	}


	/**
	 * Send a PING message to this <code>Neighbour</code> and wait timeout
	 * milliseconds for the PONG response.
	 * 
	 * @param timeout
	 *            timeout for PONG response in millliseconds
	 * @return true if the PONG response arrived within the timeout; false
	 *         otherwise
	 * 
	 */
	public boolean isReachable( int timeout ){

		PingTask ping = new PingTask( this, timeout );
		return ping.sendPINGreceivePONG();
	}


	/**
	 * Compare two <code>Neighbour</code> objects based on their ip address
	 * and port number.
	 * 
	 * @param n
	 *            to <code>Neighbour</code> to which this should be compared
	 * @return true, if this <code>Neighbour</code> and the argument have the
	 *         same ip address and port number; false otherwise
	 */
	public boolean equals( Neighbour n ){

		if( n.ip.equals( ip ) && ( n.port == port ) ){
			return true;
		}else{
			return false;
		}
	}


	/**
	 * Compare the IP address and the port of this <code>Neighbour</code> to
	 * the given arguments.
	 * 
	 * @param ip
	 * @param port
	 * @return true if they match; false otherwise
	 */
	public boolean equals( InetAddress ip, int port ){

		if( ip.equals( this.ip ) && ( port == this.port ) ){
			return true;
		}else{
			return false;
		}
	}


	/**
	 * Get the average round trip time of this <code>Neighbour</code>.
	 * 
	 * @return the average round trip time of this <code>Neighbour</code>
	 */
	public int getRTT(){

		return rtt;
	}


	/**
	 * Update the average round trip time of this <code>Neighbour</code>. A
	 * moving average estimation is used. The current RTT is calculates using<br>
	 * 0.1 * newRTT + 0.9 * oldRTT.
	 * 
	 * @param rtt
	 *            the new RTT which should be added to the existing one
	 * @return the current RTT of this <code>Neighbour</code>
	 */
	public int updateRTT( int rtt ){

		if( this.rtt == -1 ){
			this.rtt = rtt;
		}else{
			this.rtt = (int)( rtt * 0.1 + this.rtt * 0.9 );
		}
		return this.rtt;
	}


	/**
	 * Get the time in milliseconds since when this Neighbour is known.
	 * 
	 * @return the time in milliseconds since when this Neighbour is known
	 */
	public long getKnownSince(){

		return knownSince;
	}

}
