/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.peers;

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.UnknownHostException;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.dcg.p2p.exceptions.PersistencyManagerException;
import ch.ethz.dcg.p2p.gui.NEPGUI;
import ch.ethz.dcg.p2p.handlers.NEPJoinByeHandler;
import ch.ethz.dcg.p2p.utilities.DummyNode;
import ch.ethz.dcg.p2p.utilities.PersistencyManager;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.SortedNeighbourList;

/**
 * The core element in the Swistry network. The Network Entry Point acts similar
 * to a bittorrent tracker in that every new node joines over him and he
 * provides the new node with a set of inital neighbours. The Network Entry
 * Point also keeps track of the available levels and puts the joining nodes in
 * the levels to which they belong according to their reported outbound
 * bandwidth.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class NetworkEntryPoint implements Runnable {

	private final Log logger = LogFactory.getLog( this.getClass() );

	// states
	public static final String idle = "idle";
	public static final String configured = "configured";
	public static final String connected = "connected";
	protected String status = idle;

	// thread pool with threads that handle incoming requests
	private ExecutorService threadPool = Executors.newCachedThreadPool();

	InetAddress nepIp = null;
	private ServerSocket serverSocket;

	// existing levels (each level has exactly one source and zero or more
	// peers)
	private Hashtable<Integer, SortedNeighbourList> levels = new Hashtable<Integer, SortedNeighbourList>();
	// existing sources
	private Hashtable<Integer, NEPNeighbour> sources = new Hashtable<Integer, NEPNeighbour>();
	// number of initial neighbours
	private int nInitNeighbours = 3;
	// minimum amount of neighbours each peer should have
	private int minNeighs = 2;
	// maximum amount of neighbours each peer is allowed to have
	private int maxNeighs = 6;
	private int pingTimeout = 3000;
	// average speed per level (for statistics)
	private Hashtable<Integer, int[]> avgSpeeds = new Hashtable<Integer, int[]>();
	private byte streamType = ProtocolConstants.UNKNOWN_TYPE;
	private int minSpeed = Integer.MAX_VALUE;
	// peers need levelSpeed*capacityRequired to join
	private float capacityRequired = (float)1.3;

	private volatile boolean running;
	private NEPGUI gui;

	// backup of state into a file in case the NEP has to restarted or crashes
	private PersistencyManager persistencyManager;


	/**
	 * Creates a new <code>NetworkEntryPoint</code>.<br>
	 * This constructor is called when run in gui mode.
	 * 
	 * @param gui
	 *            gui the gui object that represents this
	 *            <code>NetworkEntryPoint</code>
	 */
	public NetworkEntryPoint( NEPGUI gui ){

		this.gui = gui;
		persistencyManager = new PersistencyManager( this );
	}


	/**
	 * Creates a new <code>NetworkEntryPoint</code>.<br>
	 * This constructor is called when run in console mode.
	 * 
	 * @param nInitNeighbours
	 *            number of inital neighbours
	 * @param minNeighs
	 *            minimum number of neighbours required
	 * @param maxNeighs
	 *            maximum nubmer of neighbours allowed
	 * @param recover
	 *            use the information stored in the file or discard that
	 *            information and start over
	 */
	public NetworkEntryPoint( int nInitNeighbours, int minNeighs, int maxNeighs, boolean recover ){

		persistencyManager = new PersistencyManager( this );
		if( recover ){
			boolean success = recover();
			if( success ){
				return;
			}
		}
		this.nInitNeighbours = nInitNeighbours;
		this.minNeighs = minNeighs;
		this.maxNeighs = maxNeighs;
	}


	/**
	 * Creates a new <code>NetworkEntryPoint</code>.<br>
	 * This constructor is called when run in console mode.
	 * 
	 * @param ip
	 *            IP address to which this <code>NetworkEntryPoint</code>
	 *            should be bound
	 * @param nInitNeighbours
	 *            number of inital neighbours
	 * @param minNeighs
	 *            minimum number of neighbours required
	 * @param maxNeighs
	 *            maximum nubmer of neighbours allowed
	 * @param recover
	 *            use the information stored in the file or discard that
	 *            information and start over
	 */
	public NetworkEntryPoint( String ip, int nInitNeighbours, int minNeighs, int maxNeighs, boolean recover ){

		persistencyManager = new PersistencyManager( this );
		if( recover ){
			boolean success = recover();
			if( success ){
				return;
			}
		}
		try{
			nepIp = InetAddress.getByName( ip );
			this.nInitNeighbours = nInitNeighbours;
			this.minNeighs = minNeighs;
			this.maxNeighs = maxNeighs;
		}catch( UnknownHostException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( "unknown host " + ip );
			}
		}
	}


	/**
	 * Starts the <code>NetworkEntryPoint</code>.
	 */
	public void run(){

		// open ServerSocket
		try{
			if( nepIp != null ){
				serverSocket = new ServerSocket( ProtocolConstants.NEPPORT, 50, nepIp );
			}else{
				serverSocket = new ServerSocket( ProtocolConstants.NEPPORT );
			}
		}catch( IOException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( "cannot open ServerSocket; stopping" );
			}
			return;
		}

		running = true;

		if( logger.isDebugEnabled() ){
			logger.debug( "NetworkEntryPoint running" );
		}

		// wait for incoming connections
		while( running ){
			try{
				if( logger.isDebugEnabled() ){
					logger.debug( "waiting for incoming connections" );
				}
				// create a new thread to handle the incoming connection
				threadPool.execute( new NEPJoinByeHandler( serverSocket.accept(), this ) );
				if( logger.isDebugEnabled() ){
					logger.debug( "new incoming connection accepted, handler started" );
				}
			}catch( IOException e ){
				if( logger.isFatalEnabled() ){
					logger.fatal( "cannot accept socket connection ( " + e.getMessage() + " )" );
				}
			}
		}
	}


	/**
	 * Stop this <code>NetworkEntryPoint</code>.<br>
	 * The <code>ServerSocket</code> will be closed.
	 */
	public void stop(){

		running = false;
		try{
			serverSocket.close();
		}catch( Exception e ){
		}
	}


	/**
	 * Insert a new <code>Source</code> as source of a specified level. The
	 * new source will only be inserted if it is not already there. The
	 * comparison is done based on the level of the source because each level or
	 * speed can only have one source.
	 * 
	 * @param newSource
	 *            the new <code>Source</code> to insert
	 * @return true if the <code>Source</code> was inserted; false if the
	 *         <code>Source</code> was already there
	 */
	public synchronized boolean insertSource( NEPNeighbour newSource ){

		int sourceKey = newSource.getLevel();
		if( !sources.containsKey( sourceKey ) ){
			sources.put( sourceKey, newSource );
			writeStatePersistently();
			return true;
		}else{
			return false;
		}
	}


	/**
	 * Remove the <code>Source</code> with the given level.
	 * 
	 * @param level
	 *            the level from which the <code>Source</code> should be
	 *            removed
	 * @return true if it was removed; false otherwise
	 */
	private synchronized boolean removeSource( int level ){

		if( sources.containsKey( level ) ){
			sources.remove( level );
			writeStatePersistently();
			return true;
		}else{
			return false;
		}
	}


	/**
	 * Check if the source for the given level already exists.
	 * 
	 * @param level
	 *            the level in question
	 * @return true if the source exists; false otherwise
	 */
	public synchronized boolean containsSource( Integer level ){

		return sources.containsKey( level );
	}


	/**
	 * Insert a <code>Neighbour</code> into the list of peers that have joined
	 * so far.
	 * 
	 * @param n
	 *            the new neighbour to insert
	 * @return true on success and false if the <code>NEPNeighbour</code>
	 *         already exists or if it is slower than the slowest level
	 */
	public synchronized boolean insertNEPNeighbour( NEPNeighbour n ){

		SortedNeighbourList peers;
		int levelSpeed = findLevel( n );
		if( levels.containsKey( levelSpeed ) ){
			peers = levels.get( levelSpeed );
		}else{
			// peer is too slow for all levels; he cannot participate
			return false;
		}
		if( !peers.contains( n ) ){
			boolean result = peers.insert( n );
			if( result && logger.isDebugEnabled() ){
				if( n.getNodeType() == ProtocolConstants.SOURCE ){
					updateAvgSpeeds( levelSpeed, n.getLevel() );
				}else{
					updateAvgSpeeds( levelSpeed, n.getSpeed() );
				}
				logger.debug( "NEPNeighbour " + n.getIp().getHostAddress() + ":" + n.getPort() + " inserted at level " + levelSpeed );
				writeStatePersistently();
			}else if( logger.isDebugEnabled() ){
				logger.debug( "NEPNeighbour " + n.getIp().getHostAddress() + ":" + n.getPort() + " has re-joined" );
			}
			return result;
		}else{
			return false;
		}
	}


	/**
	 * Remove the <code>NEPNeighbour</code> with the given IP address and
	 * port.
	 * 
	 * @param ip
	 *            IP of the neighbour to remove
	 * @param port
	 *            port of the neighbour to remove
	 * @return true if it was found and removed; false otherwise
	 */
	public synchronized boolean removeNEPNeighbour( InetAddress ip, int port ){

		Iterator<SortedNeighbourList> it = levels.values().iterator();
		while( it.hasNext() ){
			SortedNeighbourList peers = it.next();
			NEPNeighbour n = peers.contains( ip, port );
			if( n != null ){
				peers.remove( n );
				writeStatePersistently();
				if( logger.isDebugEnabled() ){
					logger.debug( "NEPNeighbour " + ip.getHostAddress() + ":" + port + " removed" );
				}
				return true;
			}
		}
		return false;
	}


	/**
	 * Check if the host with the given IP and port is registered at this
	 * <code>NetworkEntryPoint</code>.
	 * 
	 * @param ip
	 *            IP address in question
	 * @param port
	 * 
	 * @return the <code>NEPNeighbour</code> object that represents this host;
	 *         null if it does not exists
	 */
	public synchronized NEPNeighbour containsNEPNeighbour( InetAddress ip, int port ){

		Iterator<SortedNeighbourList> it = levels.values().iterator();
		while( it.hasNext() ){
			SortedNeighbourList peers = it.next();
			NEPNeighbour n = peers.contains( ip, port );
			if( n != null ){
				return n;
			}
		}
		return null;
	}


	/**
	 * Check if the host with given IP and port is registered at this
	 * <code>NetworkEntryPoint</code> in a certain level.
	 * 
	 * @param ip
	 *            IP address in question
	 * @param port
	 *            port in question
	 * @param level
	 *            level in question
	 * @return the <code>NEPNeighbour</code> object that represents this host
	 *         at this level; null if it does not exists
	 * 
	 */
	public synchronized NEPNeighbour containsNEPNeighbourAtLevel( InetAddress ip, int port, int level ){

		if( !levels.containsKey( level ) ){
			return null;
		}else{
			SortedNeighbourList peers = levels.get( level );
			return peers.contains( ip, port );
		}
	}


	/**
	 * Get the inital neighbours that are closest to the argument in terms of
	 * speed.
	 * 
	 * @param n
	 *            the <code>NEPNeighbour</code> for which initial neighbours
	 *            should be found
	 * @return a <code>Vector</code> containing the initial
	 *         <code>Neighbour</code>s - it might be empty!
	 */
	public synchronized Vector<Neighbour> getInitNeighbours( NEPNeighbour n ){

		Vector<Neighbour> result = new Vector<Neighbour>();
		SortedNeighbourList peers;
		int leftOut = 0;
		int levelSpeed;

		levelSpeed = findLevel( n );
		if( levels.containsKey( levelSpeed ) ){
			peers = levels.get( levelSpeed );
		}else{
			// peer is too slow for all levels; he cannot participate
			return result;
		}

		// check if list is empty
		if( peers.size() == 0 ){
			return result;
		}

		// if the NEPNeighbour is a source, he will get the fastest nodes
		// available in his level
		if( n.getNodeType() == ProtocolConstants.SOURCE ){
			int i = peers.size() - 1;
			while( ( i >= 0 ) && ( result.size() < nInitNeighbours ) ){
				NEPNeighbour next = peers.elementAt( i-- );
				if( next.equals( n ) ){
					i--;
					continue;
				}else if( next.isReachable( pingTimeout ) ){
					result.add( peers.elementAt( i ) );
				}else{
					// TODO: this might be inappropriate
					if( logger.isWarnEnabled() ){
						logger.warn( "peer " + peers.elementAt( i ).getIp().getHostAddress() + ":" + peers.elementAt( i ).getPort() + " not reachable, he will be removed" );
					}
					peers.remove( peers.elementAt( i ) );
				}
			}
			return result;
		}

		int closest = peers.findClosest( n.getSpeed() );
		int i = closest;

		// add peers with the same speed as neigh
		// (choose the ones that have joined last)
		while( ( i < peers.size() ) && ( result.size() < nInitNeighbours ) && ( peers.elementAt( i ).getSpeed() == n.getSpeed() ) ){
			NEPNeighbour next = peers.elementAt( i );
			if( next.equals( n ) ){
				// do not propose himself as a neighbour of his
				leftOut++;
				i++;
				continue;
			}else if( next.isReachable( pingTimeout ) ){
				result.add( peers.elementAt( i++ ) );
			}else{
				// TODO: this might be inappropriate
				if( logger.isWarnEnabled() ){
					logger.warn( "peer " + peers.elementAt( i ).getIp().getHostAddress() + ":" + peers.elementAt( i ).getPort() + " not reachable, he will be removed" );
				}
				peers.remove( peers.elementAt( i ) );
			}
		}

		// insert peers that are slightly slower or faster (closest in the list)
		// only insert them if they are reachable!
		closest--;
		while( ( result.size() < nInitNeighbours ) && ( ( result.size() + leftOut ) < peers.size() ) ){
			// add faster peer
			if( i < peers.size() ){
				NEPNeighbour faster = peers.elementAt( i );
				if( faster.equals( n ) ){
					// do not propose himself as a neighbour of his
					leftOut++;
					i++;
					continue;
				}else if( faster.isReachable( pingTimeout ) ){
					result.add( peers.elementAt( i++ ) );
				}else{
					// TODO: this might be inappropriate
					if( logger.isWarnEnabled() ){
						logger.warn( "peer " + peers.elementAt( i ).getIp() + ":" + peers.elementAt( i ).getPort() + " not reachable" );
					}
					peers.remove( peers.elementAt( i ) );
				}
			}
			// add slower peer
			if( ( closest >= 0 ) && ( result.size() < nInitNeighbours ) ){
				NEPNeighbour slower = peers.elementAt( closest );
				if( slower.equals( n ) ){
					leftOut++;
					closest--;
					continue;
				}else if( slower.isReachable( pingTimeout ) ){
					result.add( peers.elementAt( closest-- ) );
				}else{
					// TODO: this might be inappropriate
					if( logger.isWarnEnabled() ){
						logger.warn( "peer " + peers.elementAt( closest ).getIp() + ":" + peers.elementAt( closest ).getPort() + " not reachable" );
					}
					peers.remove( peers.elementAt( closest-- ) );
					i--;
				}
			}
		}
		return result;
	}


	/**
	 * Print the currently available <code>NEPNeighbours</code> to the console
	 * using the logger. If the log level is above DEBUG the neighbours will not
	 * be printed.<br>
	 * The neighbours are always updated in the <code>NEPGUI</code> if run in
	 * gui mode.
	 */
	public void printNeighbours(){

		int count = 0;
		String s = "Known hosts:";
		Iterator<Integer> it = levels.keySet().iterator();
		while( it.hasNext() ){
			int levelSpeed = it.next();
			s += "\nlevel [ " + levelSpeed + "kbps ]: ";
			SortedNeighbourList peers = levels.get( levelSpeed );
			NEPNeighbour n = null;
			for( int i = 0; i < peers.size(); i++ ){
				n = peers.elementAt( i );
				s += "\n" + n.getIp().getHostAddress() + ":" + n.getPort();
				if( peers.elementAt( i ).getNodeType() == ProtocolConstants.SOURCE ){
					s += " (SOURCE)";
				}
				count++;
			}
		}
		if( gui != null ){
			gui.updateNodes( s, count );
		}
		if( logger.isWarnEnabled() ){
			if( s.equals( "" ) ){
				logger.warn( "no nodes have joined so far" );
			}else{
				logger.warn( s );
			}
		}
	}


	/**
	 * Find the level to which this <code>NEPNeighbour</code> belongs.
	 * 
	 * @param n
	 *            the <code>NEPNeighbour</code> in question
	 * @return the level (key for <code>levels</code> hashmap) with the
	 *         greatest speed smaller than the node's speed; a return value of 0
	 *         means that there is no level into which this
	 *         <code>NEPNeighbour</code> fits
	 */
	private int findLevel( NEPNeighbour n ){

		if( n.getNodeType() == ProtocolConstants.SOURCE ){
			return n.getLevel();
		}
		// find the level to which this NEPNeighbour belongs
		int nodeSpeed = n.getSpeed();
		int curLevel = 0;
		Iterator<Integer> it = levels.keySet().iterator();
		while( it.hasNext() ){
			int levelSpeed = it.next();
			if( ( levelSpeed * capacityRequired <= nodeSpeed ) && ( levelSpeed > curLevel ) ){
				curLevel = levelSpeed;
			}
		}
		return curLevel;
	}


	/**
	 * Create a new entry into the <code>levels</code> Hashmap. A new empty
	 * <code>SortedNeighboursList</code> value with speed as the key will be
	 * created.
	 * 
	 * @param speed
	 *            the speed of the new level
	 * @return true if a new level was created; false otherwise. This might be
	 *         because the level already exists.
	 */
	public synchronized boolean createLevel( int speed ){

		if( levels.containsKey( speed ) ){
			return false;
		}
		levels.put( speed, new SortedNeighbourList() );
		if( speed < minSpeed ){
			minSpeed = speed;
		}
		writeStatePersistently();
		if( logger.isDebugEnabled() ){
			logger.debug( "new level with minimum speed " + speed + " created" );
		}
		return true;
	}


	/**
	 * Remove the <code>Source</code> of this level and all the nodes it
	 * contains.
	 * 
	 * @param level
	 *            the level which should be removed completely
	 */
	public synchronized void removeLevel( int level ){

		removeSource( level );
		levels.remove( level );
		if( level == minSpeed ){
			// lowest level was removed; update min required speed
			int newMin = Integer.MAX_VALUE;
			Iterator<Integer> it = levels.keySet().iterator();
			while( it.hasNext() ){
				int cur = it.next();
				if( cur < newMin ){
					newMin = cur;
				}
			}
			setMinSpeed( newMin );
		}
	}


	/**
	 * Creates a copy of the existing levels.
	 * 
	 * @return a copy of the existing levels
	 */
	public Hashtable<Integer, Vector<DummyNode>> copyLevels(){

		synchronized( levels ){
			Hashtable<Integer, Vector<DummyNode>> copy = new Hashtable<Integer, Vector<DummyNode>>();
			Iterator<Integer> it = levels.keySet().iterator();
			while( it.hasNext() ){
				int levelSpeed = it.next();
				SortedNeighbourList origPeers = levels.get( levelSpeed );
				int nPeers = origPeers.size();
				Vector<DummyNode> copyPeers = new Vector<DummyNode>( nPeers );
				for( int i = 0; i < nPeers; i++ ){
					Neighbour curN = origPeers.elementAt( i );
					copyPeers.add( new DummyNode( curN.getIp(), curN.getPort() ) );
				}
				copy.put( levelSpeed, copyPeers );
			}
			return copy;
		}
	}


	/**
	 * Write the current state into a file for recovery.
	 */
	private void writeStatePersistently(){

		try{
			persistencyManager.writeCurrentState();
		}catch( PersistencyManagerException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
				if( gui != null ){
					gui.addStatusLine( "[ERROR] " + e.getMessage() );
				}
			}
		}
	}


	/**
	 * Use the state information from the file to set up this
	 * <code>NetworkEntryPoint</code>.
	 * 
	 * @return true on success; false otherwise
	 */
	public boolean recover(){

		try{
			persistencyManager.recoverLastState();
			printNeighbours();
			return true;
		}catch( PersistencyManagerException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( e.getMessage() );
			}
			return false;
		}
	}


	/**
	 * Test if there is a file from which a recory would be possible.
	 * 
	 * @return true if recovery is possible; false otherwise
	 */
	public boolean recoveryPossible(){

		return persistencyManager.recoveryPossible();
	}


	/**
	 * Update the status information in the gui.
	 * 
	 * @param statusLine
	 *            a new line of information that will be added to the status tab
	 */
	public void write2gui( String statusLine ){

		if( gui != null ){
			gui.addStatusLine( statusLine );
		}
	}


	// GETTERS AND SETTERS

	public int getMinSpeed(){

		return minSpeed;
	}


	public int getNInitNeighbours(){

		return nInitNeighbours;
	}


	public void setNInitNeighbours( int initNeighbours ){

		if( initNeighbours > 0 ){
			nInitNeighbours = initNeighbours;
		}
		writeStatePersistently();
	}


	public int getPingTimeout(){

		return pingTimeout;
	}


	public void setPingTimeout( int pingTimeout ){

		if( pingTimeout > 0 ){
			this.pingTimeout = pingTimeout;
		}
	}


	public int getMaxNeighs(){

		return maxNeighs;
	}


	public void setMaxNeighs( int maxNeighs ){

		this.maxNeighs = maxNeighs;
		writeStatePersistently();
	}


	public int getMinNeighs(){

		return minNeighs;
	}


	public void setMinNeighs( int minNeighs ){

		this.minNeighs = minNeighs;
		writeStatePersistently();
	}


	public byte getStreamType(){

		return streamType;
	}


	public void setStreamType( byte streamType ){

		if( this.streamType == ProtocolConstants.UNKNOWN_TYPE ){
			this.streamType = streamType;
			writeStatePersistently();
		}
	}


	public String getStatus(){

		return status;
	}


	public void setStatus( String status ){

		this.status = status;
	}


	public Hashtable<Integer, SortedNeighbourList> getLevels(){

		return levels;
	}


	public void setLevels( Hashtable<Integer, SortedNeighbourList> levels ){

		this.levels = levels;
	}


	public void setMinSpeed( int minSpeed ){

		this.minSpeed = minSpeed;
		writeStatePersistently();
	}


	public float getCapacityRequired(){

		return capacityRequired;
	}


	// methods below are used for statistics...

	private boolean updateAvgSpeeds( int level, int newSpeed ){

		int[] avgSpeed = avgSpeeds.get( level );
		if( avgSpeed == null ){
			if( levels.containsKey( level ) ){
				avgSpeed = new int[2];
				avgSpeed[0] = 1;
				avgSpeed[1] = newSpeed;
				avgSpeeds.put( level, avgSpeed );
				return true;
			}else{
				return false;
			}
		}else{
			avgSpeed[1] += newSpeed;
			avgSpeed[0] += 1;
			return true;
		}
	}


	public float getAvgSpeedForLevel( int level ){

		if( avgSpeeds.containsKey( level ) ){
			int[] avgSpeed = avgSpeeds.get( level );
			return (float)avgSpeed[1] / (float)avgSpeed[0];
		}else{
			return 0;
		}
	}


	public float getAvgOverallSpeed(){

		float sum = 0;
		int count = 0;
		Iterator<int[]> it = avgSpeeds.values().iterator();
		while( it.hasNext() ){
			int[] cur = it.next();
			sum += cur[1];
			count += cur[0];
		}
		return (float)sum / (float)count;
	}
}
