/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.peers;

import java.net.InetAddress;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.dcg.p2p.exceptions.DispatcherException;
import ch.ethz.dcg.p2p.exceptions.DispatcherNotRunningException;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.exceptions.UDPSocketManagerException;
import ch.ethz.dcg.p2p.gui.GUI;
import ch.ethz.dcg.p2p.packets.Notification;
import ch.ethz.dcg.p2p.packets.P2PPacket;
import ch.ethz.dcg.p2p.tasks.ByeTask;
import ch.ethz.dcg.p2p.tasks.GetNewNeighboursTask;
import ch.ethz.dcg.p2p.tasks.JoinTask;
import ch.ethz.dcg.p2p.tasks.NotificationTask;
import ch.ethz.dcg.p2p.tasks.RequestTask;
import ch.ethz.dcg.p2p.tasks.ZombieKiller;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.Statistic;
import ch.ethz.dcg.p2p.utilities.UDPSocketManager;

/**
 * Base class for the <code>Peer</code> and the <code>Source</code>.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see Peer
 * @see Source
 */
public abstract class Node {

	protected final Log logger = LogFactory.getLog( this.getClass() );

	// status
	public static final String idle = "idle";
	public static final String configured = "configured";
	public static final String connected = "connected";
	public static final String started = "started";
	public static final String error = "error";
	protected String status = idle;
	protected String oldStatus = idle;

	// settings
	protected InetAddress ip;
	protected int port;
	protected int inSpeed; // (Kbit/s)
	protected int outSpeed; // (Kbit/s)
	protected int minNeighs = 2;
	protected int maxNeighs = 6;
	protected InetAddress nepAddress;

	// neighbours
	protected Hashtable<String, Neighbour> neighbours = new Hashtable<String, Neighbour>();
	protected ZombieKiller zombieKiller;

	// packet stuff
	protected long packetBufferSize = -1;
	protected long lowestPacketId = Long.MAX_VALUE;
	protected Hashtable<Long, P2PPacket> availablePackets = new Hashtable<Long, P2PPacket>();
	protected Hashtable<Long, Notification> notifications = new Hashtable<Long, Notification>();
	protected Hashtable<Long, Notification> requests = new Hashtable<Long, Notification>();
	protected RequestTask requestTask;
	// set in gui or with INITNEIGHBOURS
	protected byte streamType = ProtocolConstants.UNKNOWN_TYPE;
	// set in gui or with ACCEPT or REDIRECT
	protected int packetSize = -1;
	// set in gui or with ACCEPT or REDIRECT
	protected int bitrate = -1;
	// set by DePacketizer
	protected long startPacketId = -1;
	// set by DePacketizer
	protected long startTime = -1;
	protected long curPacketId = -1;

	private Object lock = new Object();
	private Object notificationLock = new Object();

	// dispatchers
	protected TCPDispatcher tcpDispatcher;
	protected UDPDispatcher udpDispatcher;

	// udp socket manager
	protected UDPSocketManager socketManager;

	// flood id 'memory'
	protected Vector<Integer> floodIds = new Vector<Integer>();

	// gui
	protected GUI gui;

	// statistic
	public static Statistic statistic = new Statistic();


	/**
	 * Creates a new <code>Node</code>.<br>
	 * This constructor is called when run in gui mode.
	 * 
	 * @param gui
	 *            the gui object that represents this <code>Node</code>
	 */
	public Node( GUI gui ){

		this.gui = gui;
	}


	/**
	 * Creates a new <code>Node</code>.<br>
	 * This constructor is called when run in console mode.
	 * 
	 * @param ip
	 *            IP address of this <code>Node</code>
	 * @param port
	 *            port number of this <code>Node</code>
	 * @param inSpeed
	 *            inbound bandwidth
	 * @param outSpeed
	 *            outbound bandwidth
	 * @param nepAddress
	 *            IP address of the <code>NetworkEntryPoint</code> which
	 *            manages this network
	 */
	public Node( InetAddress ip, int port, int inSpeed, int outSpeed, InetAddress nepAddress ){

		configure( ip, port, inSpeed, outSpeed, nepAddress );
	}


	/**
	 * Sets the parameters that are not set by the constructor when run in gui
	 * mode.
	 * 
	 * @param ip
	 *            IP address of this <code>Node</code>
	 * @param port
	 *            port number of this <code>Node</code>
	 * @param inSpeed
	 *            inbound bandwidth
	 * @param outSpeed
	 *            outbound bandwidth
	 * @param nepAddress
	 *            IP address of the <code>NetworkEntryPoint</code> which
	 *            manages this network
	 */
	public void configure( InetAddress ip, int port, int inSpeed, int outSpeed, InetAddress nepAddress ){

		this.ip = ip;
		this.port = port;
		this.inSpeed = inSpeed;
		this.outSpeed = outSpeed;
		this.nepAddress = nepAddress;
		setStatus( Node.configured );
	}


	/**
	 * Starts and initialises the UDP and TCP dispatchers as well as the
	 * UDPSocketManager.
	 * 
	 * @throws DispatcherNotRunningException
	 *             if there is a problem creating and starting one of the
	 *             dispatchers; in this case the node will not be able to
	 *             correctly receive any messages
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 * @throws UDPSocketManagerException
	 *             if the UDPSocketManager could not successfully open a UDP
	 *             socket
	 */
	public void startDispatchers() throws DispatcherNotRunningException, NodeNotConfiguredException, UDPSocketManagerException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot start Dispatcher on unconfigured node" );
		}
		try{
			tcpDispatcher = new TCPDispatcher( this );
			tcpDispatcher.start();
			udpDispatcher = new UDPDispatcher( this );
			socketManager = new UDPSocketManager( port, udpDispatcher );
			socketManager.start();
		}catch( DispatcherException e ){
			throw new DispatcherNotRunningException( e.getMessage() );
		}catch( SocketException e ){
			throw new UDPSocketManagerException( e.getMessage() );
		}
	}


	/**
	 * Starts and initializes a thread that sends requests upon receiving
	 * <code>Notification</code>s.
	 * 
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 * 
	 */
	public void startRequestTask() throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot start RequestTask on unconfigured node" );
		}
		requestTask = new RequestTask( this, notifications );
		new Thread( requestTask, "RequestTask" ).start();
	}


	/**
	 * Connect to the <code>NetworkEntryPoint</code> and send him a JOIN
	 * message. The <code>NetworkEntryPoint</code> will respond with a list of
	 * initial <code>Neighbour</code> to which HELLO messages will be sent.
	 * 
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public void joinNetwork() throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot start JoinTask on unconfigured node" );
		}
		JoinTask jt;
		if( this instanceof Source ){
			jt = new JoinTask( this, nepAddress, outSpeed, ProtocolConstants.SOURCE, streamType );
		}else{
			jt = new JoinTask( this, nepAddress, outSpeed, ProtocolConstants.PEER );
		}
		new Thread( jt, "JoinTask" ).start();
	}


	/**
	 * Leave the network to which this node is currently connected. A
	 * <code>ByeTask</code> will be started to inform the existing neighbours
	 * and the Network Entry Point of my disappearance.
	 * 
	 */
	public void leaveNetwork(){

		new Thread( new ByeTask( this ), "ByeTask" ).start();
		try{
			Thread.sleep( 3000 );
		}catch( InterruptedException e ){
		}
		resetNode();
	}


	/**
	 * Unleashes the <code>ZombieKiller</code> which is a thread that runs in
	 * a given interval and sends PING messages to all <code>Neighbour</code>s.
	 * 
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public void unleashZombieKiller() throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot unleash ZombieKiller on unconfigured node" );
		}
		zombieKiller = new ZombieKiller( 30000, 3, this );
		new Thread( zombieKiller, "ZombieKiller" ).start();
	}


	/**
	 * Adds a <code>Neighbour</code> to the list of current neighbours if it
	 * is not already there; otherwise this operation will have no effect.
	 * 
	 * @param n
	 *            <code>Neighbour</code> to be added
	 * @return true if the Neighbour n was added; false otherwise
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public boolean addNeighbour( Neighbour n ) throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot add neighbour to unconfigured node" );
		}
		synchronized( neighbours ){
			// test if the neighbour is this node
			if( n.getIp().equals( ip ) && n.getPort() == port ){
				if( logger.isDebugEnabled() ){
					logger.debug( "will not add myself as my neighbour" );
				}
				return false;
			}
			// create hash key and test if the neighbour is already known
			String nKey = n.getIp().getHostAddress() + ":" + n.getPort();
			if( !neighbours.containsKey( nKey ) ){
				neighbours.put( nKey, n );
				statistic.addNeighbourStat( nKey );
				if( logger.isDebugEnabled() ){
					logger.debug( "node " + nKey + " added to neighbours" );
				}
				return true;
			}else{
				if( logger.isDebugEnabled() ){
					logger.debug( "node " + nKey + " is already a neighbour and will not be added" );
				}
				return false;
			}
		}
	}


	/**
	 * Removes the <code>Neighbour</code> with the given IP and Port from the
	 * current set of <code>Neighbour</code>s.
	 * 
	 * @param ip
	 *            IP of the <code>Neighbour</code> to be removed
	 * @param port
	 *            port of the <code>Neighbour</code> to be removed
	 * @return true if this ip:port combination represents an existing
	 *         <code>Neighbour</code> which has now been successfully removed;
	 *         false otherwise
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public boolean removeNeighbour( InetAddress ip, int port ) throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot remove neighbour from unconfigured node" );
		}
		String nKey = ip.getHostAddress() + ":" + port;
		synchronized( neighbours ){
			// remove the corresponding neighbour if it exists
			if( neighbours.containsKey( nKey ) ){
				neighbours.remove( nKey );
				statistic.removeNeighbourStat( nKey );
				if( logger.isDebugEnabled() ){
					logger.debug( "neighbour " + nKey + " removed" );
				}
				if( !hasEnoughNeighbours() ){
					new Thread( new GetNewNeighboursTask( this, ip, port ), "GetNewNeighboursTask" ).start();
				}
				return true;
			}else{
				return false;
			}
		}
	}


	/**
	 * Removes an existing <code>Neighbour</code>. If this neighbour does not
	 * exist this operation will have no effect.
	 * 
	 * @param n
	 *            <code>Neighbour</code> to be removed
	 * @return true if the Neighbour was present and has been successfully
	 *         removed; false otherwise
	 * @throws NodeNotConfiguredException
	 */
	public boolean removeNeighbour( Neighbour n ) throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot remove neighbour from unconfigured node" );
		}
		String nKey = n.getIp().getHostAddress() + ":" + n.getPort();
		synchronized( neighbours ){
			if( neighbours.containsKey( nKey ) ){
				neighbours.remove( nKey );
				statistic.removeNeighbourStat( nKey );
				if( logger.isDebugEnabled() ){
					logger.debug( "neighbour " + nKey + " removed" );
				}
				return true;
			}else{
				return false;
			}
		}
	}


	/**
	 * Removes the worst <code>Neighbour</code> of this <code>Node</code>.
	 * The worst <code>Neighbour</code> is the one with the lowest ranking. If
	 * there are two with the lowest ranking, the oldest will be removed.
	 * 
	 * @return the worst <code>Neighbour</code> or null if no
	 *         <code>Neighbour</code> is present
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public Neighbour removeWorstNeighbour() throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot remove worst neighbour from unconfigured node" );
		}
		synchronized( neighbours ){
			Iterator<Neighbour> it = neighbours.values().iterator();
			// iterate over neighbours and find the one with the worst ranking
			Neighbour cur = null;
			Neighbour worst = null;
			int worstRanking = Integer.MAX_VALUE;
			while( it.hasNext() ){
				cur = it.next();
				if( cur.getRanking() < worstRanking || ( ( cur.getRanking() == worstRanking ) && ( cur.getKnownSince() < worst.getKnownSince() ) ) ){
					// new worst found
					worstRanking = cur.getRanking();
					worst = cur;
				}
			}
			if( worst != null ){
				String worstKey = worst.getIp().getHostAddress() + ":" + worst.getPort();
				neighbours.remove( worstKey );
				statistic.removeNeighbourStat( worstKey );
				if( logger.isDebugEnabled() ){
					logger.debug( "node " + worst.getIp().getHostAddress() + " removed because he is my worst neighbour" );
				}
			}
			return worst;
		}
	}


	/**
	 * Checks if this <code>Node</code> is allowed to accept another
	 * neighbour.
	 * 
	 * @return true if the number of neighbours is smaller than the maximum
	 *         allowed; false otherwise
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public boolean hasSpareCapacity() throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "node not configured" );
		}
		synchronized( neighbours ){
			return neighbours.size() < maxNeighs;
		}
	}


	/**
	 * Checks if the source has still at least the minimum amount of required
	 * neighbours.
	 * 
	 * @return true if the amount of current neighbours is greater or equal to
	 *         the minimum required; false otherwise
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public boolean hasEnoughNeighbours() throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "node not configured" );
		}
		synchronized( neighbours ){
			return neighbours.size() >= minNeighs;
		}
	}


	/**
	 * Tests if the host with given IP address and port number is among my
	 * current neighbours.
	 * 
	 * @param ip
	 *            IP address of the host in question
	 * @param port
	 *            port of the host in question
	 * @return true if a host with the given ip address and port is in the set
	 *         of my neighbours; false otherwise
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public boolean hasNeighbour( InetAddress ip, int port ) throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "node not configured" );
		}else{
			String nKey = ip.getHostAddress() + ":" + port;
			synchronized( neighbours ){
				return neighbours.containsKey( nKey );
			}
		}
	}


	/**
	 * Adds a <code>P2PPacket</code> to the available packets. A notification
	 * for this packet is sent immediately using a <code>NotificationTask</code>.
	 * The number of packets kept in memory is defined with the packetBufferSize
	 * field.
	 * 
	 * @param packet
	 *            the new <code>P2PPacket</code>
	 * @param receptionTime
	 *            the time stamp when this <code>P2PPacket</code> was received
	 *            at the <code>PacketHandler</code>
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public void addPacket( P2PPacket packet, long receptionTime ) throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot add packet to unconfigured node" );
		}
		// add new packet (clear some space if necessary)
		synchronized( availablePackets ){
			while( availablePackets.size() >= packetBufferSize ){
				availablePackets.remove( lowestPacketId );
				lowestPacketId++;
			}
			if( packet.getId() < lowestPacketId ){
				lowestPacketId = packet.getId();
			}
			availablePackets.put( packet.getId(), packet );
		}
		// remove notification from requests
		Notification n = null;
		synchronized( notificationLock ){
			if( requests.containsKey( packet.getId() ) ){
				n = requests.remove( packet.getId() );
			}else if( notifications.containsKey( packet.getId() ) ){
				n = notifications.remove( packet.getId() );
			}
		}
		// update RTT of the neighbour that has sent this packet
		if( !( this instanceof Source ) && n != null ){
			synchronized( neighbours ){
				int newRTT = neighbours.get( packet.getReceivedFromKey() ).updateRTT( (int)( receptionTime - n.getStartTime() ) );
				statistic.setRTTOf( packet.getReceivedFromKey(), newRTT );
			}
		}
		if( logger.isDebugEnabled() ){
			logger.debug( "packet " + packet.getId() + " added" );
		}
		// update statistics
		if( statistic.getPacketSize() < 0 ){
			statistic.setPacketSize( packet.getSize() );
		}
		statistic.incPacketsReceivedFrom( packet.getReceivedFromKey() );
		new Thread( new NotificationTask( this, packet.getId() ), "NotificationTask" ).start();
	}


	/**
	 * Gets the packet with the corresponding id.
	 * 
	 * @param packetId
	 *            identifier of the desired packet
	 * @return the packet which has the specified id or null if the packet does
	 *         not exist
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public P2PPacket getPacket( long packetId ) throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot get packet from unconfigured node" );
		}
		synchronized( availablePackets ){
			return availablePackets.get( packetId );
		}
	}


	/**
	 * Checks if the packet with the given packet id is available.
	 * 
	 * @param packetId
	 *            identifier of the desired packet
	 * @return true if the packet is available; false otherwise
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public boolean hasPacket( long packetId ) throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "node not configured" );
		}
		return availablePackets.containsKey( packetId );
	}


	/**
	 * Adds a notification for a packet to the hashtable of received
	 * notifications.<br>
	 * If a notification for this packetId already exists, the new holder will
	 * be added to the list of currently known holders for that packetId. If the
	 * holder is already known, this method will have no effect.<br>
	 * If the holder who announces this packet is not part of my neighbours the
	 * notification will be dropped.<br>
	 * If I already have the packet for which a notifications should be added
	 * the notification will be dropped.
	 * 
	 * @param packetId
	 *            id of the packet that is announced
	 * @param holderIp
	 *            the ip of the holder of the announced packet
	 * @param holderPort
	 *            the port of the holder of the announced packet
	 * @return true if the notification has been added; false otherwise
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public boolean addNotification( long packetId, InetAddress holderIp, int holderPort ) throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot add notification to unconfigured node" );
		}
		if( hasPacket( packetId ) ){
			return false;
		}
		if( packetId < curPacketId ){
			return false;
		}
		// check if the notification has come from a known neighbour
		Neighbour tmpNeigh = getNeighbour( holderIp, holderPort );
		if( tmpNeigh == null ){
			if( logger.isWarnEnabled() ){
				logger.warn( "will not add notification from unknown neighbour " + holderIp.getHostAddress() + ":" + holderPort );
			}
			write2gui( "NOTIFICATION from " + holderIp.getHostAddress() + ":" + holderPort + " received which is NOT my neighbour" );
			statistic.incUnknownNotificationsReceived();
			return false;
		}
		// lock notifications and requests
		synchronized( notificationLock ){
			// check if a notification for this packetId already exists
			if( notifications.containsKey( packetId ) || requests.containsKey( packetId ) ){
				// try to get Notification from the notifications Hashtable
				Notification n = notifications.get( packetId );
				if( n == null ){
					// Notification is in the Hashtable of requested packets
					n = requests.get( packetId );
				}
				statistic.incDuplicateNotificationsReceived();
				// add new node as a new holder of the packet if not already
				// there
				if( n.addHolder( tmpNeigh ) ){
					if( logger.isDebugEnabled() ){
						logger.debug( holderIp.getHostAddress() + ":" + holderPort + " added as holder of packet " + packetId );
					}
					return true;
				}else{
					if( logger.isDebugEnabled() ){
						logger.debug( holderIp.getHostAddress() + ":" + holderPort + " already known as holder of packet " + packetId );
					}
					return false;
				}
			}
			// no notification for this id present -> create new one if packet
			// not already available or requested
			notifications.put( packetId, new Notification( packetId, tmpNeigh ) );
		}
		statistic.incNotificationsReceivedFrom( holderIp.getHostAddress() + ":" + holderPort );
		if( logger.isDebugEnabled() ){
			logger.debug( "notification for packet " + packetId + " from " + holderIp.getHostAddress() + ":" + holderPort + " added" );
		}
		// wake up the RequestTask
		if( requestTask != null ){
			// the source does not have a request task
			synchronized( requestTask ){
				requestTask.notify();
			}
		}
		return true;
	}


	/**
	 * Moves a <code>Notification</code> from the list of remaining
	 * <code>Notification</code>s to the list of <code>Notification</code>s
	 * for which a request has been sent.
	 * 
	 * @param packetId
	 *            the key of the <code>Notification</code> which should be
	 *            moved
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public void moveToRequested( long packetId ) throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "node not configured" );
		}
		synchronized( notificationLock ){
			if( notifications.containsKey( packetId ) ){
				requests.put( packetId, notifications.remove( packetId ) );
			}
		}
	}


	/**
	 * Moves a <code>Notification</code> from the list of requested packets to
	 * the list of <code>Notification</code>s for which a request still has
	 * to be sent.
	 * 
	 * @param packetId
	 *            the key of the <code>Notification</code> which should be
	 *            moved
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public void moveToNotifications( long packetId ) throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "node not configured" );
		}
		synchronized( notificationLock ){
			if( requests.containsKey( packetId ) ){
				notifications.put( packetId, requests.remove( packetId ) );
			}
		}
		// wake up the RequestTask
		if( requestTask != null ){
			// the source does not have a request task
			synchronized( requestTask ){
				requestTask.notify();
			}
		}
	}


	/**
	 * Tests if this packet id is in the list of requested packets.
	 * 
	 * @param packetId
	 *            the id of the packet in question
	 * @return true if the packet is in the list of requested packets; false
	 *         otherwise
	 */
	public boolean hasRequestedPacket( long packetId ){

		synchronized( notificationLock ){
			return requests.containsKey( packetId );
		}
	}


	/**
	 * Creates a copy of the set of <code>Neighbour</code>s. This is the only
	 * way another class can get the set of <code>Neighbour</code>s. There
	 * will be no concurrent modification exceptions if it done like this.
	 * 
	 * @return a copy of the set of current <code>Neighbour</code>s
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public ArrayList<Neighbour> getNeighboursCopy() throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot get neighbours copy on unconfigured node" );
		}
		synchronized( neighbours ){
			return new ArrayList<Neighbour>( neighbours.values() );
		}
	}


	/**
	 * Gets the number of currently available <code>Neighbour</code>s.
	 * 
	 * @return the number of currently available <code>Neighbour</code>s
	 */
	public int getNNeighbours(){

		if( !isConfigured() ){
			return 0;
		}else{
			synchronized( neighbours ){
				return neighbours.size();
			}
		}
	}


	/**
	 * Gets the <code>Neighbour</code> object that represents a neighbour with
	 * the given ip and port combination.
	 * 
	 * @param ip
	 *            IP address of the requested <code>Neighbour</code>
	 * @param port
	 *            port number of the requested <code>Neighbour</code>
	 * @return the <code>Neighbour</code> object with the given ip and port or
	 *         null if it does not exist
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public Neighbour getNeighbour( InetAddress ip, int port ) throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot get neighbour from unconfigured node" );
		}
		synchronized( neighbours ){
			String nKey = ip.getHostAddress() + ":" + port;
			return neighbours.get( nKey );
		}
	}


	/**
	 * Increase the ranking of a neighbour.
	 * 
	 * @param nKey
	 *            the IP:port combination of the peer to honor
	 */
	public void honorPeer( String nKey ){

		Neighbour n;
		synchronized( neighbours ){
			n = neighbours.get( nKey );
		}
		if( n != null ){
			n.increaseRanking();
			if( gui != null ){
				gui.updateNeighbourRanking( nKey, n.getRanking() );
			}
		}
	}


	/**
	 * Decrease the ranking of a neighbour.
	 * 
	 * @param missingPacket
	 *            the IP:port combination of the peer to blame
	 */
	public void blamePeer( long missingPacket ){

		Notification notify = null;
		synchronized( notificationLock ){
			if( requests.containsKey( missingPacket ) ){
				// packet was requested but not delivered
				notify = requests.get( missingPacket );
			}else{
				return;
			}
		}
		Neighbour n = notify.getProvider();
		if( n != null ){
			n.decreaseRanking();
			if( gui != null ){
				gui.updateNeighbourRanking( n.getIp().getHostAddress() + ":" + n.getPort(), n.getRanking() );
			}
		}
	}


	/**
	 * Prints of the currently known neighbours using the logging mechanism. If
	 * the log level is above DEBUG this method will have no effect.
	 */
	public void printNeighbours(){

		if( logger.isDebugEnabled() ){
			String s = "\nNeighbours of " + ip.getHostAddress() + ":" + port + " :";
			if( neighbours.isEmpty() ){
				s += "\nno neighbours available";
			}else{
				try{
					for( Neighbour curN : getNeighboursCopy() ){
						s += "\n" + curN.getIp().getHostAddress() + ":" + curN.getPort();
					}
				}catch( NodeNotConfiguredException e ){
					s += "\nno neighbours available (node not configured)";
				}
			}
			logger.debug( s );
		}
	}


	/**
	 * Tests if this <code>Node</code> is in configured state or further.
	 * 
	 * @return true if this <code>Node</code> is in one of the following
	 *         states: configured, connected, error or started
	 */
	public boolean isConfigured(){

		return ( status.equalsIgnoreCase( configured ) || status.equalsIgnoreCase( connected ) || status.equalsIgnoreCase( error ) || status.equalsIgnoreCase( started ) );
	}


	/**
	 * Resets this node.<br>
	 * ZombieKiller will be stopped, UDPSocketManager will be closed, the
	 * statistics will be reset etc.
	 */
	private void resetNode(){

		// stop ZombieKiller
		if( zombieKiller != null ){
			zombieKiller.close();
		}
		// stop UDPSocketManager
		if( socketManager != null ){
			socketManager.close();
		}

		// reset statistic
		statistic.reset();

		// stop dispatcher
		tcpDispatcher.close();

		// closing request task
		if( requestTask != null ){
			// source does not have a RequestTask
			requestTask.close();
		}

		// reset neighbours
		neighbours = new Hashtable<String, Neighbour>();
		// reset flood messages
		floodIds = new Vector<Integer>();

		// reset packets, requests and notifications
		lowestPacketId = Long.MAX_VALUE;
		availablePackets = new Hashtable<Long, P2PPacket>();
		notifications = new Hashtable<Long, Notification>();
		requests = new Hashtable<Long, Notification>();

		// reset stream
		streamType = ProtocolConstants.UNKNOWN_TYPE;

		packetSize = -1;
		bitrate = -1;
		startPacketId = -1;
		startTime = -1;
		curPacketId = -1;
	}


	/**
	 * Add a new line of text to the status tab in the gui.
	 * 
	 * @param statusLine
	 *            a new line of information that will be added to the status tab
	 */
	public void write2gui( String statusLine ){

		if( gui != null ){
			gui.addStatusLine( statusLine );
		}
	}


	/**
	 * Show an error dialog message in the gui.
	 * 
	 * @param message
	 *            error message
	 * @param title
	 *            title of the dialog
	 */
	public void errorDialog( String message, String title ){

		if( gui != null ){
			gui.errorDialog( message, title );
		}
	}


	/**
	 * Checks if the given flood id has already been seen.
	 * 
	 * @param id
	 *            the flood id in question
	 * @return true if this flood id has been seen before; false otherwise
	 * @throws NodeNotConfiguredException
	 *             if the node is not in configured state
	 */
	public boolean seenFloodId( int id ) throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "node not configured" );
		}
		synchronized( floodIds ){
			return floodIds.contains( id );
		}
	}


	/**
	 * Adds an id to set of seen flood ids.
	 * 
	 * @param id
	 *            the new flood id that has been received
	 */
	public void addFloodId( int id ){

		synchronized( floodIds ){
			if( !floodIds.contains( id ) ){
				floodIds.add( id );
			}
		}
	}


	// GETTERS AND SETTERS

	public InetAddress getIp(){

		return ip;
	}


	public int getPort(){

		return port;
	}


	public long getPacketBufferSize(){

		return packetBufferSize;
	}


	public InetAddress getNepAddress(){

		return nepAddress;
	}


	public int getMinNeighs(){

		return minNeighs;
	}


	public void setMinNeighs( int minNeighs ){

		this.minNeighs = minNeighs;
	}


	public int getMaxNeighs(){

		return maxNeighs;
	}


	public void setMaxNeighs( int maxNeighs ){

		this.maxNeighs = maxNeighs;
	}


	public byte getStreamType(){

		return streamType;
	}


	public void setStreamType( byte streamType ){

		this.streamType = streamType;
	}


	public String getStatus(){

		return status;
	}


	public String getOldStatus(){

		return oldStatus;
	}


	public void setStatus( String status ){

		oldStatus = this.status;
		this.status = status;
		if( gui != null ){
			gui.updateStatus();
		}
	}


	public boolean inStatus( String status ){

		return this.status.equals( status );
	}


	public long getCurPacketId(){

		synchronized( lock ){
			return curPacketId;
		}
	}


	public void setCurPacketId( long curPacketId ){

		synchronized( lock ){
			this.curPacketId = curPacketId;
		}
		if( gui != null ){
			gui.updateTotals();
		}
	}


	public UDPSocketManager getSocketManager(){

		return socketManager;
	}


	public int getBitrate(){

		return bitrate;
	}


	public void setBitrate( int bitrate ){

		this.bitrate = bitrate;
		write2gui( bitrate + "kbps stream" );
	}


	public int getPacketSize(){

		return packetSize;
	}


	public void setStartPacketId( long startPacketId ){

		this.startPacketId = startPacketId;
	}


	public abstract int setPacketSize( int packetSize );


	public long getStartTime(){

		return startTime;
	}


	public void setStartTime( long startTime ){

		this.startTime = startTime;
	}
}
