/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.peers;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Vector;

import ch.ethz.dcg.p2p.adapters.output.Mp3Adapter;
import ch.ethz.dcg.p2p.adapters.output.OutputAdapter;
import ch.ethz.dcg.p2p.adapters.output.VlcAdapter;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.gui.PeerGUI;
import ch.ethz.dcg.p2p.packets.DePacketizer;
import ch.ethz.dcg.p2p.packets.P2PPacket;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * Implementation of a peer which is a sink in the Swistry network.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class Peer extends Node {

	private Vector<OutputAdapter> adapters = new Vector<OutputAdapter>();
	private DePacketizer depacketizer;
	private int dePacketizerStartThreshold;
	private int mediaPort;


	/**
	 * Creates a new peer. This constructor is usually called when the peer is
	 * created using the console mode.
	 * 
	 * @param ip
	 *            this peers IP
	 * @param port
	 *            this peers port
	 * @param inSpeed
	 *            this peers inbound speed (KBit/s)
	 * @param outSpeed
	 *            this peers outbound speed (KBit/s)
	 * @param nepAddress
	 *            the IP of the <code>NetworkEntryPoint</code> of the network
	 *            this peers wants to join
	 * @param dePacketizerStartThreshold
	 *            the time (in seconds) that need to be available in the buffer
	 *            before the peer starts to forward data to the media player
	 * @throws UnknownHostException
	 *             if the IP address or the nepAddress are invalid
	 */
	public Peer( String ip, int port, int inSpeed, int outSpeed, String nepAddress, int dePacketizerStartThreshold ) throws UnknownHostException{

		super( InetAddress.getByName( ip ), port, inSpeed, outSpeed, InetAddress.getByName( nepAddress ) );
		this.dePacketizerStartThreshold = dePacketizerStartThreshold;
	}


	/**
	 * Create a new <code>Peer</code>. This constructor is invoked by the
	 * GUI. Thereafter the Peer needs to be configured before it can be started.
	 * 
	 * @param gui
	 *            the gui that represents this <code>Peer</code>
	 */
	public Peer( PeerGUI gui ){

		super( gui );
	}


	/**
	 * Sets the parameters of this <code>Peer</code>. This method is used
	 * when the <code>Peer</code> has been created in gui mode.
	 * 
	 * @param ip
	 *            this peers IP
	 * @param port
	 *            this peers port
	 * @param inSpeed
	 *            this peers inbound speed (KBit/s)
	 * @param outSpeed
	 *            this peers outbound speed (KBit/s)
	 * @param nepAddress
	 *            the IP of the <code>NetworkEntryPoint</code> of the network
	 *            this peers wants to join
	 * @param dePacketizerStartThreshold
	 *            the time (in seconds) that need to be available in the buffer
	 *            before the peer starts to forward data to the media player
	 */
	public void configure( InetAddress ip, int port, int inSpeed, int outSpeed, InetAddress nepAddress, int dePacketizerStartThreshold ){

		super.configure( ip, port, inSpeed, outSpeed, nepAddress );
		this.dePacketizerStartThreshold = dePacketizerStartThreshold;
		write2gui( "configured" );
	}


	/**
	 * Initialises and starts the <code>DePacketizer</code> with a pointer to
	 * the available packets and with the specified start threshold.<br>
	 * The <code>DePacketizer</code> is run as a thread.
	 * 
	 * @throws NodeNotConfiguredException
	 *             if the peer is not configured
	 * @see DePacketizer
	 */
	public void initDePacketizer() throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot init DePacketizer on unconfigured node" );
		}
		if( depacketizer == null ){
			depacketizer = new DePacketizer( availablePackets, dePacketizerStartThreshold, this );
			new Thread( depacketizer, "DePacketizer" ).start();
		}else{
			write2gui( "[ERROR] DePacketizer instance present; cannot restart it" );
		}
	}


	/**
	 * Adds a <code>P2PPacket</code> to the available packets. A notification
	 * for this packet is sent immediately using a <code>NotificationTask</code>.
	 * 
	 * @param packet
	 *            the new packet
	 * @throws NodeNotConfiguredException
	 *             if the peer is not configured
	 */
	public void addPacket( P2PPacket packet, long receptionTime ) throws NodeNotConfiguredException{

		super.addPacket( packet, receptionTime );
		if( depacketizer != null ){
			synchronized( depacketizer ){
				depacketizer.notify();
			}
		}
	}


	/**
	 * Initialises <code>OutputAdapter</code> that matches the stream type.
	 * 
	 * @throws NodeNotConfiguredException
	 *             if the peer is not configured
	 * @see OutputAdapter
	 */
	public void initOutputAdapter() throws NodeNotConfiguredException{

		if( !isConfigured() ){
			throw new NodeNotConfiguredException( "cannot init OutputAdapter on unconfigured node" );
		}
		while( streamType == ProtocolConstants.UNKNOWN_TYPE ){
			synchronized( this ){
				try{
					wait();
				}catch( InterruptedException e ){
				}
			}
		}
		switch( streamType ){
		case ProtocolConstants.HTTP_MP3:
			addOutputAdapter( new Mp3Adapter( mediaPort, this ) );
			write2gui( "mp3 adapter started at port " + mediaPort );
			break;
		case ProtocolConstants.HTTP_VLC:
			addOutputAdapter( new VlcAdapter( mediaPort, 50, this ) );
			write2gui( "video adapter started at port " + mediaPort );
			break;
		default:
			write2gui( "no appropriate adapter found" );
			break;
		}
	}


	/**
	 * Get a list of the currently set <code>OutputAdapters</code>.
	 * 
	 * @return the currently connected <code>OutputAdapters</code>
	 */
	public Vector<OutputAdapter> getOutputAdapters(){

		return adapters;
	}


	/**
	 * Adds an <code>OutputAdapters</code> to ones that are already connected.
	 * 
	 * @param adapter
	 *            the new <code>OutputAdapters</code>
	 */
	public void addOutputAdapter( OutputAdapter adapter ){

		// make sure that we do not have the same outputadapter twice
		// determine the type of the new adapter
		//TODO:
		/**
		 * Set<Class<? extends OutputAdapter>> adapterTypes = new HashSet<...>();
		addAdapter(OutputAdapter adapter) {
		    if (!adapteryTypes.contains(adapter.getClass())) {
		        // do it
		        adapterTypes.add(adapter.getClass());
		    }
		}
		  15:03:05: btw: besser es Set statt e list mache, das esch no effizinter ond wederschpieglet dini semantik besser ;-)
		  **/
		for( OutputAdapter curA : adapters ){
			if( curA.equals( adapter ) ){
				return;
			}
		}
		adapters.add( adapter );
		depacketizer.addDePacketizerListener( adapter );
	}


	/**
	 * Removes a currently connected <code>OutputAdapters</code>.
	 * 
	 * @param adapter
	 *            the <code>OutputAdapters</code> to be removed
	 */
	public void removeOutputAdapter( OutputAdapter adapter ){

		adapters.remove( adapter );
		depacketizer.removeDePacketizerListener( adapter );
	}


	/**
	 * Clears the list of connected <code>OutputAdapters</code>s.
	 */
	public void removeAllOutputAdapters(){

		for( OutputAdapter adapter : adapters ){
			adapter.close();
			depacketizer.removeDePacketizerListener( adapter );
		}
		adapters.removeAllElements();
	}


	/**
	 * Leave the network to which this <code>Peer</code> is currently
	 * connected.
	 */
	public void leaveNetwork(){

		super.leaveNetwork();
		// reset adapters
		removeAllOutputAdapters();
		// close depacketizer
		if( depacketizer != null ){
			depacketizer.close();
			depacketizer = null;
		}
	}


	// GETTERS AND SETTERS

	public int getMediaPort(){

		return mediaPort;
	}


	public void setMediaPort( int mediaPort ){

		this.mediaPort = mediaPort;
	}


	public int setPacketSize( int packetSize ){

		// set the packetSize only once
		if( packetSize != -1 ){
			this.packetSize = packetSize;
			write2gui( packetSize + "byte packets" );
			packetBufferSize = Runtime.getRuntime().maxMemory() / ( 2 * packetSize );
		}
		return packetSize;
	}
}
