/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.peers;

import java.net.InetAddress;
import java.net.UnknownHostException;

import ch.ethz.dcg.p2p.adapters.input.InputAdapter;
import ch.ethz.dcg.p2p.gui.SourceGUI;
import ch.ethz.dcg.p2p.packets.Packetizer;

/**
 * Implementation of the media source of the Swistry network.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class Source extends Node {

	private Packetizer packetizer;
	private InputAdapter adapter;


	/**
	 * Creates a new <code>Source</code>. This method is called when the
	 * <code>Source</code> is created using the console mode.
	 * 
	 * @param ip
	 *            this <code>Source</code>s IP
	 * @param port
	 *            zhis <code>Source</code>s port
	 * @param streamBitRate
	 *            bit rate of the stream this <code>Source</code> is sending
	 * @param packetSize
	 *            size in bytes of a single packet
	 * @param nepAddress
	 *            IP address of the <code>NetworkEntryPoint</code> of the
	 *            network this <code>Source</code> wants to join
	 * @param streamType
	 *            type of the stream sent by this <code>Source</code> (see
	 *            ProtocolConstants)
	 * @throws UnknownHostException
	 *             if the <code>Source</code> IP or nepAddress are invalid
	 */
	public Source( String ip, int port, int streamBitRate, int packetSize, String nepAddress, byte streamType ) throws UnknownHostException{

		super( InetAddress.getByName( ip ), port, streamBitRate, streamBitRate, InetAddress.getByName( nepAddress ) );
		this.streamType = streamType;
		this.packetSize = packetSize;
		this.bitrate = streamBitRate;
	}


	/**
	 * Creates a new <code>Source</code> using the gui mode.
	 * 
	 * @param gui
	 *            the gui that represents this <code>Source</code>
	 */
	public Source( SourceGUI gui ){

		super( gui );
	}


	/**
	 * Sets the parameters of this <code>Source</code>. This method is used
	 * when the <code>Source</code> has been created in gui mode.
	 * 
	 * @param ip
	 *            this <code>Source</code>s IP
	 * @param port
	 *            zhis <code>Source</code>s port
	 * @param streamBitRate
	 *            bit rate of the stream this <code>Source</code> is sending
	 * @param nepAddress
	 *            IP address of the <code>NetworkEntryPoint</code> of the
	 *            network this <code>Source</code> wants to join
	 * @param streamType
	 *            type of the stream sent by this <code>Source</code> (see
	 *            ProtocolConstants)
	 */
	public void configure( InetAddress ip, int port, int streamBitRate, InetAddress nepAddress, byte streamType ){

		super.configure( ip, port, streamBitRate, streamBitRate, nepAddress );
		this.streamType = streamType;
		this.bitrate = streamBitRate;
		write2gui( "configured" );
	}


	/**
	 * Initializes the <code>Packetizer</code> to create
	 * <code>P2PPackets</code> of the given size.
	 * 
	 * @param packetSize
	 *            the size in bytes of the created packets (payload only); must
	 *            be >0 or else a default size of 16k will be set
	 * @return the packet size actually used
	 * @see Packetizer
	 */
	public int initPacketizer( int packetSize ){

		if( packetSize <= 0 ){
			// invalid packetSize argument; use default of 16000
			this.packetSize = 16000;
		}else{
			// valid argument; take it
			this.packetSize = packetSize;
		}
		packetizer = new Packetizer( this.packetSize, this );
		packetBufferSize = Runtime.getRuntime().maxMemory() / ( 2 * packetSize );

		return this.packetSize;
	}


	/**
	 * Starts this <code>Source</code>. The <code>InputAdapter</code> is
	 * started as a new threads and starts gathering media data which will be
	 * forwarded to the <code>Packetizer</code>.
	 */
	public void start(){

		new Thread( adapter, "InputAdapter" ).start();
		setStatus( Node.started );
		write2gui( "stream started with packets of size " + packetSize + "bytes" );
	}


	/**
	 * Stops this <code>Source</code>. The adapters will be stopped.
	 */
	public void stop(){

		adapter.stop();
		setStatus( Node.connected );
		write2gui( "stream stopped" );
	}


	/**
	 * Leave the network to which this <code>Source</code> is currently
	 * connected. The <code>InputAdapter</code>s will be stopped.
	 */
	public void leaveNetwork(){

		super.leaveNetwork();
		adapter.stop();
	}


	// GETTERS AND SETTERS

	public int setPacketSize( int packetSize ){

		return 0;
	}


	public int getPacketSize(){

		return packetizer.getPacketSize();
	}


	public InputAdapter getInputAdapter(){

		return adapter;
	}


	public void setInputAdapter( InputAdapter adapter ){

		this.adapter = adapter;
		adapter.setTarget( packetizer );
	}

}
