/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.peers;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.dcg.p2p.exceptions.DispatcherException;
import ch.ethz.dcg.p2p.handlers.ByeConfirmHandler;
import ch.ethz.dcg.p2p.handlers.ByeHandler;
import ch.ethz.dcg.p2p.handlers.DroppedConfirmHandler;
import ch.ethz.dcg.p2p.handlers.DroppedHandler;
import ch.ethz.dcg.p2p.handlers.FloodHandler;
import ch.ethz.dcg.p2p.handlers.ForwardHandler;
import ch.ethz.dcg.p2p.handlers.HelloHandler;
import ch.ethz.dcg.p2p.handlers.DirectReportNeighboursHandler;
import ch.ethz.dcg.p2p.handlers.ReportStatisticHandler;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * Dispatcher for incoming TCP connections.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see UDPDispatcher
 */
public class TCPDispatcher extends Thread {

	private final Log logger = LogFactory.getLog( this.getClass() );

	private Node source;
	private ServerSocket serverSocket;
	private volatile boolean running = true;


	/**
	 * Creates a new <code>TCPDispatcher</code>.
	 * 
	 * @param n
	 *            the <code>Node</code> which owns this
	 *            <code>TCPDispatcher</code>
	 * @throws DispatcherException
	 *             if the server socket cannot be opened
	 */
	public TCPDispatcher( Node n ) throws DispatcherException{

		this.source = n;
		try{
			// open ServerSocket
			serverSocket = new ServerSocket( source.port );
		}catch( IOException e ){
			try{
				serverSocket.close();
			}catch( Exception e1 ){
			}
			throw new DispatcherException( "could not open ServerSocket" );
		}
	}


	/**
	 * Handles incoming TCP connections and forwards them to the dispatcher.
	 */
	public void run(){

		Socket socket;

		if( logger.isDebugEnabled() ){
			logger.debug( "TCPDispatcher running" );
		}

		while( running ){
			try{
				socket = serverSocket.accept();
				if( logger.isDebugEnabled() ){
					logger.debug( "Incoming connection from " + socket.getInetAddress().getHostAddress() );
				}
				dispatch( socket );
			}catch( IOException e ){
				if( logger.isErrorEnabled() ){
					logger.error( "IOException at ServerSocket ( " + e.getMessage() + " )" );
				}
			}catch( DispatcherException e ){
				if( logger.isErrorEnabled() ){
					logger.error( e );
				}
			}
		}
	}


	/**
	 * Dispatches the incoming messages.
	 * 
	 * @param socket
	 *            the TCP connection to the sender of the message
	 * @throws DispatcherException
	 *             if a communication error between this dispatcher and the
	 *             sender occurred
	 */
	public void dispatch( Socket socket ) throws DispatcherException{

		InputStream inStream;

		try{
			// get InputStream
			inStream = socket.getInputStream();
			// determine message type
			byte type = (byte)inStream.read();
			// dispatch it
			switch( type ){

			// PING
			case ProtocolConstants.PING:
				if( logger.isInfoEnabled() ){
					logger.info( "PING message received from " + socket.getInetAddress().getHostAddress() );
				}
				try{
					sendPONG( socket );
				}catch( IOException e ){
					if( logger.isErrorEnabled() ){
						logger.error( "cannot send PONG message to " + socket.getInetAddress().getHostAddress() );
					}
				}
				break;

			// HELLO
			case ProtocolConstants.HELLO:
				new Thread( new HelloHandler( socket, source ), "HelloHandler" ).start();
				break;

			// FORWARD
			case ProtocolConstants.FORWARD:
				new Thread( new ForwardHandler( socket, source ), "ForwardHandler" ).start();
				break;

			// DROPPED
			case ProtocolConstants.DROPPED:
				new Thread( new DroppedHandler( socket, source ), "DroppedHandler" ).start();
				break;

			// DROPPEDCONFIRM
			case ProtocolConstants.DROPPEDCONFIRM:
				new Thread( new DroppedConfirmHandler( socket, source ), "DroppedConfirmHandler" ).start();
				break;

			// DIRECTREPORTNEIGHBOURS
			case ProtocolConstants.DIRECTREPORTNEIGHBOURS:
				new Thread( new DirectReportNeighboursHandler( socket, source ), "ReportNeighboursHandler" ).start();
				break;

			// FLOOD
			case ProtocolConstants.FLOOD:
				new Thread( new FloodHandler( socket, source ), "FloodHandler" ).start();
				break;

			// REPORTSTATISTIC
			case ProtocolConstants.REPORTSTATISTIC:
				new Thread( new ReportStatisticHandler( socket, source, null ), "ReportStatisticHandler" ).start();
				break;

			// BYE
			case ProtocolConstants.BYE:
				new Thread( new ByeHandler( socket, source, null ), "ByeHandler" ).start();
				break;

			// BYECONFIRM
			case ProtocolConstants.BYECONFIRM:
				new Thread( new ByeConfirmHandler( socket, source ), "ByeConfirmHandler" ).start();
				break;

			// DEFAULT
			default:
				if( logger.isErrorEnabled() ){
					logger.error( "invalid message type received " + type );
				}
				break;
			}
		}catch( IOException e ){
			throw new DispatcherException( "cannot dispatch the current socket" );
		}
	}


	/**
	 * Sends PONG message as answer for PING message.
	 * 
	 * @param socket
	 *            the TCP connection to the host that has sent the PING message
	 * @throws IOException
	 *             if a communication error occurred
	 */
	public synchronized void sendPONG( Socket socket ) throws IOException{

		OutputStream outStream = socket.getOutputStream();
		outStream.write( ProtocolConstants.PONG );
		outStream.flush();
		if( logger.isInfoEnabled() ){
			logger.info( "PONG message sent to " + socket.getInetAddress().getHostAddress() );
		}
	}


	public void close(){

		running = false;
		try{
			serverSocket.close();
		}catch( Exception e ){
		}
	}
}
