/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.peers;

import java.net.DatagramPacket;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.dcg.p2p.exceptions.DispatcherException;
import ch.ethz.dcg.p2p.handlers.NotificationHandler;
import ch.ethz.dcg.p2p.handlers.PacketHandler;
import ch.ethz.dcg.p2p.handlers.PacketNotAvailableHandler;
import ch.ethz.dcg.p2p.handlers.RequestHandler;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * Dispatcher for incoming UPC packets.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see TCPDispatcher
 */
public class UDPDispatcher {

	private final Log logger = LogFactory.getLog( this.getClass() );

	private Node source;


	/**
	 * Creates a new <code>UDPDispatcher</code>.
	 * 
	 * @param n
	 *            the <code>Node</code> that owns this
	 *            <code>UDPDispatcher</code>
	 * @throws DispatcherException
	 *             if the server socket cannot be opened
	 */
	public UDPDispatcher( Node n ) throws DispatcherException{

		this.source = n;
	}


	/**
	 * Dispatches the incoming messages.
	 * 
	 * @param packet
	 *            the received UDP packet
	 */
	public void dispatchPacket( DatagramPacket packet ){

		if( logger.isDebugEnabled() ){
			logger.debug( "UDP packet received from " + packet.getAddress().getHostAddress() + ":" + packet.getPort() );
		}

		// get the payload
		byte[] udpData = packet.getData();

		// determine message type
		byte type = udpData[0];

		// dispatch it
		switch( type ){

		// NOTIFICATION ( 17 bytes )
		case ProtocolConstants.NOTIFICATION:
			new Thread( new NotificationHandler( udpData, source ), "NotificationHandler" ).start();
			break;

		// REQUEST ( 17 bytes )
		case ProtocolConstants.REQUEST:
			new Thread( new RequestHandler( udpData, source ), "RequestHandler" ).start();
			break;

		// PACKET ( 21+packetSize bytes )
		case ProtocolConstants.P2PPACKET:
			new Thread( new PacketHandler( udpData, source ), "PacketHandler" ).start();
			break;

		// PACKETNOTAVAILABLE ( 17 bytes )
		case ProtocolConstants.PACKETNOTAVAILABLE:
			new Thread( new PacketNotAvailableHandler( udpData, source ), "PacketNotAvailableHandler" ).start();
			break;

		// DEFAULT
		default:
			if( logger.isErrorEnabled() ){
				logger.error( "invalid message type received " + type );
			}
			break;
		}
	}
}
