/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.tasks;

import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Thread for sending BYE messages to the neighbours and the NetworkEntryPoint.
 * This is done when the owner leaves the network.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see Task
 */
public class ByeTask extends Task {

	/**
	 * Creates a new <code>ByeTask</code>.
	 * 
	 * @param source
	 *            the <code>Node</code> that owns this <code>Task</code>
	 */
	public ByeTask( Node source ){

		this.source = source;
	}


	/**
	 * Starts this task.
	 */
	public void run(){

		if( !source.isConfigured() ){
			return;
		}

		if( logger.isDebugEnabled() ){
			logger.debug( "ByeTask started" );
		}

		source.setStatus( Node.configured );
		source.write2gui( "network left" );

		try{
			InetAddress ip;
			int port;

			// send BYE to NetworkEntryPoint
			ip = source.getNepAddress();
			port = ProtocolConstants.NEPPORT;
			sendByeTo( ip, port );
			try{
				Thread.sleep( 100 );
			}catch( InterruptedException e ){
			}

			// send BYE to my current neighbours
			for( Neighbour curN : source.getNeighboursCopy() ){
				sendByeTo( curN.getIp(), curN.getPort() );
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}


	/**
	 * Send a BYE message to the specified IP and port. <byte:BYE><InetAddress:myIp><int:myPort>
	 * 
	 * @param targetIp
	 *            target IP
	 * @param targetPort
	 *            target port
	 */
	public void sendByeTo( InetAddress targetIp, int targetPort ){

		try{
			socket = new Socket( targetIp, targetPort );
			DataOutputStream out = new DataOutputStream( socket.getOutputStream() );
			out.write( ProtocolConstants.BYE );
			Utilities.writeIP( source.getIp(), out );
			out.writeInt( source.getPort() );
			out.flush();
			if( logger.isInfoEnabled() ){
				logger.info( "BYE message sent to " + targetIp.getHostAddress() + ":" + targetPort );
			}
		}catch( IOException e ){
			if( logger.isWarnEnabled() ){
				logger.warn( "cannot send BYE message to " + targetIp.getHostAddress() + ":" + targetPort );
			}
		}finally{
			try{
				socket.close();
			}catch( Exception e ){
			}
		}
	}

}
