/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.tasks;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.util.Hashtable;
import java.util.Iterator;

import ch.ethz.dcg.p2p.exceptions.NeighbourUnavailableException;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.exceptions.WrongProtocolMessageException;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.peers.Source;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Thread for sending finding new neighbours.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see Task
 */
public class GetNewNeighboursTask extends Task {

	private Hashtable<String, Neighbour> candidates = new Hashtable<String, Neighbour>();
	private String sourceKey;
	private String removedKey;


	/**
	 * Creates a new <code>GetNewNeighboursTask</code>.
	 * 
	 * @param source
	 *            the <code>Node</code> that owns this <code>Task</code>
	 * @param leavingIp
	 *            the IP address of the neighbour that has just left
	 * @param leavingPort
	 *            the port of that neighbour
	 */
	public GetNewNeighboursTask( Node source, InetAddress leavingIp, int leavingPort ){

		this.source = source;
		sourceKey = source.getIp().getHostAddress() + ":" + source.getPort();
		removedKey = leavingIp.getHostAddress() + ":" + leavingPort;
	}


	/**
	 * Starts this thread.<br>
	 * Get all distance-2-neighbours from the remaining distance-1-neighbours.
	 * If I do not have any distance-1-neighbours I will re-join at the
	 * <code>NetworkEntryPoint</code>.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "GetNewNeighboursTask started" );
		}

		try{
			// the Source gets new neighbours directly from the NEP
			if( source instanceof Source ){
				source.joinNetwork();
				source.write2gui( "getting currently fastest neighbours from the Network Entry Point" );
				return;
			}

			// send DIRECTREPORTNEIGHBOURS message to all neighbours
			for( Neighbour curN : source.getNeighboursCopy() ){
				getNeighboursOf( curN );
			}

			// re-join the network if my neighbours do not have new neighbours
			// for me
			if( candidates.isEmpty() ){
				if( logger.isDebugEnabled() ){
					logger.debug( "no new neighbours left; re-joining at the NetworkEntryPoint" );
				}
				source.joinNetwork();
				return;
			}

			// send HELLO messages to candidates and add new neighbours to
			// source
			Iterator<Neighbour> it = candidates.values().iterator();
			while( it.hasNext() && source.hasSpareCapacity() ){
				Neighbour curN = it.next();
				HelloTask hello = new HelloTask( source, curN );
				try{
					hello.openConnection();
					hello.sendHello();
					Neighbour newNeigh = hello.getAnswer();
					if( newNeigh != null ){
						source.addNeighbour( newNeigh );
					}
				}catch( NeighbourUnavailableException e ){
					if( logger.isErrorEnabled() ){
						logger.error( e.getMessage() );
					}
				}catch( IOException e ){
					if( logger.isErrorEnabled() ){
						logger.error( e.getMessage() );
					}
				}catch( WrongProtocolMessageException e ){
					if( logger.isErrorEnabled() ){
						logger.error( e.getMessage() );
					}
				}finally{
					hello.closeConnection();
				}
			}
			
			// re-join if there are still not enough neighbours available
			if( !source.hasEnoughNeighbours() ){
				if( logger.isDebugEnabled() ){
					logger.debug( "not enough new neighbours found; re-joining at the NetworkEntryPoint" );
				}
				source.joinNetwork();
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}


	/**
	 * Sends a DIRECTREPORTNEIGHBOURS message to <code>Neighbour</code> n and
	 * gets his answer. The reported <code>Neighbour</code>s are added to the
	 * candidates <code>Vector</code>.
	 * 
	 * @param n
	 *            the distance-1-neighbour to ask for his neighbours
	 */
	private void getNeighboursOf( Neighbour n ){

		Socket socket = null;
		DataOutputStream out = null;
		DataInputStream in = null;

		try{

			socket = new Socket( n.getIp(), n.getPort() );
			out = new DataOutputStream( socket.getOutputStream() );
			in = new DataInputStream( socket.getInputStream() );

			out.writeByte( ProtocolConstants.DIRECTREPORTNEIGHBOURS );
			out.flush();

			if( logger.isInfoEnabled() ){
				logger.info( "DIRECTREPORTNEIGHBOURS message sent to " + n.getIp().getHostAddress() + ":" + n.getPort() );
			}

			receiveNeighbours( in );

		}catch( IOException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( e.getMessage() );
			}
		}catch( WrongProtocolMessageException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}finally{
			try{
				socket.close();
			}catch( Exception e ){
			}
		}
	}


	/**
	 * Gets the answer from the distance-1-neighbour and adds his neighbours to
	 * the candidates <code>Vector</code>.
	 * 
	 * @param in
	 *            the <code>InputStream</code> on the TCP connection to the
	 *            distance-1-neighbour
	 * @throws IOException
	 *             if a communication error occurred
	 * @throws WrongProtocolMessageException
	 *             if an unexpected answer was received
	 */
	private void receiveNeighbours( DataInputStream in ) throws IOException, WrongProtocolMessageException{

		byte type = in.readByte();
		if( type != ProtocolConstants.DIRECTNEIGHBOURS ){
			throw new WrongProtocolMessageException( "wrong message type received: " + type );
		}

		if( logger.isInfoEnabled() ){
			logger.info( "DIRECTNEIGHBOURS message received" );
		}
		int n = in.readInt();
		if( n > source.getMaxNeighs() ){
			n = source.getMaxNeighs();
		}
		if( n < 0 ){
			throw new IOException( "invalid number of neighbours received" );
		}
		while( n > 0 ){
			InetAddress neighbourIp = Utilities.readIP( in );
			int neighbourPort = in.readInt();
			addCandidate( neighbourIp, neighbourPort );
			n--;
		}
	}


	/**
	 * Adds a possible new neighbour to the <code>Vector</code> of candidates.
	 * 
	 * @param ip
	 *            the candidate's IP address
	 * @param port
	 *            the candidate's port
	 */
	private void addCandidate( InetAddress ip, int port ){

		String key = ip.getHostAddress() + ":" + port;
		if( ( !key.equals( sourceKey ) ) && !key.equals( removedKey ) && !candidates.containsKey( key ) ){
			candidates.put( key, new Neighbour( ip, port ) );
			if( logger.isDebugEnabled() ){
				logger.debug( ip.getHostAddress() + ":" + port + " added to possible new neighbours" );
			}
		}
	}
}
