/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.tasks;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;

import ch.ethz.dcg.p2p.exceptions.NeighbourUnavailableException;
import ch.ethz.dcg.p2p.exceptions.WrongProtocolMessageException;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.utilities.Utilities;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * Thread for sending HELLO messages to a new neighbour. This is done when the
 * owner joins the network.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see Task
 */
public class HelloTask extends Task {

	private DataOutputStream outStream = null;
	private DataInputStream inStream = null;


	/**
	 * Creates a new <code>HelloTask</code>.
	 * 
	 * @param source
	 *            the <code>Node</code> that owns this <code>Task</code>
	 */
	public HelloTask( Node source, Neighbour target ){

		this.source = source;
		this.target = target;
	}


	/**
	 * Starts this task.<br>
	 * A HELLO message is sent to the target neighbour which will answer with
	 * either an ACCEPT or a REDIRECT message.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "HelloTask started with target " + target.getIp() );
		}

		// open a TCP connection to the target host
		// send HELLO message
		// receive an ACCEPT or REDIRECT message from the target host
		try{
			openConnection();
			sendHello();
			getAnswer();
		}catch( NeighbourUnavailableException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}catch( WrongProtocolMessageException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}finally{
			try{
				source.printNeighbours();
				socket.close();
			}catch( IOException e ){
			}
		}
	}


	/**
	 * Opens TCP connection to the target host.
	 * 
	 * @throws NeighbourUnavailableException
	 *             if a communication error occurred
	 */
	public void openConnection() throws NeighbourUnavailableException{

		try{
			socket = new Socket( target.getIp(), target.getPort() );
			inStream = new DataInputStream( socket.getInputStream() );
			outStream = new DataOutputStream( socket.getOutputStream() );
		}catch( IOException e ){
			throw new NeighbourUnavailableException( "cannot connect to neighbour " + target.getIp() );
		}
	}


	/**
	 * Closes the TCP connection to the target host.
	 * 
	 */
	public void closeConnection(){

		try{
			socket.close();
		}catch( Exception e ){
		}
	}


	/**
	 * Sends the HELLO message.
	 * 
	 * @throws IOException
	 *             if a communication error occurred
	 */
	public void sendHello() throws IOException{

		try{
			outStream.writeByte( ProtocolConstants.HELLO );
			Utilities.writeIP( source.getIp(), outStream );
			outStream.writeInt( source.getPort() );
			outStream.flush();
			if( logger.isInfoEnabled() ){
				logger.info( "HELLO message sent to " + target.getIp().getHostAddress() + ":" + target.getPort() );
			}
		}catch( IOException e ){
			throw new IOException( "cannot send HELLO message to " + target.getIp() );
		}
	}


	/**
	 * Receives the answer from the target host.
	 * 
	 * @return the new <code>Neighbour</code>
	 * @throws WrongProtocolMessageException
	 *             if an unexpected message was received
	 * @throws IOException
	 *             if a communication error occurred
	 */
	public Neighbour getAnswer() throws WrongProtocolMessageException, IOException{

		InetAddress remoteIp;
		int remotePort;
		int packetSize, bitrate;
		try{
			byte type = inStream.readByte();
			if( type == ProtocolConstants.ACCEPT ){
				// the target has accepted me as a direct neighbour
				remoteIp = Utilities.readIP( inStream );
				remotePort = inStream.readInt();
				packetSize = inStream.readInt();
				bitrate = inStream.readInt();
				if( remoteIp.equals( target.getIp() ) && ( remotePort == target.getPort() ) ){
					if( logger.isInfoEnabled() ){
						logger.info( "ACCEPT message received from " + remoteIp.getHostAddress() + ":" + remotePort );
					}
					if( source.getPacketSize() == -1 ){
						source.setPacketSize( packetSize );
					}
					if( source.getBitrate() == -1 ){
						source.setBitrate( bitrate );
					}
					return new Neighbour( remoteIp, remotePort );
				}
			}else if( type == ProtocolConstants.REDIRECT ){
				// the target has redirected me to one of his neighbours
				remoteIp = Utilities.readIP( inStream );
				remotePort = inStream.readInt();
				packetSize = inStream.readInt();
				bitrate = inStream.readInt();
				if( logger.isInfoEnabled() ){
					logger.info( "REDIRECT message received from " + socket.getInetAddress().getHostAddress() );
				}
				if( logger.isDebugEnabled() ){
					logger.debug( target.getIp().getHostAddress() + " has redirected me to be a direct neighbour of " + remoteIp.getHostAddress() );
				}
				if( source.getPacketSize() == -1 ){
					source.setPacketSize( packetSize );
				}
				if( source.getBitrate() == -1 ){
					source.setBitrate( bitrate );
				}
				return new Neighbour( remoteIp, remotePort );
			}else{
				throw new WrongProtocolMessageException( "wrong message type received: " + type );
			}
		}catch( IOException e ){
			throw new IOException( "cannot get answer on HELLO message from " + target.getIp() );
		}
		return null;
	}
}
