/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.tasks;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.util.Vector;

import ch.ethz.dcg.p2p.exceptions.JoinException;
import ch.ethz.dcg.p2p.exceptions.NeighbourUnavailableException;
import ch.ethz.dcg.p2p.exceptions.NepUnavailableException;
import ch.ethz.dcg.p2p.exceptions.NoInitNeighboursException;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.exceptions.WrongProtocolMessageException;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.peers.Peer;
import ch.ethz.dcg.p2p.peers.Source;
import ch.ethz.dcg.p2p.utilities.Utilities;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * This Task connects to the <code>NetworkEntryPoint</code> and requests a
 * list of inital neighbours. These will be added to the neighbours list of the
 * <code>Node</code> that has started this task. After that, the node sends
 * HELLO messages to all these initial neighbours.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see Task
 */
public class JoinTask extends Task {

	DataInputStream inStream;
	DataOutputStream outStream;

	int speed;
	InetAddress nepAddress;
	byte nodeType;
	byte streamType = ProtocolConstants.UNKNOWN_TYPE;
	Vector<Neighbour> candidates = new Vector<Neighbour>();


	/**
	 * Creates a new <code>JoinTask</code>.<br>
	 * Used by <code>Peer</code>s.
	 * 
	 * @param source
	 *            the owner of this task
	 * @param nepAddress
	 *            the IP address of the <code>NetworkEntryPoint</code> to join
	 *            with
	 * @param speed
	 *            the reported outbound speed (used by the
	 *            <code>NetworkEntryPoint</code> to determine the correct
	 *            level)
	 * @param nodeType
	 *            the type of the owner <code>Node</code> (<code>Source</code>
	 *            or <code>Peer</code>)
	 */
	public JoinTask( Node source, InetAddress nepAddress, int speed, byte nodeType ){

		this.source = source;
		this.nepAddress = nepAddress;
		this.speed = speed;
		this.nodeType = nodeType;
	}


	/**
	 * Creates a new <code>JoinTask</code>.<br>
	 * Used by <code>Source</code>s.
	 * 
	 * @param source
	 *            the owner of this task
	 * @param nepAddress
	 *            the IP address of the <code>NetworkEntryPoint</code> to join
	 *            with
	 * @param speed
	 *            the reported outbound speed (used by the
	 *            <code>NetworkEntryPoint</code> to determine the correct
	 *            level)
	 * @param nodeType
	 *            the type of the owner <code>Node</code> (<code>Source</code>
	 *            or <code>Peer</code>)
	 * @param streamType
	 *            the type of the media stream
	 */
	public JoinTask( Node source, InetAddress nepAddress, int speed, byte nodeType, byte streamType ){

		this.source = source;
		this.nepAddress = nepAddress;
		this.speed = speed;
		this.nodeType = nodeType;
		this.streamType = streamType;
	}


	/**
	 * Starts this task.<br>
	 * A JOIN message is sent to the <code>NetworkEntryPoint</code> which will
	 * respond with either a INITNEIGHBOURS message or an error message if
	 * something does not work.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "JoinTask started" );
		}

		// open connection to the NetworkEntryPoint
		// send JOIN message
		// receive list of initial neighbours and add them to the invoking node
		// send HELLO message to each neighbour
		try{
			openConnectionToNep();
			sendJOIN();
			if( getAnswer() ){
				sayHelloToInitNeighbours();
				if( source instanceof Peer ){
					Peer p = (Peer)source;
					p.initDePacketizer();
					p.initOutputAdapter();
				}
				source.printNeighbours();
				synchronized( source ){
					source.notify();
				}
			}
		}catch( NepUnavailableException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( "NetworkEntryPoint unavailable ( " + e.getMessage() + " )" );
			}
			source.setStatus( Node.error );
			source.write2gui( "[ERROR] network entry point unavailable ( " + e.getMessage() + " )" );
		}catch( JoinException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( "join failed ( " + e.getMessage() + " )" );
			}
			source.write2gui( "[ERROR] join failed ( " + e.getMessage() + " )" );
		}catch( WrongProtocolMessageException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( e.getMessage() );
			}
			source.setStatus( Node.error );
			source.write2gui( "[ERROR] " + e.getMessage() );
		}catch( NoInitNeighboursException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( e.getMessage() );
			}
			if( !( source instanceof Source ) ){
				source.setStatus( Node.error );
				source.write2gui( "[ERROR] " + e.getMessage() );
			}
		}catch( Exception e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( "join failed (" + e.getMessage() + " )" );
			}
			source.setStatus( Node.error );
			source.write2gui( "[ERROR] join failed ( " + e.getMessage() + " )" );
		}finally{
			try{
				socket.close();
			}catch( Exception e ){
			}
		}
	}


	/**
	 * Opens a TCP connection to the NetworkEntryPoint.
	 * 
	 * @throws NepUnavailableException
	 *             if the <code>NetworkEntryPoint</code> is not available
	 */
	private void openConnectionToNep() throws NepUnavailableException{

		try{
			socket = new Socket( nepAddress, ProtocolConstants.NEPPORT );
			inStream = new DataInputStream( socket.getInputStream() );
			outStream = new DataOutputStream( socket.getOutputStream() );
		}catch( IOException e ){
			try{
				socket.close();
			}catch( Exception e1 ){
			}
			throw new NepUnavailableException( e.getMessage() );
		}
	}


	/**
	 * Sends the actual JOIN message to the NetworkEntryPoint.
	 * 
	 * @throws JoinException
	 *             if a communication error occurred
	 */
	private void sendJOIN() throws JoinException{

		try{
			outStream.writeByte( ProtocolConstants.JOIN );
			outStream.writeInt( source.getPort() );
			outStream.writeInt( speed );
			outStream.writeByte( nodeType );
			if( nodeType == ProtocolConstants.SOURCE ){
				outStream.writeByte( streamType );
			}
			outStream.write( ProtocolConstants.VERSION.getBytes() );
			outStream.flush();
			if( logger.isInfoEnabled() ){
				logger.info( "JOIN message sent to " + socket.getInetAddress().getHostAddress() );
			}
		}catch( IOException e ){
			throw new JoinException( e.getMessage() );
		}
	}


	/**
	 * Waits for the message from the <code>NetworkEntryPoint</code>
	 * containing the initial neighbours or an error message.
	 * 
	 * @throws WrongProtocolMessageException
	 *             if an unexpected message was received
	 * @throws JoinException
	 *             if a communication error occurred
	 * @throws NoInitNeighboursException
	 *             if the set of initial neighbours is empty
	 * @return true if an INITNEIGHBOURS response was received; false otherwise
	 */
	private boolean getAnswer() throws WrongProtocolMessageException, JoinException, NoInitNeighboursException{

		try{
			byte type = (byte)inStream.readByte();

			switch( type ){
			case ProtocolConstants.INITNEIGHBOURS:
				// everything ok
				if( logger.isInfoEnabled() ){
					logger.info( "INITNEIGHBOURS message received from " + socket.getInetAddress().getHostAddress() );
				}
				if( source.getStatus().equals( Node.configured ) ){
					source.setStatus( Node.connected );
				}
				source.write2gui( "successfully (re-)connected to the p2p network" );
				break;
			case ProtocolConstants.PEERTOOSLOW:
				if( logger.isFatalEnabled() ){
					logger.fatal( "too slow to join network; try again later" );
				}
				source.setStatus( Node.error );
				source.write2gui( "[ERROR] too slow to join; try again later as speed requirements may vary" );
				source.errorDialog( "Your Internet connection is too slow for this stream.\nPlease try again later as speed requirements may vary.", "Connection too slow" );
				return false;
			case ProtocolConstants.DUPLICATESOURCE:
				if( logger.isFatalEnabled() ){
					logger.fatal( "a source with this speed already exists; cannot join" );
				}
				source.setStatus( Node.error );
				source.write2gui( "[ERROR] duplicate host; cannot join" );
				return false;
			case ProtocolConstants.VERSIONMISMATCH:
				byte[] v = new byte[255];
				int len = inStream.read( v );
				String version = new String( v, 0, len );
				if( logger.isFatalEnabled() ){
					logger.fatal( "invalid software version: " + version + " is required for this network but this is " + ProtocolConstants.VERSION );
				}
				source.setStatus( Node.error );
				source.write2gui( "invalid software version: " + version + " is required for this network but this is " + ProtocolConstants.VERSION );
				source.errorDialog( "Version " + version + " required (this is " + ProtocolConstants.VERSION + ")\n\n Please install the required version to join this network.", "Software Version Mismatch" );
				return false;
			default:
				throw new WrongProtocolMessageException( "wrong message type received: " + type );
			}

		}catch( IOException e ){
			throw new JoinException( e.getMessage() );
		}

		try{
			// parse INITNEIGHBOURS message
			InetAddress ip = null;
			int port = 0;
			int minNeighs = inStream.readInt();
			int maxNeighs = inStream.readInt();
			if( ( minNeighs < 0 ) || ( minNeighs > maxNeighs ) ){
				return false;
			}else{
				source.setMinNeighs( minNeighs );
			}
			if( ( maxNeighs < 0 ) ){
				return false;
			}else{
				source.setMaxNeighs( maxNeighs );
			}
			byte streamType = inStream.readByte();
			switch( streamType ){
			case ProtocolConstants.HTTP_MP3:
				source.setStreamType( streamType );
				break;
			case ProtocolConstants.HTTP_VLC:
				source.setStreamType( streamType );
				break;
			default:
				return false;
			}
			int n = inStream.readInt();
			if( n == 0 ){
				throw new NoInitNeighboursException( "set of initial neighbours is empty" );
			}
			if( n > source.getMaxNeighs() ){
				n = source.getMaxNeighs();
			}
			if( n < 0 ){
				throw new JoinException( "invalid number of initial neighbours received" );
			}
			for( int i = 0; i < n; i++ ){
				ip = Utilities.readIP( inStream );
				port = inStream.readInt();
				candidates.add( new Neighbour( ip, port ) );
			}
			return true;
		}catch( IOException e ){
			throw new JoinException( e.getMessage() );
		}
	}


	/**
	 * Sends a HELLO message to each neighbour and updates the neighbours set of
	 * the source.
	 * 
	 */
	private void sayHelloToInitNeighbours() throws NodeNotConfiguredException{

		for( Neighbour curN : candidates ){
			if( matchesSource( curN ) || source.hasNeighbour( curN.getIp(), curN.getPort() ) ){
				// do not send a HELLO message to myself or to my existing
				// neighbours
				continue;
			}
			HelloTask hello = new HelloTask( source, curN );
			try{
				hello.openConnection();
				hello.sendHello();
				Neighbour newNeigh = hello.getAnswer();
				if( newNeigh != null ){
					source.addNeighbour( newNeigh );
				}
			}catch( NeighbourUnavailableException e ){
				if( logger.isErrorEnabled() ){
					logger.error( e.getMessage() );
				}
			}catch( IOException e ){
				if( logger.isErrorEnabled() ){
					logger.error( e.getMessage() );
				}
			}catch( WrongProtocolMessageException e ){
				if( logger.isErrorEnabled() ){
					logger.error( e.getMessage() );
				}
			}catch( NodeNotConfiguredException e ){
				if( logger.isErrorEnabled() ){
					logger.error( e.getMessage() );
				}
			}finally{
				hello.closeConnection();
			}
		}
		source.printNeighbours();
	}


	/**
	 * Tests if the given neighbour matches the owner of this task in IP address
	 * and port number.
	 * 
	 * @param n
	 *            the possibly matching neighbour
	 * @return true if they match; false otherwise
	 */
	private boolean matchesSource( Neighbour n ){

		if( n.getIp().equals( source.getIp() ) && n.getPort() == source.getPort() ){
			return true;
		}else{
			return false;
		}
	}
}
