/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.tasks;

import java.io.IOException;
import java.net.DatagramPacket;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.packets.P2PPacket;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.UDPSocketManager;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Thread for sending NOTIFICATION messages to the neighbours. This is done when
 * a new <code>P2PPacket</code> has become available.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see Task
 */
public class NotificationTask extends Task {

	private long packetId;
	private UDPSocketManager socketManager;


	/**
	 * Creates a new <code>NotificationTask</code>.
	 * 
	 * @param source
	 *            the <code>Node</code> that owns this <code>Task</code>
	 * @param packetId
	 *            the id of the newly available packet
	 */
	public NotificationTask( Node source, long packetId ){

		this.source = source;
		this.packetId = packetId;
		this.socketManager = source.getSocketManager();
	}


	/**
	 * Starts this task.<br>
	 * Sends a NOTIFICATION message to all neighbours except for the one which
	 * has sent the packet. The messages are sent using UDP.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "NotificationTask started for packet " + packetId );
		}

		try{
			P2PPacket packet = source.getPacket( packetId );
			if( packet == null ){
				if( logger.isWarnEnabled() ){
					logger.warn( "trying to send notification for a packet that is not available anymore; abort" );
				}
				return;
			}
			Neighbour receivedFrom = packet.getReceivedFrom();
			for( Neighbour curN : source.getNeighboursCopy() ){
				try{
					// do not send a notification to the node from which I
					// had received the packet
					if( ( receivedFrom == null ) || !curN.equals( receivedFrom ) ){
						sendNotificationTo( curN );
					}
				}catch( IOException e ){
					if( logger.isErrorEnabled() ){
						logger.error( "cannot send NOTIFICATION message to " + curN.getIp().getHostAddress() + ":" + curN.getPort() );
					}
				}
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}


	/**
	 * Actually sends the NOTIFICATION packet to a neighbour.
	 * 
	 * @param n
	 *            the neighbour to which the NOTIFICATION message should be sent
	 * @throws IOException
	 *             if a communication error occurred
	 */
	private void sendNotificationTo( Neighbour n ) throws IOException{

		byte[] buffer = new byte[17];
		buffer[0] = ProtocolConstants.NOTIFICATION;
		System.arraycopy( Utilities.longToByteArray( packetId ), 0, buffer, 1, 8 );
		System.arraycopy( source.getIp().getAddress(), 0, buffer, 9, 4 );
		System.arraycopy( Utilities.intToByteArray( source.getPort() ), 0, buffer, 13, 4 );
		DatagramPacket packet = new DatagramPacket( buffer, buffer.length, n.getIp(), n.getPort() );
		socketManager.sendPacket( packet );
		if( logger.isInfoEnabled() ){
			logger.info( "NOTIFICATION message for packet " + packetId + " sent to " + n.getIp().getHostAddress() + ":" + n.getPort() );
		}
		Node.statistic.incNotificationsSentTo( n.getIp().getHostAddress() + ":" + n.getPort() );
	}
}
