/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.tasks;

import java.io.IOException;
import java.net.DatagramPacket;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.packets.P2PPacket;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.UDPSocketManager;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Thread for sending P2PPACKET messages to the neighbours.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see Task
 */
public class PacketTask extends Task {

	private long packetId;
	private UDPSocketManager socketManager;


	/**
	 * Creates a new <code>PacketTask</code>.
	 * 
	 * @param source
	 *            the <code>Node</code> that owns this <code>Task</code>
	 * @param target
	 *            the neighbour to which this packet should be sent
	 * @param packetId
	 *            the id of the packet to be sent
	 */
	public PacketTask( Node source, Neighbour target, long packetId ){

		this.source = source;
		this.target = target;
		this.packetId = packetId;
		this.socketManager = source.getSocketManager();
	}


	/**
	 * Starts the task.<br>
	 * A P2PPACKET message is sent using a UDP packet.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "PacketTask started for packet " + packetId );
		}

		try{
			P2PPacket packet = source.getPacket( packetId );
			if( packet == null ){
				if( logger.isErrorEnabled() ){
					logger.error( "P2PPACKET " + packetId + " not available; cannot send it to " + target.getIp().getHostAddress() + ":" + target.getPort() );
				}
				source.write2gui( "P2PPACKET " + packetId + " requested by " + target.getIp().getHostAddress() + ":" + target.getPort() + " not available anymore" );
				Node.statistic.incPacketNotAvailable();
				return;
			}
			try{
				byte[] buffer = new byte[21 + packet.getSize()];
				buffer[0] = ProtocolConstants.P2PPACKET;
				System.arraycopy( source.getIp().getAddress(), 0, buffer, 1, 4 );
				System.arraycopy( Utilities.intToByteArray( source.getPort() ), 0, buffer, 5, 4 );
				System.arraycopy( Utilities.longToByteArray( packetId ), 0, buffer, 9, 8 );
				System.arraycopy( Utilities.intToByteArray( packet.getSize() ), 0, buffer, 17, 4 );
				System.arraycopy( packet.getData(), 0, buffer, 21, packet.getSize() );
				DatagramPacket udpPacket = new DatagramPacket( buffer, buffer.length, target.getIp(), target.getPort() );
				socketManager.sendPacket( udpPacket );
				if( logger.isInfoEnabled() ){
					logger.info( "P2PPACKET " + packetId + " sent to " + target.getIp().getHostAddress() + ":" + target.getPort() );
				}
				Node.statistic.incPacketsSentTo( target.getIp().getHostAddress() + ":" + target.getPort() );
			}catch( IOException e ){
				if( logger.isErrorEnabled() ){
					logger.error( "IOException: cannot sent packet " + packetId + " to " + target.getIp().getHostAddress() + ":" + target.getPort() );
				}
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}
}
