/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.tasks;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketTimeoutException;

import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;

/**
 * Task to send a PING message to a neighbour in order to see whether he is
 * still alive.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see Task
 */
public class PingTask extends Task {

	private int timeout; // ms
	private boolean done = false;
	private boolean result = false;
	private InetAddress ip;
	private int port;


	/**
	 * Creates a new <code>PingTask</code>.
	 * 
	 * @param target
	 *            the <code>Node</code> that owns this <code>Task</code>
	 * @param timeout
	 *            the time in ms after which a node is reported as non-reachable
	 */
	public PingTask( Neighbour target, int timeout ){

		ip = target.getIp();
		port = target.getPort();
		this.timeout = timeout;
	}


	/**
	 * Creates a new <code>PingTask</code>.
	 * 
	 * @param ip
	 *            the IP address of the target host
	 * @param port
	 *            the port of the target host
	 * @param timeout
	 *            the time in ms after which a node is reported as non-reachable
	 */
	public PingTask( InetAddress ip, int port, int timeout ){

		this.ip = ip;
		this.port = port;
		this.timeout = timeout;
	}


	/**
	 * Starts this task. (Never actually used like this though.)
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "PingTask started with target " + ip.getHostAddress() + ":" + port );
		}
		result = sendPINGreceivePONG();
		done = true;
	}


	/**
	 * Sends a PING message and waits for timeout ms for the PONG answer.
	 */
	public boolean sendPINGreceivePONG(){

		try{
			Socket socket = null;
			// open socket (try to open new TCP connection to see if peer is
			// still there)
			socket = new Socket();
			socket.connect( new InetSocketAddress( ip, port ), timeout );
			// set timeout
			socket.setSoTimeout( timeout );
			OutputStream outStream = socket.getOutputStream();
			InputStream inStream = socket.getInputStream();
			// send PING
			outStream.write( ProtocolConstants.PING );
			outStream.flush();
			if( logger.isInfoEnabled() ){
				logger.info( "PING sent to " + ip.getHostAddress() + ":" + port );
			}
			// wait for PONG
			byte type = (byte)inStream.read();
			if( type == ProtocolConstants.PONG ){
				if( logger.isInfoEnabled() ){
					logger.info( "PONG received from " + ip.getHostAddress() + ":" + port );
				}
				return true;
			}else{
				return false;
			}
		}catch( SocketTimeoutException e ){
			if( logger.isDebugEnabled() ){
				logger.debug( "timeout of " + timeout + "ms reached" );
			}
			return false;
		}catch( IOException e ){
			return false;
		}finally{
			try{
				socket.close();
			}catch( Exception e ){
			}
		}
	}


	/**
	 * Tells if the PONG has arrived, the timeout has been reached or if it is
	 * still waiting.
	 * 
	 * @return true if the task has finished; false otherwise
	 */
	public boolean isDone(){

		return done;
	}


	/**
	 * Tells if the target host is reachable.
	 * 
	 * @return true if it is reachable; false otherwise
	 */
	public boolean getResult(){

		return result;
	}
}
