/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.tasks;

import java.io.IOException;
import java.net.DatagramPacket;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.NoSuchElementException;
import java.util.Vector;

import ch.ethz.dcg.p2p.exceptions.NeighbourUnavailableException;
import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.packets.Notification;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.Node;
import ch.ethz.dcg.p2p.utilities.ProtocolConstants;
import ch.ethz.dcg.p2p.utilities.UDPSocketManager;
import ch.ethz.dcg.p2p.utilities.Utilities;

/**
 * Thread for sending REQUEST messages to the neighbours.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see Task
 */
public class RequestTask extends Task {

	//TODO: better notification synchronization required!
	private Hashtable<Long, Notification> notifications;
	private UDPSocketManager socketManager;
	private boolean running = true;
	private int requestThreshold = 3;
	private boolean thresholdReached = false;


	/**
	 * Creates a new <code>RequestTask</code>.
	 * 
	 * @param source
	 *            the <code>Node</code> that owns this <code>Task</code>
	 * @param notifications
	 *            the notifications that are available at the owner
	 */
	public RequestTask( Node source, Hashtable<Long, Notification> notifications ){

		this.source = source;
		this.notifications = notifications;
		this.socketManager = source.getSocketManager();
	}


	/**
	 * Starts this task. A REQUEST message is sent using UDP.
	 */
	public void run(){

		Long nextKey;
		Notification nextNotification;
		Neighbour provider;

		if( logger.isDebugEnabled() ){
			logger.debug( "RequestTask running" );
		}

		try{
			while( running ){
				if( !thresholdReached ){
					// check if enough notifications have been received to start
					// requesting packets
					if( notifications.size() > requestThreshold ){
						thresholdReached = true;
					}else{
						synchronized( this ){
							try{
								wait();
							}catch( InterruptedException e ){
							}
						}
					}
				}else{
					// threshold reached; request packets if notifications
					// available
					if( !notifications.isEmpty() ){
						// TODO: better strategy than picking at random
						nextKey = getNextPacketKey();
						if( nextKey == null ){
							continue;
						}
						nextNotification = notifications.get( nextKey );
						try{
							provider = getProvider( nextNotification );
						}catch( NeighbourUnavailableException e ){
							if( logger.isWarnEnabled() ){
								logger.warn( e );
							}
							notifications.remove( nextKey );
							// packet for which the depacketizer is waiting is
							// no longer available -> move on to the next packet
							if( nextKey == source.getCurPacketId() ){
								source.setCurPacketId( source.getCurPacketId() + 1 );
							}
							continue;
						}
						if( sendRequest( nextNotification, provider ) ){
							source.moveToRequested( nextKey );
						}
					}else{
						synchronized( this ){
							try{
								wait();
							}catch( InterruptedException e ){
							}
						}
					}
				}
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
	}


	// private Long getNextPacketKeyAtRandom(){
	//
	// int rnd = rand.nextInt( notifications.size() );
	// Iterator<Long> it = notifications.keySet().iterator();
	// for( int i = 0; i < rnd; i++ ){
	// it.next();
	// }
	// return it.next();
	// }

	/**
	 * Selects the next packet that should be requested.
	 */
	private Long getNextPacketKey(){

		Long notId = null;
		try{
			notId = notifications.keySet().iterator().next();
		}catch( NoSuchElementException e ){
		}
		return notId;
	}


	/**
	 * Selects a provider for the selected <code>Notification</code>
	 * 
	 * @param n
	 *            the <code>Notification</code> for which a provider is needed
	 * @return a <code>Neighbour</code> that can provide the packet
	 * @throws NeighbourUnavailableException
	 *             if there are no <code>Neighbour</code>s left that have the
	 *             packet
	 */
	private Neighbour getProvider( Notification n ) throws NeighbourUnavailableException{

		// TODO: request from best holder instead of the first might be better
		Neighbour provider = null;
		try{
			Enumeration<Neighbour> holders = n.getHolders().elements();
			Vector<Neighbour> zombies = null;
			provider = holders.nextElement();
			// do not choose a provider which is no longer available
			while( !source.hasNeighbour( provider.getIp(), provider.getPort() ) ){
				if( zombies == null ){
					zombies = new Vector<Neighbour>();
				}
				zombies.add( provider );
				provider = holders.nextElement();
			}
			n.setProvider( provider );
			// remove unavailable neighbours
			if( zombies != null ){
				for( Neighbour zombie : zombies ){
					n.removeHolder( zombie );
				}
			}
		}catch( NoSuchElementException e ){
			throw new NeighbourUnavailableException( "no more Neighbours available for this Notification" );
		}catch( NodeNotConfiguredException e ){
		}
		return provider;
	}


	/**
	 * Sends a request to the target neighbour that has the desired packet.
	 * 
	 * @param n
	 *            the <code>Notification</code> for the packet that is desired
	 * @param target
	 *            the <code>Neighbour</code> which has the packet
	 * @return true if the REQUEST has been successfully sent
	 */
	private boolean sendRequest( Notification n, Neighbour target ){

		try{
			byte[] buffer = new byte[17];
			buffer[0] = ProtocolConstants.REQUEST;
			System.arraycopy( Utilities.longToByteArray( n.getPacketId() ), 0, buffer, 1, 8 );
			System.arraycopy( source.getIp().getAddress(), 0, buffer, 9, 4 );
			System.arraycopy( Utilities.intToByteArray( source.getPort() ), 0, buffer, 13, 4 );
			DatagramPacket packet = new DatagramPacket( buffer, buffer.length, target.getIp(), target.getPort() );
			socketManager.sendPacket( packet );
			n.setStartTime( System.currentTimeMillis() );
			if( logger.isInfoEnabled() ){
				logger.info( "REQUEST message for packet " + n.getPacketId() + " sent to " + target.getIp().getHostAddress() + ":" + target.getPort() );
			}
			Node.statistic.incRequestsSentTo( target.getIp().getHostAddress() + ":" + target.getPort() );
		}catch( IOException e ){
			if( logger.isErrorEnabled() ){
				logger.error( "sending REQUEST for packet " + n.getPacketId() + " to " + target.getIp().getHostAddress() + ":" + target.getPort() + " failed" );
			}
			return false;
		}
		return true;
	}


	/**
	 * Stops this <code>RequestTask</code>.
	 */
	public void close(){

		running = false;
		synchronized( this ){
			notify();
		}
	}
}
