/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.tasks;

import java.net.Socket;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.Node;

/**
 * Base class for all Tasks.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see Runnable
 */
public abstract class Task implements Runnable {

	protected final Log logger = LogFactory.getLog( this.getClass() );

	protected Node source = null;
	protected Neighbour target = null;
	protected Socket socket = null;

	private int pingTimeout = 3000;


	/**
	 * Sends a PING message to all neighbours to check their availability. If
	 * one is no longer available, he will be removed from the neighbour list of
	 * the source.
	 * 
	 * @return true if all neighbours are still reachable; false otherwise
	 */
	protected boolean allNeighboursAvailable(){

		PingTask ping;
		boolean result = true;

		if( source == null ){
			return true;
		}
		try{
			// start ping threads
			for( Neighbour curN : source.getNeighboursCopy() ){
				ping = new PingTask( curN, pingTimeout );
				if( !ping.sendPINGreceivePONG() ){
					if( logger.isDebugEnabled() ){
						logger.debug( curN.getIp() + " does not respond to PING" );
					}
					source.removeNeighbour( curN );
					result = false;
				}
			}
		}catch( NodeNotConfiguredException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}
		return result;
	}
}
