/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.tasks;

import ch.ethz.dcg.p2p.exceptions.NodeNotConfiguredException;
import ch.ethz.dcg.p2p.peers.Neighbour;
import ch.ethz.dcg.p2p.peers.Node;

/**
 * Thread for removing dead neighbours.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 * @see Task
 */
public class ZombieKiller extends Task {

	private int interval = 30000; // ms
	private boolean running = false;
	private int pingTimeout = 3000;
	private int retries = 2;


	/**
	 * Creates a new <code>ZombieKiller</code>.
	 * 
	 * @param interval
	 *            the interval in ms between two killing sprees
	 * @param retries
	 *            the number of retries if a neighbour does not reply to the
	 *            PING message
	 * @param source
	 *            the <code>Node</code> that owns this <code>Task</code>
	 */
	public ZombieKiller( int interval, int retries, Node source ){

		if( interval > 0 ){
			this.interval = interval;
		}
		if( retries > 0 ){
			this.retries = retries;
		}
		this.source = source;
	}


	/**
	 * Starts this task.<br>
	 * Every interval milliseconds is a PING message sent to all neighbours. If
	 * a neighbour does not answer after retries PINGs he will be removed.
	 */
	public void run(){

		PingTask ping;
		running = true;

		if( logger.isDebugEnabled() ){
			logger.debug( "ZombieKiller running" );
		}

		while( running ){
			if( logger.isDebugEnabled() ){
				logger.debug( "ZombieKiller is killing zombies" );
			}
			try{
				for( Neighbour curN : source.getNeighboursCopy() ){
					boolean success = false;
					int i = 0;
					while( i < retries && !success ){
						ping = new PingTask( curN, pingTimeout );
						success = ping.sendPINGreceivePONG();
						i++;
					}
					if( !success ){
						source.removeNeighbour( curN );
						source.write2gui( "ZombieKiller exterminated neighbour " + curN.getIp().getHostAddress() + ":" + curN.getPort() + " after " + i + " retries" );
					}
				}
				try{
					Thread.sleep( interval );
				}catch( InterruptedException e ){
				}
			}catch( NodeNotConfiguredException e ){
				if( logger.isErrorEnabled() ){
					logger.error( e.getMessage() );
				}
				return;
			}
		}
	}


	/**
	 * Stop the massacre.
	 */
	public void close(){

		running = false;
	}

}
