/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.utilities;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.dcg.p2p.exceptions.WrongProtocolMessageException;

/**
 * Gather diagnostical and statistical information about the network.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class Diagnostics {

	protected final Log logger = LogFactory.getLog( this.getClass() );

	private ServerSocket serverSocket = null;

	private int maxNeighboursAllowed;
	private int hopTimeout;
	private Hashtable<String, DummyNode> dummyNodes = new Hashtable<String, DummyNode>();


	/**
	 * Creates a new Diagnostics object.
	 * 
	 * @param maxNeighboursAllowed
	 *            upper bound defined by the <code>NetworkEntryPoint</code>
	 *            for the amount of neighbours a peer can have
	 * @param hopTimeout
	 *            timeout in milliseconds between two directly connected peers
	 */
	public Diagnostics( int maxNeighboursAllowed, int hopTimeout ){

		try{
			serverSocket = new ServerSocket( ProtocolConstants.DIAGNOSTICSPORT );
			this.maxNeighboursAllowed = maxNeighboursAllowed;
			this.hopTimeout = hopTimeout;
		}catch( IOException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( "cannot open ServerSocket" );
			}
			System.exit( -1 );
		}
	}


	/**
	 * Tests if the graph is connected. A token is sent around the network to
	 * which each node has to reply with an ack message.
	 * 
	 * @param diagIp
	 *            IP address of the host running the diagnostics
	 * @param firstIp
	 *            IP address of the node that initiates the token sending
	 * @param firstPort
	 *            port of that node
	 * @param floodId
	 *            token id; each token has a unique id in order to support
	 *            several tokens at a time and to prevent duplicate responses
	 * @param total
	 *            the total number of nodes in the network
	 * @return true if the graph is connected; false otherwise.
	 */
	public boolean connected( InetAddress diagIp, InetAddress firstIp, int firstPort, int floodId, int total ){

		Socket sendSocket = null;
		DataOutputStream out = null;

		if( logger.isDebugEnabled() ){
			logger.debug( "connectivity test started with " + total + " nodes" );
		}

		try{
			sendSocket = new Socket( firstIp, firstPort );
			out = new DataOutputStream( sendSocket.getOutputStream() );
			out.write( ProtocolConstants.FLOOD );
			out.writeInt( floodId );
			out.writeByte( ProtocolConstants.CONNECTIVITYTOKEN );
			Utilities.writeIP( diagIp, out );
			out.writeInt( ProtocolConstants.DIAGNOSTICSPORT );
			out.flush();
			if( logger.isDebugEnabled() ){
				logger.debug( "CONNECTIVITYTOKEN message sent to " + firstIp.getHostAddress() );
			}
			boolean connected = receiveConnectivityAcks( total );
			return connected;
		}catch( IOException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( e.getMessage() );
			}
		}catch( WrongProtocolMessageException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}finally{
			try{
				sendSocket.close();
			}catch( Exception e ){
			}
		}
		return false;
	}


	/**
	 * Waits for incoming acks as response to the token. This method blocks
	 * until the expected amount of messages has arrived or a timeout has
	 * expired.
	 * 
	 * @param total
	 *            the total number of nodes in the network
	 * @return true if the number of received acks matches the total number of
	 *         nodes in the network
	 * @throws IOException
	 *             if a communication error occurred
	 * @throws WrongProtocolMessageException
	 *             if an unexpected message was received
	 */
	private boolean receiveConnectivityAcks( int total ) throws IOException, WrongProtocolMessageException{

		Socket receiveSocket = null;
		DataInputStream in = null;
		HashSet<String> acks = new HashSet<String>();

		if( logger.isDebugEnabled() ){
			logger.debug( "waiting for CONNECTIVITYACK messages" );
		}

		int timeout = total * hopTimeout;
		long startTime = System.currentTimeMillis();
		serverSocket.setSoTimeout( hopTimeout );
		while( ( total > 0 ) && ( ( System.currentTimeMillis() - startTime ) <= timeout ) ){
			try{
				receiveSocket = serverSocket.accept();
			}catch( IOException e ){
				// timeout expired probably
				continue;
			}
			in = new DataInputStream( receiveSocket.getInputStream() );
			byte type = in.readByte();
			if( type != ProtocolConstants.CONNECTIVITYACK ){
				throw new WrongProtocolMessageException( "wrong message type received: " + type );
			}
			if( logger.isDebugEnabled() ){
				logger.debug( "CONNECTIVITYACK message received from " + receiveSocket.getInetAddress() );
			}
			String newAck = Utilities.readIP( in ).getHostAddress() + ":" + in.readInt();
			if( acks.contains( newAck ) ){
				if( logger.isWarnEnabled() ){
					logger.warn( "duplicate CONNECTIVITYACK received" );
				}
			}else{
				total--;
				acks.add( newAck );
				if( logger.isDebugEnabled() ){
					logger.debug( total + " CONNECTIVITYACKS missing" );
				}
			}
		}
		if( total == 0 ){
			return true;
		}else{
			return false;
		}
	}


	/**
	 * Sends a REPORTNEIGHBOURS message to nodes asking them to report their
	 * current neighbours. This is done using the flooding mechanism.
	 * 
	 * @param diagIp
	 *            IP of the node that is collecting the reported neighbours
	 * @param firstIp
	 *            starting point of the flooding
	 * @param firstPort
	 *            the corresponding port a node is listening for flood messages
	 * @param floodId
	 *            the id of the current flood message; cannot be used twice or
	 *            the nodes will not answer
	 * @param nNodes
	 *            the number of nodes that are expected to answer
	 */
	public void reportNeighbours( InetAddress diagIp, InetAddress firstIp, int firstPort, int floodId, int nNodes ){

		Socket sendSocket = null;
		DataOutputStream out = null;

		if( logger.isDebugEnabled() ){
			logger.debug( "neighbours reporting started" );
		}

		try{
			sendSocket = new Socket( firstIp, firstPort );
			out = new DataOutputStream( sendSocket.getOutputStream() );
			out.writeByte( ProtocolConstants.FLOOD );
			out.writeInt( floodId );
			out.writeByte( ProtocolConstants.REPORTNEIGHBOURS );
			Utilities.writeIP( diagIp, out );
			out.writeInt( ProtocolConstants.DIAGNOSTICSPORT );
			out.flush();
			if( logger.isDebugEnabled() ){
				logger.debug( "REPORTNEIGHBOURS FLOOD message sent to " + firstIp );
			}
			receiveNeighbours( nNodes );
		}catch( IOException e ){
			if( logger.isFatalEnabled() ){
				logger.fatal( e.getMessage() );
			}
		}catch( WrongProtocolMessageException e ){
			if( logger.isErrorEnabled() ){
				logger.error( e.getMessage() );
			}
		}finally{
			try{
				sendSocket.close();
			}catch( Exception e ){
			}
		}
	}


	/**
	 * Handles incoming NEIGHBOUR messages and prints the reported neighbours
	 * using the logger. The nodes with their respective neighbours are stored
	 * in the dummyNodes vector for further diagnosis.
	 * 
	 * @param nNodes
	 *            the number of nodes that are expected to send an answer; the
	 *            method will block until all of the answers were received.
	 * @throws IOException
	 *             if a communication error occurred
	 * @throws WrongProtocolMessageException
	 *             if an unexpected message was received
	 */
	private void receiveNeighbours( int nNodes ) throws IOException, WrongProtocolMessageException{

		Socket receiveSocket = null;
		DataInputStream in = null;

		if( logger.isDebugEnabled() ){
			logger.debug( "waiting for NEIGHBOURS messages" );
		}

		InetAddress nodeIp, neighbourIp;
		int nodePort, neighbourPort;
		DummyNode dn;
		while( nNodes > 0 ){
			receiveSocket = serverSocket.accept();
			in = new DataInputStream( receiveSocket.getInputStream() );
			byte type = in.readByte();
			if( type != ProtocolConstants.NEIGHBOURS ){
				throw new WrongProtocolMessageException( "wrong message type received: " + type );
			}
			nNodes--;
			nodeIp = Utilities.readIP( in );
			nodePort = in.readInt();
			dn = new DummyNode( nodeIp, nodePort );
			if( logger.isDebugEnabled() ){
				logger.debug( "NEIGHBOURS message received from " + nodeIp + ":" + nodePort + "; " + nNodes + " missing" );
			}
			int n = in.readInt();
			if( ( n < 0 ) || ( n > maxNeighboursAllowed ) ){
				throw new IOException( "invalid number of neighbours received" );
			}
			String s = nodeIp + ": [";
			while( n > 0 ){
				neighbourIp = Utilities.readIP( in );
				neighbourPort = in.readInt();
				dn.addNeighbour( neighbourIp, neighbourPort );
				s += " " + neighbourIp + ":" + neighbourPort;
				n--;
			}
			s += " ]";
			dummyNodes.put( hashKey( dn.getIp(), dn.getPort() ), dn );
			if( logger.isDebugEnabled() ){
				logger.debug( s );
			}
		}
	}


	/**
	 * Checks if the neighbourhood of the nodes in the graph is consistent.
	 * 
	 * @return true if each node is known by all of its neighbours and vice
	 *         versa; false otherwise
	 */
	public boolean checkNeighbourhoodConsistency(){

		// iterate over all nodes
		Iterator<DummyNode> dit = dummyNodes.values().iterator();
		while( dit.hasNext() ){
			DummyNode curN = dit.next();
			// iterate over all neighbours of the nodes
			for( DummyNode neigh : curN.getNeighboursCopy() ){
				if( !checkReverseNeighbourhood( curN, neigh ) ){
					return false;
				}
			}
		}
		return true;
	}


	/**
	 * Checks if neighbour n has dummy node dn as neighbour.
	 * 
	 * @param dn
	 *            the dummy node for which a reverse neighbourhood test should
	 *            be done
	 * @param n
	 *            the neighbour from which we want to know if he has dummy node
	 *            dn as neighbour
	 * @return true if neighbour n has dummy node dn as neighbour; false
	 *         otherwise
	 */
	private boolean checkReverseNeighbourhood( DummyNode dn, DummyNode n ){

		// check if dn is a neighbour of n
		return n.hasNeighbour( dn.getIp(), dn.getPort() );
	}


	/**
	 * Determines the diameter of the network using a round based algorithm.
	 * 
	 * @param firstIp
	 *            IP of the node from which the distance to the farthest node
	 *            should be determined
	 * @param firstPort
	 *            port of the above node
	 * @return the distance from the start node to the node in the network with
	 *         maximum hop count
	 */
	public int calcDiameter( InetAddress firstIp, int firstPort ){

		Socket socket = null;
		DataInputStream in = null;
		DataOutputStream out = null;

		int diameter = -1;
		Vector<DummyNode> done = new Vector<DummyNode>();
		Vector<DummyNode> cur = new Vector<DummyNode>();
		DummyNode first = new DummyNode( firstIp, firstPort );
		cur.add( first );
		done.add( first );

		if( logger.isDebugEnabled() ){
			logger.debug( "diameter calculation started at " + firstIp );
		}

		while( !cur.isEmpty() ){
			diameter++;
			Vector<DummyNode> tmp = new Vector<DummyNode>();
			for( DummyNode curN : cur ){
				// send DIRECTREPORTNEIGHBOURS message to curN
				try{
					socket = new Socket( curN.getIp(), curN.getPort() );
					out = new DataOutputStream( socket.getOutputStream() );
					out.write( ProtocolConstants.DIRECTREPORTNEIGHBOURS );
					out.flush();
					if( logger.isDebugEnabled() ){
						logger.debug( "DIRECTREPORTNEIGHBOURS message sent to " + curN.getIp().getHostAddress() + ":" + curN.getPort() );
					}
				}catch( IOException e ){
					if( logger.isErrorEnabled() ){
						logger.error( "cannot send DIRECTREPORTNEIGHBOURS message to " + curN.getIp().getHostAddress() + ":" + curN.getPort() + ": " + e.getMessage() );
					}
					return 0;
				}
				// receive neighbours of curN
				try{
					in = new DataInputStream( socket.getInputStream() );
					byte type = in.readByte();
					if( type != ProtocolConstants.DIRECTNEIGHBOURS ){
						if( logger.isErrorEnabled() ){
							logger.error( "wrong message type received: " + type );
						}
						return 0;
					}
					int n = in.readInt();
					if( ( n < 0 ) || ( n > maxNeighboursAllowed ) ){
						if( logger.isErrorEnabled() ){
							logger.error( "invalid amount of neighbours received from " + socket.getInetAddress().getHostAddress() );
						}
						return 0;
					}
					InetAddress neighbourIp;
					int neighbourPort;
					if( logger.isDebugEnabled() ){
						logger.debug( "DIRECTNEIGHBOURS message received from " + curN.getIp().getHostAddress() + ":" + curN.getPort() );
					}
					while( n > 0 ){
						neighbourIp = Utilities.readIP( in );
						neighbourPort = in.readInt();
						if( !neighbourVectorContains( done, neighbourIp, neighbourPort ) ){
							DummyNode tmpNeigh = new DummyNode( neighbourIp, neighbourPort );
							tmp.add( tmpNeigh );
							done.add( tmpNeigh );
						}
						n--;
					}
				}catch( IOException e ){
					if( logger.isErrorEnabled() ){
						logger.error( "cannot receive DIRECTNEIGHBOURS message from " + curN.getIp() + ": " + e.getMessage() );
					}
				}
			}
			cur = tmp;
		}
		return diameter;
	}


	/**
	 * Test if a given IP and port combination is stored as a dummy node in a
	 * given <code>Vector</code>
	 * 
	 * @param v
	 *            the haystack
	 * @param ip
	 *            the needle
	 * @param port
	 *            the neele
	 * @return true if the haystack contains the needle; false otherwise
	 */
	private boolean neighbourVectorContains( Vector<DummyNode> v, InetAddress ip, int port ){

		for( DummyNode curN : v ){
			if( curN.getIp().equals( ip ) && ( curN.getPort() == port ) ){
				return true;
			}
		}
		return false;
	}


	/**
	 * Combine the IP address and the port number to a hashkey of the form
	 * 'IP:port'
	 * 
	 * @param ip
	 *            IP address for the hashkey
	 * @param port
	 *            port for the hashkey
	 * @return a <code>String</code> representation of the hashkey
	 */
	private String hashKey( InetAddress ip, int port ){

		return ip.getHostAddress() + ":" + port;
	}

}
