/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.utilities;

import java.net.InetAddress;
import java.util.ArrayList;
import java.util.Hashtable;

/**
 * Minimal representation of a node.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class DummyNode {

	private InetAddress ip;
	private int port;

	private Hashtable<String, DummyNode> neighbours = new Hashtable<String, DummyNode>();


	/**
	 * Creates a new <code>DummyNode</code> with the given IP and port
	 * 
	 * @param ip
	 *            IP address of the new node
	 * @param port
	 *            port number of the new node
	 */
	public DummyNode( InetAddress ip, int port ){

		this.ip = ip;
		this.port = port;
	}


	/**
	 * Adds a neighbour to this node.
	 * 
	 * @param ip
	 *            IP of the neighbour node
	 * @param port
	 *            port of the neighbour node
	 */
	public void addNeighbour( InetAddress ip, int port ){

		synchronized( neighbours ){
			neighbours.put( hashKey( ip, port ), new DummyNode( ip, port ) );
		}
	}


	/**
	 * Removes a neighbour from this node.
	 * 
	 * @param ip
	 *            IP of the neighbour node
	 * @param port
	 *            port of the neighbour node
	 */
	public void removeNeighbour( InetAddress ip, int port ){

		synchronized( neighbours ){
			neighbours.remove( hashKey( ip, port ) );
		}
	}


	/**
	 * Checks if this node has a node with the given IP and port combination
	 * among his neighbours.
	 * 
	 * @param ip
	 *            IP of the neighbour node in question
	 * @param port
	 *            port of the neighbour node in question
	 */
	public boolean hasNeighbour( InetAddress ip, int port ){

		synchronized( neighbours ){
			return neighbours.containsKey( hashKey( ip, port ) );
		}
	}


	/**
	 * Returns a copy of the set of neighbours.
	 * 
	 * @return a copy of the set of neighbours.
	 */
	public ArrayList<DummyNode> getNeighboursCopy(){

		synchronized( neighbours ){
			return new ArrayList<DummyNode>( neighbours.values() );
		}
	}


	/**
	 * Combine the IP address and the port number to a hashkey of the form
	 * 'IP:port'
	 * 
	 * @param ip
	 *            IP address for the hashkey
	 * @param port
	 *            port for the hashkey
	 * @return a <code>String</code> representation of the hashkey
	 */
	private String hashKey( InetAddress ip, int port ){

		return ip.getHostAddress() + ":" + port;
	}


	// GETTERS

	public InetAddress getIp(){

		return ip;
	}


	public int getPort(){

		return port;
	}
}
