/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.utilities;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Calendar;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

/**
 * Mechanism to track the status of the network.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class Logger {

	// NEP
	private InetAddress nepIp;
	private int nepPort;

	// totals
	private int nPeersTotal = 0;
	private int nLevels = 0;
	private int[] averageTotalNeighCount = new int[2];
	private float averageTotalOutSpeed = 0;
	private float[] averageTotalRTT = new float[2];
	private int nInitNeighs = 0;
	private int minNeighs = 0;
	private int maxNeighs = 0;

	// per level
	private Hashtable<Integer, LevelStatistic> perLevelStats = new Hashtable<Integer, LevelStatistic>();

	// diagnostics
	private Diagnostics diag = new Diagnostics( 10000, 2000 );
	private InetAddress myIp = null;
	private int floodId = 0;

	// file
	private String fileEnding = ".log";
	private String baseNameLevel = "Level";
	private String nameTotal = "LevelTotals" + fileEnding;


	/**
	 * Creates a new <code>Logger</code>
	 * 
	 * @param nepIp
	 *            IP address of the <code>NetworkEntryPoint</code>
	 * @param nepPort
	 *            port of the <code>NetworkEntryPoint</code>
	 * @param myIp
	 *            IP address of this host
	 * @param startFloodId
	 *            the first flood ID to use (incremented automatically
	 *            afterwards)
	 */
	public Logger( InetAddress nepIp, int nepPort, InetAddress myIp, int startFloodId ){

		this.nepIp = nepIp;
		this.nepPort = nepPort;
		this.myIp = myIp;
		this.floodId = startFloodId;
	}


	/**
	 * Starts this <code>Logger</code>.
	 * 
	 * @param args
	 *            the arguments are [0]nepIp [1]nepPort [2]myIp [3]startFloodId
	 *            [4]interval
	 */
	public static void main( String[] args ){

		InetAddress nepIp;
		int nepPort;
		InetAddress myIp;
		int startFloodId;
		int interval; // seconds
		Logger logger = null;

		if( args.length < 5 ){
			printUsage();
			return;
		}

		try{
			nepIp = InetAddress.getByName( args[0] );
			nepPort = Integer.parseInt( args[1] );
			myIp = InetAddress.getByName( args[2] );
			startFloodId = Integer.parseInt( args[3] );
			interval = Integer.parseInt( args[4] );
			logger = new Logger( nepIp, nepPort, myIp, startFloodId );
		}catch( NumberFormatException e ){
			System.out.println( e.getMessage() );
			printUsage();
			return;
		}catch( UnknownHostException e ){
			System.out.println( e.getMessage() );
			printUsage();
			return;
		}

		while( true ){
			System.out.println( "*** LOGGING STARTED (every " + interval + "s) *********" );
			// NEP
			try{
				logger.getNEPInfos();
			}catch( IOException e ){
				System.out.println( "IOException at NEP: " + e.getMessage() );
				System.exit( -1 );
			}
			// Levels with Peers
			logger.getLevelInfos();
			// print to console
			logger.printInfos();
			// print to files
			logger.writeTotals2File();
			Iterator<Integer> it = logger.perLevelStats.keySet().iterator();
			while( it.hasNext() ){
				logger.writeLevel2File( it.next() );
			}
			System.out.println( "*** TAKING TOPOLOGY SNAPSHOT ************" );
			logger.networkTopologySnapshot();
			System.out.println( "*****************************************" );
			logger.reset();
			System.out.println( "*** LOGGING FINISHED (ID=" + ( logger.floodId - 1 ) + ") ************\n" );
			try{
				Thread.sleep( interval * 1000 );
			}catch( InterruptedException e ){
			}

		}
	}


	/**
	 * Prints the usage to the console.
	 */
	private static void printUsage(){

		System.out.println( "USAGE: java Logger <NEP-IP> <NEP-Port> <Logger-IP> <firstFloodId> <interval>" );
		System.out.println( "<NEP-IP>: IP address of the Network Entry Point" );
		System.out.println( "<NEP-Port>: port number" );
		System.out.println( "<Logger-IP>: IP address of the host that is logging (this host)" );
		System.out.println( "<firstFloodId>: number that represents the first flood id; peers will drop connectivity- and diametermessages if they receive a duplicate flood id" );
		System.out.println( "<interval>: the interval in seconds between the logging" );
	}


	/**
	 * Resets the logger to the start values.
	 */
	private void reset(){

		nPeersTotal = 0;
		nLevels = 0;
		averageTotalNeighCount = new int[2];
		averageTotalOutSpeed = 0;
		averageTotalRTT = new float[2];
		nInitNeighs = 0;
		minNeighs = 0;
		maxNeighs = 0;
		perLevelStats = new Hashtable<Integer, LevelStatistic>();
	}


	/**
	 * Gather information from the <code>NetworkEntryPoint</code>.
	 * 
	 * @throws IOException
	 *             if a communication error occurred
	 */
	private void getNEPInfos() throws IOException{

		// open TCP connection
		Socket s = new Socket( nepIp, nepPort );
		s.setSoTimeout( 5000 );
		// send request
		OutputStream out = s.getOutputStream();
		out.write( ProtocolConstants.REPORTSTATISTIC );
		// get answer
		DataInputStream in = new DataInputStream( s.getInputStream() );
		byte type = in.readByte();
		if( type != ProtocolConstants.NEPSTATISTIC ){
			System.out.println( "invalid answer received from NEP - exiting" );
			System.exit( -1 );
		}
		nInitNeighs = in.readInt();
		maxNeighs = in.readInt();
		minNeighs = in.readInt();
		nLevels = in.readInt();
		// get per-level-infos
		for( int level = 0; level < nLevels; level++ ){
			LevelStatistic ls = new LevelStatistic();
			ls.bitrate = in.readInt();
			ls.nPeers = in.readInt();
			nPeersTotal += ls.nPeers;
			for( int peer = 0; peer < ls.nPeers; peer++ ){
				ls.peers.add( new DummyNode( Utilities.readIP( in ), in.readInt() ) );
			}
			ls.averageOutSpeed = in.readFloat();
			perLevelStats.put( ls.bitrate, ls );
		}
		averageTotalOutSpeed = in.readFloat();
		s.close();
	}


	/**
	 * Gather information about each individual level.
	 */
	private void getLevelInfos(){

		// each level
		Iterator<Integer> it = perLevelStats.keySet().iterator();
		while( it.hasNext() ){
			int levelSpeed = it.next();
			LevelStatistic ls = perLevelStats.get( levelSpeed );
			int nPeers = ls.peers.size();
			// each peer in this level
			for( int i = 0; i < nPeers; i++ ){
				DummyNode curN = ls.peers.elementAt( i );
				try{
					getPeerInfos( curN.getIp(), curN.getPort(), perLevelStats.get( levelSpeed ) );
				}catch( IOException e ){
					System.out.println( "IOException: cannot get peer infos from " + curN.getIp().getHostAddress() + ":" + curN.getPort() );
				}
			}
			ls.connected = diag.connected( myIp, ls.peers.firstElement().getIp(), ls.peers.firstElement().getPort(), floodId++, nPeers );
			ls.diameter = diag.calcDiameter( ls.peers.firstElement().getIp(), ls.peers.firstElement().getPort() );
		}
	}


	/**
	 * Gather information about a specific peer.
	 * 
	 * @param ip
	 *            IP address of the target peer
	 * @param port
	 *            port of the target peer
	 * @param ls
	 *            the <code>LevelStatistic</code> to update with the new
	 *            information
	 * @throws IOException
	 *             if a communication error occurred
	 */
	private void getPeerInfos( InetAddress ip, int port, LevelStatistic ls ) throws IOException{

		// open TCP connection
		Socket s = new Socket( ip, port );
		s.setSoTimeout( 4000 );
		// send request
		OutputStream out = s.getOutputStream();
		out.write( ProtocolConstants.REPORTSTATISTIC );
		// get answer
		DataInputStream in = new DataInputStream( s.getInputStream() );
		// TYPE
		byte type = in.readByte();
		if( type != ProtocolConstants.PEERSTATISTIC ){
			System.out.println( "invalid answer received from peer " + ip.getHostAddress() + ":" + port + " - exiting" );
			System.exit( -1 );
		}
		// nNeighs
		int nNeighs = in.readInt();
		ls.averageNeighbourCount[1] += nNeighs;
		ls.averageNeighbourCount[0] += 1;
		averageTotalNeighCount[1] += nNeighs;
		averageTotalNeighCount[0] += 1;
		// packetSize
		int packetSize = in.readInt();
		if( packetSize > 0 && ls.packetSize == 0 ){
			ls.packetSize = packetSize;
		}
		// averageRTT
		float averageRTT = in.readFloat();
		if( averageRTT > 0 ){
			averageTotalRTT[1] += averageRTT;
			averageTotalRTT[0] += 1;
			ls.averageRTT[1] += averageRTT;
			ls.averageRTT[0] += 1;
		}
	}


	/**
	 * Prints the gathered informations
	 */
	private void printInfos(){

		System.out.println( "*** T O T A L S *************************" );
		System.out.println( "nLevels:                " + nLevels );
		System.out.println( "nPeersTotal:            " + nPeersTotal );
		System.out.println( "averageTotalNeighCount: " + calcAverage( averageTotalNeighCount ) );
		System.out.println( "averageTotalOutSpeed:   " + averageTotalOutSpeed );
		System.out.println( "averageTotalRTT:        " + calcAverage( averageTotalRTT ) );
		System.out.println( "nInitNeighs:            " + nInitNeighs );
		System.out.println( "minNeighs:              " + minNeighs );
		System.out.println( "maxNeighs:              " + maxNeighs );
		System.out.println( "*****************************************" );

		Iterator<Integer> it = perLevelStats.keySet().iterator();
		while( it.hasNext() ){
			int curLevel = it.next();
			System.out.println( "*** L E V E L  " + curLevel + " **********************" );
			LevelStatistic ls = perLevelStats.get( curLevel );
			System.out.println( "bitrate [kbps]:         " + ls.bitrate );
			System.out.println( "packetSize [bytes]:     " + ls.packetSize );
			System.out.println( "nPeers:                 " + ls.nPeers );
			System.out.println( "averageNeighbourCount:  " + calcAverage( ls.averageNeighbourCount ) );
			System.out.println( "averageOutSpeed:        " + ls.averageOutSpeed );
			System.out.println( "averageRTT:             " + calcAverage( ls.averageRTT ) );
			System.out.println( "connected:              " + ls.connected );
			System.out.println( "diameter:               " + ls.diameter );
			System.out.println( "*****************************************" );
		}
	}


	/**
	 * Calculates average values from gathered float values.
	 * 
	 * @param x
	 *            float array containing the number of values at [0] and the sum
	 *            of them in [1]
	 * @return the calculated average
	 */
	private float calcAverage( float[] x ){

		if( x[0] == 0 ){
			return -1;
		}else{
			return (float)x[1] / (float)x[0];
		}
	}


	/**
	 * Calculates average values from gathered int values.
	 * 
	 * @param x
	 *            int array containing the number of values at [0] and the sum
	 *            of them in [1]
	 * @return the calculated average
	 */
	private float calcAverage( int[] x ){

		if( x[0] == 0 ){
			return -1;
		}else{
			return (float)x[1] / (float)x[0];
		}
	}


	/**
	 * Writes the gathered information about a level into a file.
	 * 
	 * @param levelSpeed
	 *            the speed of the level to be stored
	 */
	private void writeLevel2File( int levelSpeed ){

		String fileName = baseNameLevel + levelSpeed + fileEnding;
		File file = new File( fileName );
		boolean fileExists = file.exists();
		PrintWriter writer = null;
		try{
			writer = new PrintWriter( new FileWriter( file, true ) );
			if( !fileExists ){
				writer.println( getLevelSignature() );
			}
			writer.println( getTimeStamp() + getLevelLog( levelSpeed ) );
		}catch( IOException e ){
			System.out.println( "Warning: Cannot write level " + levelSpeed + " to file!" );
		}finally{
			try{
				writer.close();
			}catch( Exception e ){
			}
		}
	}


	/**
	 * Writes the gathered total information into a file.
	 */
	private void writeTotals2File(){

		String fileName = nameTotal;
		File file = new File( fileName );
		boolean fileExists = file.exists();
		PrintWriter writer = null;
		try{
			writer = new PrintWriter( new FileWriter( file, true ) );
			if( !fileExists ){
				writer.println( getTotalSignature() );
			}
			writer.println( getTimeStamp() + getTotalLog() );
		}catch( IOException e ){
			System.out.println( "Warning: Cannot write log of totals to file!" );
		}finally{
			try{
				writer.close();
			}catch( Exception e ){
			}
		}
	}


	/**
	 * Gets the explanation of the individual values in the per level recording.
	 * 
	 * @return a <code>String</code> describing what value means what
	 */
	private String getLevelSignature(){

		return "[ TIMESTAMP ] - 0:bitrate,1:packetSize,2:nPeers,3:averageNeighbourCount,4:averageOutSpeed,5:averageRTT,6:connected,7:diameter";
	}


	/**
	 * Gets the explanation of the individual values in the totals recording.
	 * 
	 * @return a <code>String</code> describing what value means what
	 */
	private String getTotalSignature(){

		return "[ TIMESTAMP ] - 0:nLevels,1:nPeersTotal,2:averageTotalNeighCount,3:averageTotalOutSpeed,4:averageTotalRTT,5:nInitNeighs,6:minNeighs,7:maxNeighs";
	}


	/**
	 * Gets the explanation of the individual values in the neighbour recording.
	 * 
	 * @return a <code>String</code> describing what value means what
	 */
	private String getNeighboursSignature(){

		return "[ TIMESTAMP ] - [ID] - {peer;neighbour_1,neighbour_2,...,neighbour_n}";
	}


	/**
	 * Gets a line to insert into the file which contains the recorded values.
	 * 
	 * @param levelSpeed
	 *            the speed of the level to record
	 * @return a <code>String</code> which contains the gathered information
	 *         in a compact way
	 */
	private String getLevelLog( int levelSpeed ){

		LevelStatistic ls = perLevelStats.get( levelSpeed );
		String line = "";
		line += "0:" + ls.bitrate + ",";
		line += "1:" + ls.packetSize + ",";
		line += "2:" + ls.nPeers + ",";
		line += "3:" + calcAverage( ls.averageNeighbourCount ) + ",";
		line += "4:" + ls.averageOutSpeed + ",";
		line += "5:" + calcAverage( ls.averageRTT ) + ",";
		line += "6:" + ls.connected + ",";
		line += "7:" + ls.diameter;
		return line;
	}


	/**
	 * Gets a line to insert into the file which contains the recorded total
	 * values.
	 * 
	 * @return a <code>String</code> which contains the gathered information
	 *         in a compact way
	 */
	private String getTotalLog(){

		String line = "";
		line += "0:" + nLevels + ",";
		line += "1:" + nPeersTotal + ",";
		line += "2:" + calcAverage( averageTotalNeighCount ) + ",";
		line += "3:" + averageTotalOutSpeed + ",";
		line += "4:" + calcAverage( averageTotalRTT ) + ",";
		line += "5:" + nInitNeighs + ",";
		line += "6:" + minNeighs + ",";
		line += "7:" + maxNeighs;
		return line;
	}


	/**
	 * Gets the current time formatted as a human readable time stamp.
	 * 
	 * @return a String representation of the time stamp
	 */
	private String getTimeStamp(){

		return "[ " + Calendar.getInstance().getTime().toString() + " ] - ";
	}


	/**
	 * Records a snapshot of the current network topology.
	 */
	private void networkTopologySnapshot(){

		// iterate over all levels
		Iterator<Integer> it = perLevelStats.keySet().iterator();
		while( it.hasNext() ){
			int curLevel = it.next();
			LevelStatistic ls = perLevelStats.get( curLevel );
			// iterate over all peers
			for( int i = 0; i < ls.peers.size(); i++ ){
				DummyNode curP = ls.peers.elementAt( i );
				Vector<DummyNode> neighs = getNeighboursOf( curP );
				writeNeighbours2File( floodId, curLevel, curP, neighs );
			}
		}
	}


	/**
	 * Asks a node about his neighbours.
	 * 
	 * @param p
	 *            the node from which we want to learn about his neighbours
	 * @return a <code>Vector</code> containing the neighbours of the argument
	 *         node
	 */
	private Vector<DummyNode> getNeighboursOf( DummyNode p ){

		try{
			Socket socket = new Socket( p.getIp(), p.getPort() );
			socket.setSoTimeout( 3000 );
			DataOutputStream out = new DataOutputStream( socket.getOutputStream() );
			DataInputStream in = new DataInputStream( socket.getInputStream() );
			out.write( ProtocolConstants.DIRECTREPORTNEIGHBOURS );
			byte type = in.readByte();
			if( type != ProtocolConstants.DIRECTNEIGHBOURS ){
				System.out.println( "DIRECTNEIGHBOURS answer from " + p.getIp().getHostAddress() + ":" + p.getPort() + " invalid" );
				return new Vector<DummyNode>( 0 );
			}else{
				int nNeighs = in.readInt();
				Vector<DummyNode> neighs = new Vector<DummyNode>( nNeighs );
				for( int i = 0; i < nNeighs; i++ ){
					neighs.add( new DummyNode( Utilities.readIP( in ), in.readInt() ) );
				}
				return neighs;
			}
		}catch( IOException e ){
			return new Vector<DummyNode>( 0 );
		}
	}


	/**
	 * Writes the neighbours of the peers into a file.
	 * 
	 * @param iteration
	 *            the number of the iteration
	 * @param levelSpeed
	 *            the speed of the level of these nodes (for filename only)
	 * @param node
	 *            the nodes from wich the neighbours should be recorded
	 * @param neighs
	 *            his neighbours
	 */
	private void writeNeighbours2File( int iteration, int levelSpeed, DummyNode node, Vector<DummyNode> neighs ){

		String fileName = baseNameLevel + levelSpeed + "Neighbours" + fileEnding;
		File file = new File( fileName );
		boolean fileExists = file.exists();
		PrintWriter writer = null;
		try{
			writer = new PrintWriter( new FileWriter( file, true ) );
			if( !fileExists ){
				writer.println( getNeighboursSignature() );
			}
			String neighsString = getTimeStamp() + "[" + iteration + "] - {" + node.getIp().getHostAddress() + ":" + node.getPort() + ";";
			for( int i = 0; i < neighs.size(); i++ ){
				DummyNode curN = neighs.elementAt( i );
				neighsString += curN.getIp().getHostAddress() + ":" + curN.getPort();
				if( i < ( neighs.size() - 1 ) ){
					neighsString += ",";
				}
			}
			neighsString += "}";
			writer.println( neighsString );
		}catch( IOException e ){
			System.out.println( "Warning: Cannot write neighbours of  " + node.getIp().getHostAddress() + ":" + node.getPort() + " to file!" );
		}finally{
			try{
				writer.close();
			}catch( Exception e ){
			}
		}
	}
}
