/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.utilities;

import java.net.InetAddress;
import java.util.Iterator;
import java.util.Vector;

import ch.ethz.dcg.p2p.peers.Neighbour;

/**
 * This class implements a heap. The neighbour with the LOWEST ranking is at the
 * HEAD of the heap.
 * 
 *  --> DEPRECATED <--
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class NeighbourHeap implements Iterable<Neighbour> {

	Vector<Neighbour> elements;


	public NeighbourHeap( NeighbourHeap n ){

		this.elements = new Vector<Neighbour>( n.elements );
	}


	public NeighbourHeap(){

		elements = new Vector<Neighbour>();
	}


	/**
	 * Insert a neighbour in order according to his ranking. The slowest one is
	 * at index 0.
	 * 
	 * @param n
	 *            the Neighbour that should be inserted into the heap.
	 * @return true if the neighbour was inserted successfully; false otherwise.
	 */
	public boolean insert( Neighbour n ){

		// check if neighbour already exists
		if( elements.contains( n ) ){
			return false;
		}

		// check if heap is empty
		if( elements.size() == 0 ){
			elements.add( n );
			return true;
		}

		// look for first neighbour that is faster than the new one
		for( int i = 0; i < elements.size(); i++ ){
			if( elements.elementAt( i ).getRanking() >= n.getRanking() ){
				// insert new neighbour at position i and shift rest towards the
				// end
				elements.insertElementAt( n, i );
				return true;
			}
		}

		// neighbour has a new highest ranking -> append it to the end
		elements.add( n );
		return true;
	}


	/**
	 * Remove a neighbour from the heap.
	 * 
	 * @param n
	 *            the Neighbour that should be removed from the heap.
	 * @return true if the neighbour was a component of this heap; false
	 *         otherwise.
	 */
	public boolean remove( Neighbour n ){

		return elements.removeElement( n );
	}


	/**
	 * Removes the head of this heap which is the neighbour with the lowest
	 * ranking.
	 * 
	 * @return the neighbour with the lowest ranking contained in this heap.
	 */
	public Neighbour removeWorst(){

		if( elements.isEmpty() ){
			return null;
		}else{
			return elements.remove( 0 );
		}
	}


	/**
	 * Return the index of the first Neighbour in the heap that has a ranking
	 * equal to or higher than the given ranking
	 * 
	 * @param ranking
	 *            the ranking for which the closest neighbour should be found.
	 * @return index into the heap where this neighbour resides.
	 */
	public int findClosest( int ranking ){

		// find first neighbour that has equal or better ranking than the given
		// one
		for( int i = 0; i < elements.size(); i++ ){
			if( elements.elementAt( i ).getRanking() >= ranking ){
				return i;
			}
		}

		// neighbour is faster than all neighbours in the heap -> return last
		// neighbour
		return elements.size() - 1;
	}


	/**
	 * Return number of neighbours currently in the heap.
	 * 
	 * @return the number of neighbours in this heap.
	 */
	public int size(){

		return elements.size();
	}


	/**
	 * Return neighbour at position i.
	 * 
	 * @param i
	 *            an index into this heap
	 * @return the neighbour at position i
	 */
	public Neighbour elementAt( int i ){

		return elements.elementAt( i );
	}


	/**
	 * Check if current neighbour already exists.
	 * 
	 * @param n
	 *            the neighbour to look for.
	 * @return true if the neighbours is contained in the heap; false otherwise.
	 */
	public boolean contains( Neighbour n ){

		for( Neighbour curN : elements ){
			if( ( n.getIp().equals( curN.getIp() ) && ( n.getPort() == curN.getPort() ) ) ){
				return true;
			}
		}
		return false;
	}


	/**
	 * Check if current neighbour already exists.
	 * 
	 * @param ip
	 *            the ip address to look for
	 * @param port
	 *            the port to look for
	 * @return true if there is a neighbour in the heap with the given ip
	 *         address and port
	 */
	public boolean contains( InetAddress ip, int port ){

		for( Neighbour curN : elements ){
			if( curN.getIp().equals( ip ) && curN.getPort() == port ){
				return true;
			}
		}
		return false;
	}


	/**
	 * Check if the heap is currently empty.
	 * 
	 * @return true if the heap contains no elements; false otherwise.
	 */
	public boolean isEmpty(){

		return elements.isEmpty();
	}


	/**
	 * Part of the Iterable interface.
	 * 
	 * @return an iterator over the heap.
	 */
	public Iterator<Neighbour> iterator(){

		return elements.iterator();
	}

	
	public int updateRTTOf( Neighbour toUpdate, int newRTT ){
		
		for( Neighbour curN : elements ){
			if( curN.getIp().equals( toUpdate.getIp() ) && curN.getPort() == toUpdate.getPort() ){
				return curN.updateRTT( newRTT );
			}
		}
		return -1;
	}
}
