/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.utilities;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Hashtable;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.dcg.p2p.exceptions.PersistencyManagerException;
import ch.ethz.dcg.p2p.peers.NetworkEntryPoint;

/**
 * Helper class to store the state of the <code>NetworkEntryPoint</code> into
 * a file.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class PersistencyManager {

	private final Log logger = LogFactory.getLog( this.getClass() );

	private String fileName = "nepState.p2p";
	private File nepStateFile;
	private FileOutputStream fileOut;
	private FileInputStream fileIn;
	private ObjectOutputStream objectOut;
	private ObjectInputStream objectIn;

	private NetworkEntryPoint nep;

	private boolean doUpdate = true;


	/**
	 * Creates a new <code>PersistencyManager</code>.
	 * 
	 * @param nep
	 *            the <code>NetworkEntryPoint</code> whose state should be
	 *            written into a file
	 */
	public PersistencyManager( NetworkEntryPoint nep ){

		nepStateFile = new File( fileName );
		this.nep = nep;
	}


	/**
	 * Opens an <code>ObjectOutputStream</code> to a file.
	 * 
	 * @throws IOException
	 *             if the stream cannot be opened
	 */
	private void openWrite() throws IOException{

		if( !nepStateFile.exists() ){
			nepStateFile.createNewFile();
		}
		fileOut = new FileOutputStream( nepStateFile );
		objectOut = new ObjectOutputStream( fileOut );
	}


	/**
	 * Opens an <code>ObjectInputStream</code> to a file.
	 * 
	 * @throws IOException
	 *             if the stream cannot be opened
	 */
	private void openRead() throws IOException{

		if( !nepStateFile.exists() ){
			throw new IOException( "file " + nepStateFile.getName() + " does not exist; cannot open it for reading" );
		}
		fileIn = new FileInputStream( nepStateFile );
		objectIn = new ObjectInputStream( fileIn );
	}


	/**
	 * Closes the <code>ObjectOutputStream</code>.
	 * 
	 * @throws IOException
	 *             if the stream cannot be closed
	 */
	private void closeWrite() throws IOException{

		objectOut.close();
		fileOut.close();
	}


	/**
	 * Closes the <code>ObjectInputStream</code>.
	 * 
	 * @throws IOException
	 *             if the stream cannot be closed
	 */
	private void closeRead() throws IOException{

		objectIn.close();
		fileIn.close();
	}


	/**
	 * Writes the current state into the file.
	 * 
	 * @throws PersistencyManagerException
	 *             if the state cannot be written
	 */
	public void writeCurrentState() throws PersistencyManagerException{

		if( !doUpdate ){
			return;
		}
		try{
			openWrite();
			synchronized( nep ){
				// (1) Hashtable<Integer, SortedNeighbourList>
				objectOut.writeObject( nep.getLevels() );
				// (2) int minSpeed
				objectOut.writeInt( nep.getMinSpeed() );
				// (3) byte streamType
				objectOut.writeByte( nep.getStreamType() );
				// (4) int nInitNeighbours
				objectOut.writeInt( nep.getNInitNeighbours() );
				// (5) int minNeighs
				objectOut.writeInt( nep.getMinNeighs() );
				// (6) int maxNeighs
				objectOut.writeInt( nep.getMaxNeighs() );
			}
			closeWrite();
		}catch( Exception e ){
			throw new PersistencyManagerException( "writing state failed: " + e.getMessage() );
		}
		if( logger.isDebugEnabled() ){
			logger.debug( "NetworkEntryPoint state written persistently" );
		}
	}


	/**
	 * Read the state from an existing file and restore it.
	 * 
	 * @throws PersistencyManagerException
	 *             if the saved state cannot be restored
	 */
	public void recoverLastState() throws PersistencyManagerException{

		try{
			doUpdate = false;
			openRead();
			synchronized( nep ){
				// (1) Hashtable<Integer, SortedNeighbourList>
				nep.setLevels( (Hashtable<Integer, SortedNeighbourList>)objectIn.readObject() );
				// (2) int minSpeed
				nep.setMinSpeed( objectIn.readInt() );
				// (3) byte streamType
				nep.setStreamType( objectIn.readByte() );
				// (4) int nInitNeighbours
				nep.setNInitNeighbours( objectIn.readInt() );
				// (5) int minNeighs
				nep.setMinNeighs( objectIn.readInt() );
				// (6) int maxNeighs
				nep.setMaxNeighs( objectIn.readInt() );
			}
			closeRead();
		}catch( Exception e ){
			throw new PersistencyManagerException( "recovering failed: " + e.getMessage() );
		}finally{
			doUpdate = true;
		}
		if( logger.isDebugEnabled() ){
			logger.debug( "NetworkEntryPoint state recovered" );
		}
	}


	/**
	 * Checks if there is a file from which a recovery would be possible.
	 * 
	 * @return true if a file to recover from exists; false otherwise
	 */
	public boolean recoveryPossible(){

		return nepStateFile.exists();
	}
}
