/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.utilities;

import java.io.Serializable;
import java.net.InetAddress;
import java.util.Vector;

import ch.ethz.dcg.p2p.peers.NEPNeighbour;

/**
 * This class implements a sorted list. The NEPNeighbour with the greatest speed
 * is at the END of the list.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class SortedNeighbourList implements Serializable {

	Vector<NEPNeighbour> elements = new Vector<NEPNeighbour>();


	/**
	 * Insert <code>NEPNeighbour</code> n in order according to his speed.
	 * 
	 * @param n
	 *            the <code>NEPNeighbour</code> to insert
	 * @return true if the <code>NEPNeighbour</code> was inserted
	 *         successfully; false otherwise
	 */
	public boolean insert( NEPNeighbour n ){

		// check if node already exists
		if( contains( n ) ){
			return false;
		}

		// check if vector is empty
		if( elements.size() == 0 ){
			elements.add( n );
			return true;
		}

		// look for first node that is faster than the new one
		for( int i = 0; i < elements.size(); i++ ){
			if( ( elements.elementAt( i ).getSpeed() >= n.getSpeed() ) ){
				// insert new node at position i and shift rest towards the end
				elements.insertElementAt( n, i );
				return true;
			}
		}

		// element has a new greatest speed -> append it at the end
		elements.add( n );
		return true;
	}


	/**
	 * Remove <code>NEPNeighbour</code> n from the vector.
	 * 
	 * @param n
	 *            the <code>NEPNeighbour</code> to remove
	 * @return true if it was removed; false otherwise
	 */
	public boolean remove( NEPNeighbour n ){

		return elements.removeElement( n );
	}


	/**
	 * Return the index of the first <code>NEPNeighbour</code> in the list
	 * that has a speed equal to or higher than the given speed.
	 * 
	 * @param speed
	 *            the speed which should be matched
	 * @return the node with a speed as close as possible to the argument
	 */
	public int findClosest( int speed ){

		// find first element that has equal or greater speed than the given
		// one
		for( int i = 0; i < elements.size(); i++ ){
			if( elements.elementAt( i ).getSpeed() >= speed ){
				return i;
			}
		}

		// node is faster than all elements in the vector -> return last element
		return elements.size() - 1;
	}


	/**
	 * Return number of <code>NEPNeighbour</code> currently in the list.
	 * 
	 * @return the number of contained <code>NEPNeighbour</code>s
	 */
	public int size(){

		return elements.size();
	}


	/**
	 * Return element at position i.
	 * 
	 * @param i
	 *            the position of the requested element
	 * @return the <code>NEPNeighbour</code> at position i
	 */
	public NEPNeighbour elementAt( int i ){

		return elements.elementAt( i );
	}


	/**
	 * Check if current <code>NEPNeighbour</code> already exists.
	 * 
	 * @return true if the <code>NEPNeighbour</code> exists; false otherwise
	 */
	public boolean contains( NEPNeighbour n ){

		for( NEPNeighbour curN : elements ){
			if( ( n.getIp().equals( curN.getIp() ) && ( n.getPort() == curN.getPort() ) ) ){
				return true;
			}
		}
		return false;
	}


	/**
	 * Check if the <code>NEPNeighbour</code> with given IP and port exists.
	 * 
	 * @param ip
	 *            IP address in question
	 * @param port
	 *            port in question
	 * @return the <code>NEPNeigbhour</code> if it exists; null otherwise
	 */
	public NEPNeighbour contains( InetAddress ip, int port ){

		for( NEPNeighbour curN : elements ){
			if( curN.getIp().equals( ip ) && ( curN.getPort() == port ) ){
				return curN;
			}
		}
		return null;
	}


	/**
	 * Check if the list is currently empty.
	 * 
	 * @return true if there are no <code>NEPNeighbour</code> available; false
	 *         otherwise
	 */
	public boolean isEmpty(){

		return elements.isEmpty();
	}

}
