/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.utilities;

import java.util.Hashtable;

import ch.ethz.dcg.p2p.gui.GUI;

/**
 * Collection of statistical values about a Swistry peer.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class Statistic {

	private long totalPacketsReceived = 0;
	private long totalPacketsSent = 0;

	private long totalRequestsSent = 0;
	private long totalRequestsReceived = 0;

	private long totalNotificationsReceived = 0;
	private long totalNotificationsSent = 0;

	private long duplicatePacketsReceived = 0;
	private long duplicateNotificationsReceived = 0;

	private long unknownNotificationsReceived = 0;
	private long unknownRequestsReceived = 0;
	private long unknownPacketsReceived = 0;
	private long packetNotAvailable = 0;
	private long invalidNotificationsReceived = 0;
	private long invalidRequestsReceived = 0;
	private long invalidPacketsReceived = 0;

	private int averageRTT = Integer.MAX_VALUE;

	private int packetSize = -1;

	private GUI gui;

	private Hashtable<String, NeighbourStat> perNeighbourStats = new Hashtable<String, NeighbourStat>();


	/**
	 * Sets the gui.
	 * 
	 * @param gui
	 *            the gui to set
	 */
	public void setGui( GUI gui ){

		this.gui = gui;
	}


	/**
	 * Resets the statistic.
	 */
	public void reset(){

		totalPacketsReceived = 0;
		totalPacketsSent = 0;
		totalRequestsReceived = 0;
		totalRequestsSent = 0;
		totalNotificationsReceived = 0;
		totalNotificationsSent = 0;
		duplicatePacketsReceived = 0;
		duplicateNotificationsReceived = 0;
		averageRTT = Integer.MAX_VALUE;
		packetSize = -1;
	}


	/**
	 * Adds a neighbour to the statistic and insert it into the GUI.
	 * 
	 * @param neighbourKey
	 *            the key of the neighbour which is of the form ip:port
	 */
	public void addNeighbourStat( String neighbourKey ){

		perNeighbourStats.put( neighbourKey, new NeighbourStat() );
		if( gui != null ){
			gui.addNeighbour( neighbourKey );
		}
	}


	/**
	 * Removes a neighbour from the statistic and removes it from the GUI as
	 * well.
	 * 
	 * @param neighbourKey
	 *            the key of the neighbour which is of the form ip:port
	 */
	public void removeNeighbourStat( String neighbourKey ){

		perNeighbourStats.remove( neighbourKey );
		if( gui != null ){
			gui.removeNeighbour( neighbourKey );
		}
	}


	// PACKETS RECEIVED
	public long getTotalPacketsReceived(){

		return totalPacketsReceived;
	}


	public void incPacketsReceivedFrom( String neighbourKey ){

		if( perNeighbourStats.containsKey( neighbourKey ) ){
			Long n = perNeighbourStats.get( neighbourKey ).packetsReceived++;
			totalPacketsReceived++;
			if( gui != null ){
				gui.updatePacketsReceived( neighbourKey, n );
				gui.updateTotals();
			}
		}
		if( neighbourKey.equals( "" ) ){
			// packet should have come from the input adapter
			totalPacketsReceived++;
			if( gui != null ){
				gui.updateTotals();
			}
		}
	}


	// PACKETS SENT
	public long getTotalPacketsSent(){

		return totalPacketsSent;
	}


	public void incPacketsSentTo( String neighbourKey ){

		if( perNeighbourStats.containsKey( neighbourKey ) ){
			Long n = perNeighbourStats.get( neighbourKey ).packetsSent++;
			totalPacketsSent++;
			if( gui != null ){
				gui.updatePacketsSent( neighbourKey, n );
				gui.updateTotals();
			}
		}
	}


	// REQUESTS RECEIVED
	public long getTotalRequestsReceived(){

		return totalRequestsReceived;
	}


	public void incRequestsReceivedFrom( String neighbourKey ){

		if( perNeighbourStats.containsKey( neighbourKey ) ){
			Long n = perNeighbourStats.get( neighbourKey ).requestsReceived++;
			totalRequestsReceived++;
			if( gui != null ){
				gui.updateRequestsReceived( neighbourKey, n );
				gui.updateTotals();
			}
		}
	}


	// REQUESTS SENT
	public long getTotalRequestsSent(){

		return totalRequestsSent;
	}


	public void incRequestsSentTo( String neighbourKey ){

		if( perNeighbourStats.containsKey( neighbourKey ) ){
			Long n = perNeighbourStats.get( neighbourKey ).requestsSent++;
			totalRequestsSent++;
			if( gui != null ){
				gui.updateRequestsSent( neighbourKey, n + 1 );
				gui.updateTotals();
			}
		}
	}


	// NOTIFICATOINS RECEIVED
	public long getTotalNotificationsReceived(){

		return totalNotificationsReceived;
	}


	public void incNotificationsReceivedFrom( String neighbourKey ){

		if( perNeighbourStats.containsKey( neighbourKey ) ){
			Long n = perNeighbourStats.get( neighbourKey ).notificationsReceived++;
			totalNotificationsReceived++;
			if( gui != null ){
				gui.updateNotificationsReceived( neighbourKey, n + 1 );
				gui.updateTotals();
			}
		}
	}


	// NOTIFICATIONS SENT
	public long getTotalNotificationsSent(){

		return totalNotificationsSent;
	}


	public void incNotificationsSentTo( String neighbourKey ){

		if( perNeighbourStats.containsKey( neighbourKey ) ){
			Long n = perNeighbourStats.get( neighbourKey ).notificationsSent++;
			totalNotificationsSent++;
			if( gui != null ){
				gui.updateNotificationsSent( neighbourKey, n );
				gui.updateTotals();
			}
		}
	}


	// DUPLICATE PACKETS RECEIVED
	public long getDuplicatePacketsReceived(){

		return duplicatePacketsReceived;
	}


	public void incDuplicatePacketsReceived(){

		duplicatePacketsReceived++;
		if( gui != null ){
			gui.updateTotals();
		}
	}


	// DUPLICATE NOTIFICATIONS RECEIVED
	public long getDuplicateNotificationsReceived(){

		return duplicateNotificationsReceived;
	}


	public void incDuplicateNotificationsReceived(){

		duplicateNotificationsReceived++;
		if( gui != null ){
			gui.updateTotals();
		}
	}


	// RATIO
	public float getSendReceiveRatioFor( String neighbourKey ){

		if( perNeighbourStats.containsKey( neighbourKey ) ){
			return (float)perNeighbourStats.get( neighbourKey ).packetsReceived / Math.max( (float)perNeighbourStats.get( neighbourKey ).packetsSent, 1 );
		}else{
			return 0;
		}
	}


	// RTT
	public int getAverageRTT(){

		return averageRTT;
	}


	public void setRTTOf( String neighbourKey, int rtt ){

		if( perNeighbourStats.containsKey( neighbourKey ) ){
			perNeighbourStats.get( neighbourKey ).rtt = rtt;
			if( gui != null ){
				gui.updateNeighbourRTT( neighbourKey, rtt );
			}
		}
	}


	// PACKET SIZE
	public void setPacketSize( int packetSize ){

		this.packetSize = packetSize;
	}


	public int getPacketSize(){

		return packetSize;
	}


	// UNKNOWN NOTIFICATIONS
	public void incUnknownNotificationsReceived(){

		unknownNotificationsReceived++;
	}


	public long getUnknownNotificationsReceived(){

		return unknownNotificationsReceived;
	}


	// UNKNOWN REQUESTS
	public void incUnknownRequestsReceived(){

		unknownRequestsReceived++;
	}


	public long getUnknownRequestsReceived(){

		return unknownRequestsReceived;
	}


	// UNKNOWN PACKETS
	public void incUnknownPacketsReceived(){

		unknownPacketsReceived++;
	}


	public long getUnknownPacketsReceived(){

		return unknownPacketsReceived;
	}


	// UNAVAILABLE PACKETS
	public void incPacketNotAvailable(){

		packetNotAvailable++;
	}


	public long getPacketNotAvailable(){

		return packetNotAvailable;
	}


	// INVALID NOTIFICATIONS
	public void incInvalidNotificationReceived(){

		invalidNotificationsReceived++;
	}


	public long getInvalidNotificationReceived(){

		return invalidNotificationsReceived;
	}


	// INVALID REQUESTS
	public void incInvalidRequestReceived(){

		invalidRequestsReceived++;
	}


	public long getInvalidRequestReceived(){

		return invalidRequestsReceived;
	}


	// INVALID PACKETS
	public void incInvalidPacketReceived(){

		invalidPacketsReceived++;
	}


	public long getInvalidPacketReceived(){

		return invalidPacketsReceived;
	}


	// KB SENT
	public long getTotalKBytesSent(){

		if( packetSize > 0 ){
			return totalPacketsSent * packetSize / 1000;
		}else{
			return 0;
		}
	}


	// KB RECEIVED
	public long getTotalKBytesReceived(){

		if( packetSize > 0 ){
			return totalPacketsReceived * packetSize / 1000;
		}else{
			return 0;
		}
	}


	/**
	 * Per neighbour statistics.
	 * 
	 * @author Philip Frey (philip.frey@alumni.ethz.ch)
	 */
	class NeighbourStat {

		long packetsReceived = 0;
		long packetsSent = 0;
		long requestsReceived = 0;
		long requestsSent = 0;
		long notificationsReceived = 0;
		long notificationsSent = 0;
		int rtt = Integer.MAX_VALUE;
	}
}
