/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.utilities;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.SocketException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ch.ethz.dcg.p2p.peers.UDPDispatcher;

/**
 * Helper class to handle the UDP socket.
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class UDPSocketManager extends Thread {

	protected final Log logger = LogFactory.getLog( this.getClass() );

	private DatagramSocket udpSocket;
	private byte[] buffer;
	private DatagramPacket inPacket;
	private UDPDispatcher dispatcher;
	private boolean running = false;
	private int packetSize = 65500;


	/**
	 * Creates a new <code>UDPSocketManager</code>
	 * 
	 * @param port
	 *            the UDP port to listen on
	 * @param dispatcher
	 *            the dispatcher for incoming UDP messages
	 * @throws SocketException
	 *             if the <code>DatagramSocket</code> cannot be opened
	 */
	public UDPSocketManager( int port, UDPDispatcher dispatcher ) throws SocketException{

		udpSocket = new DatagramSocket( port );
		this.dispatcher = dispatcher;
	}


	/**
	 * Sends a UDP packet.
	 * 
	 * @param packet
	 *            the packet to be sent
	 * @throws IOException
	 *             if the packet cannot be sent
	 */
	public void sendPacket( DatagramPacket packet ) throws IOException{

		udpSocket.send( packet );
	}


	/**
	 * Starts this manager.<br>
	 * It waits for incoming UDP packets and forwards them to the dispatcher.
	 */
	public void run(){

		if( logger.isDebugEnabled() ){
			logger.debug( "UDPSocketManager running" );
		}

		running = true;
		while( running ){
			try{
				buffer = new byte[packetSize];
				inPacket = new DatagramPacket( buffer, buffer.length );
				udpSocket.receive( inPacket );
				if( ( packetSize == 65500 ) && ( inPacket.getData()[0] == ProtocolConstants.P2PPACKET ) ){
					// set correct packetSize as soon as the first P2PPACKET has
					// arrived
					packetSize = inPacket.getLength();
				}
				dispatcher.dispatchPacket( inPacket );
			}catch( IOException e ){
				if( logger.isErrorEnabled() ){
					logger.error( "cannot receive UDP packet (" + e.getMessage() + ")" );
				}
			}
		}
	}


	/**
	 * Closes this manager.<br>
	 * The UDP socket will be closed.
	 */
	public void close(){

		running = false;
		try{
			udpSocket.close();
		}catch( Exception e ){
		}
	}
}
