/**
 * Swistry is a 3rd generation Peer-to-Peer streaming system 
 * that supports live audio and video broadcasting over the Internet.
 * 
 * Copyright (C) 2006  Philip Frey
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 * 
 * Contact the author: philip.frey@alumni.ethz.ch
 */

package ch.ethz.dcg.p2p.utilities;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 * Several little functions that make life easier ;)
 * 
 * @author Philip Frey (philip.frey@alumni.ethz.ch)
 */
public class Utilities {

	/**
	 * Writes an IP address to the given OutputStream.
	 * 
	 * @param ip
	 *            IP to be written
	 * @param out
	 *            out stream to use
	 * @throws IOException
	 *             if the IP cannot be written to the out stream
	 */
	public static void writeIP( InetAddress ip, DataOutputStream out ) throws IOException{

		if( out == null ){
			throw new IOException( "target OutputStream not available." );
		}

		out.write( ip.getAddress() );

	}


	/**
	 * Reads an IP address from the given InputStream.
	 * 
	 * @param in
	 *            the in stream to read from
	 * @return an <code>InetAddress</code> object for the received IP address
	 * @throws IOException
	 *             if reading from the in stream failed
	 */
	public static InetAddress readIP( DataInputStream in ) throws IOException{

		if( in == null ){
			throw new IOException( "target InputStream not available." );
		}

		byte[] buffer = new byte[4];
		in.read( buffer );

		return InetAddress.getByAddress( buffer );

	}


	/**
	 * Converts a byte array into an <code>InetAddress</code> object
	 * 
	 * @param b
	 *            the byte array containing an IP address
	 * @return <code>InetAddress</code> object that represents the IP address
	 *         given in the byte array
	 * @throws UnknownHostException
	 *             if the IP address in the byte array is invalid
	 */
	public static InetAddress byteArrayToInetAddress( byte[] b ) throws UnknownHostException{

		return InetAddress.getByAddress( b );
	}


	/**
	 * Converts an <code>InetAddress</code> object into a byte array.
	 * 
	 * @param ip
	 *            the <code>InetAddress</code> object to be converted
	 * @return a byte array representing the IP address in the
	 *         <code>InetAddress</code> object
	 */
	public static byte[] InetAddressToByteArray( InetAddress ip ){

		return ip.getAddress();
	}


	/**
	 * Converst an int into a byte array of length 4; high byte first.
	 * 
	 * @param i
	 *            the int to convert
	 * @return a byte array representing the given int
	 */
	public static byte[] intToByteArray( int i ){

		byte[] b = new byte[4];
		b[0] = (byte)( ( i >>> 24 ) & 0xFF );
		b[1] = (byte)( ( i >>> 16 ) & 0xFF );
		b[2] = (byte)( ( i >>> 8 ) & 0xFF );
		b[3] = (byte)( ( i >>> 0 ) & 0xFF );
		return b;
	}


	/**
	 * Converts a byte array of length 4 into an int.
	 * 
	 * @param b
	 *            the byte array to be converted
	 * @return int representing the values stored in the byte array
	 */
	public static int byteArrayToInt( byte[] b ){

		return ( ( (int)b[0] << 24 ) + ( (int)( b[1] & 255 ) << 16 ) + ( (int)( b[2] & 255 ) << 8 ) + ( (int)( b[3] & 255 ) << 0 ) );
	}


	/**
	 * Converts a long into a byte array of length 8; high byte first.
	 * 
	 * @param l
	 *            the long value to be converted
	 * @return the byte array representing the long value
	 */
	public static byte[] longToByteArray( long l ){

		byte[] b = new byte[8];
		b[0] = (byte)( l >>> 56 );
		b[1] = (byte)( l >>> 48 );
		b[2] = (byte)( l >>> 40 );
		b[3] = (byte)( l >>> 32 );
		b[4] = (byte)( l >>> 24 );
		b[5] = (byte)( l >>> 16 );
		b[6] = (byte)( l >>> 8 );
		b[7] = (byte)( l >>> 0 );
		return b;
	}


	/**
	 * Converts a byte array of length 8 int a long.
	 * 
	 * @param b
	 *            the byte array to be converted
	 * @return a long representing the value stored in the byte array
	 */
	public static long byteArrayToLong( byte[] b ){

		return ( ( (long)b[0] << 56 ) + ( (long)( b[1] & 255 ) << 48 ) + ( (long)( b[2] & 255 ) << 40 ) + ( (long)( b[3] & 255 ) << 32 ) + ( (long)( b[4] & 255 ) << 24 ) + ( ( b[5] & 255 ) << 16 ) + ( ( b[6] & 255 ) << 8 ) + ( ( b[7] & 255 ) << 0 ) );
	}


	/**
	 * Gets the subarray of length len extracted from b starting at start. The
	 * elements are b[start]..b[start+len-1].
	 * 
	 * @param b
	 *            the target array
	 * @param start
	 *            start index
	 * @param len
	 *            length of the subarray
	 * @return the subarray
	 */
	public static byte[] subArray( byte[] b, int start, int len ){

		byte[] res = new byte[len];
		int i = 0;
		while( i < len ){
			res[i] = b[start + i];
			i++;
		}
		return res;
	}


	/**
	 * Gets the int value which is stored at bytes [start,start+len-1].
	 * 
	 * @param b
	 *            the target array
	 * @param start
	 *            start index
	 * @param len
	 *            length of the subarray
	 * @return the int stored in this subarray
	 */
	public static int getInt( byte[] b, int start, int len ){

		return byteArrayToInt( subArray( b, start, len ) );
	}


	/**
	 * Gets the long value which is stored at bytes [start,start+len-1].
	 * 
	 * @param b
	 *            the target array
	 * @param start
	 *            start index
	 * @param len
	 *            length of the subarray
	 * @return the long stored in this subarray
	 */
	public static long getLong( byte[] b, int start, int len ){

		return byteArrayToLong( subArray( b, start, len ) );
	}
}
